MODULE mo_aero_massfix

  ! *mo_aero_massfix* holds routines to correct for 
  !                   inconsistent tracer fields
  !                   in a mass conserving manner
  !
  ! Author:
  ! -------
  ! Philip Stier, MPI-Met, 2003-2004

  USE mo_kind,         ONLY: dp

REAL(dp), ALLOCATABLE :: pxtte_old(:,:,:) 

CONTAINS

  SUBROUTINE xt_borrow(kproma, kbdim,  klev,  klevp1, ktrac, krow, &
                       papp1,  paphp1,                             &
                       pxtm1,  pxtte                               )

    ! *xt_borrow* borrowing scheme to correct for negative 
    !             tracer masses conserveing mass within the 
    !             column
    !
    ! Author:
    ! -------
    ! Philip Stier, MPI-Met,         06/2004
    !
    ! Method:
    ! -------
    ! If negative tracer values occur the scheme is iteratively  
    ! borrowing tracer mixing ratios from the grid-box below.
    ! If the lowest layer can not compensate for accumulated remaining
    ! corrections, the procedure is repeated from bottom to top.
    !
    ! Restrictions:
    ! -------------
    ! Columns with a negative total integrated tracer 
    ! content are set to zero despite the associated mass
    ! error.

    USE mo_constants,    ONLY: g
    USE mo_time_control, ONLY: time_step_len, delta_time
    USE mo_exception,    ONLY: finish
    USE mo_tracer,       ONLY: trlist

    IMPLICIT NONE

    INTEGER, INTENT(IN)    :: kproma, kbdim, klev, klevp1, ktrac, krow
    REAL(dp),INTENT(IN)    :: papp1(kbdim,klev), paphp1(kbdim,klevp1)
    REAL(dp),INTENT(IN)    :: pxtm1(kbdim,klev,ktrac)
    REAL(dp),INTENT(INOUT) :: pxtte(kbdim,klev,ktrac)

    !--- Local variables:

    INTEGER :: jl, jk, jt

    REAL(dp):: zxtp1, ztmst, zeps

    REAL(dp):: zxtbor(kbdim)

    REAL(dp):: zdpg(kbdim,klev)


    !--- 0) Initializations:

    ztmst=time_step_len
    zeps=10._dp*EPSILON(1.0_dp)

    !--- 1) Calculate auxiliary variable dp/g :

    !--- Uppermost level:

    zdpg(1:kproma,1)=2._dp*(paphp1(1:kproma,2)-papp1(1:kproma,1))/g

    !--- Other levels:

    DO jk=2, klev
     
       zdpg(1:kproma,jk)=(paphp1(1:kproma,jk+1)-paphp1(1:kproma,jk))/g

    END DO

    !--- 2) Borrowing scheme:

    DO jt=1, ktrac

       !--- Start borrowing scheme:

       !--- Integrate from top to bottom:

       zxtbor(1:kproma)=0.0_dp
          
       DO jk=1, klev
          DO jl=1, kproma

             !--- Check if corrected mass, including the fix of the layer above,
             !    yields a negative mixing ratio:
             !    (Convert mass correction converted to [kg kg-1] in current layer)

             zxtp1=pxtm1(jl,jk,jt)+pxtte(jl,jk,jt)*ztmst + zxtbor(jl)/zdpg(jl,jk)

             IF ( zxtp1 > 0.0_dp ) THEN

                !--- Subtract corrected tracer mass from current layer:
                !    (zxtbor is negative)

                pxtte(jl,jk,jt)=pxtte(jl,jk,jt)+zxtbor(jl)/(zdpg(jl,jk)*ztmst)

                !--- Reset mass correction:

                zxtbor(jl)=0.0_dp

             ELSE

                !--- Adjust tendency to yield zero:

                pxtte(jl,jk,jt)=-pxtm1(jl,jk,jt)/ztmst

                !--- Add correcting mass mixing ratio and convert to [kg m-2]:
                !    (implicit summation due to the inclusion
                !     of zxtbor(jk-1) in zxtp1)

                zxtbor(jl)=zxtp1*zdpg(jl,jk)

             END IF

          END DO
       END DO

       !--- If surface layer cannot compensate accumulated correction:
       !    Iterate from bottom to top
          
       DO jk=klev, 1, -1
          DO jl=1, kproma
             IF (zxtbor(jl) < -zeps) THEN

                !--- Check if corrected mass, including the fix of the layer below,
                !    yields a negative mixing ratio:
                !    (Convert mass correction converted to [kg kg-1] in current layer)

                zxtp1=pxtm1(jl,jk,jt)+pxtte(jl,jk,jt)*ztmst + zxtbor(jl)/zdpg(jl,jk)

                IF ( zxtp1 > 0.0_dp ) THEN

                   !--- Subtract corrected tracer mass from current layer:
                   !    (zxtbor is negative)

                   pxtte(jl,jk,jt)=pxtte(jl,jk,jt)+zxtbor(jl)/(zdpg(jl,jk)*ztmst)

                   !--- Reset mass correction:

                   zxtbor(jl)=0.0_dp

                ELSE

                   !--- Adjust tendency to yield zero:

                   pxtte(jl,jk,jt)=-pxtm1(jl,jk,jt)/ztmst

                   !--- Add correcting mass mixing ratio and convert to [kg m-2]:
                   !    (implicit summation due to the inclusion
                   !     of zxtbor(jk-1) in zxtp1)

                   zxtbor(jl)=zxtp1*zdpg(jl,jk)

                END IF

             END IF
          END DO
       END DO

    END DO

  END SUBROUTINE xt_borrow


  SUBROUTINE xt_massfix(kproma, kbdim,  klev,  klevp1, ktrac, krow, &
                        papp1,  paphp1, pxtte, loini,  cxtbound     )

    ! *xt_massfix* corrects the tendencies of each column to 
    !              conserve mass
    !
    ! Author:
    ! -------
    ! Philip Stier, MPI-Met,         11/2003
    !
    ! Method:
    ! -------
    ! To keep the model as closely as possible to the intended physical
    ! tendency changes, the correction of the tendencies is 
    ! imposed proportionally to the absolute values of the 
    ! applied tendency in each layer. 
    !
    ! Total mass error tendency [kg m-2 s-1]:
    !
    !    dxtdt(jl)=SUM(xtte(jl,jk)*dpg(jl,jk))+xtbound(jl,jk)
    !
    !    where xtbound is a boundary condition, i.e. deposition flux
    !
    ! Correction proportional to tendency [kg m-2 s-1]:
    !
    !                         - ABS(xtte(jl,jk)*dpg(jl,jk))
    !    dxtfix(jl,jk)   = --------------------------------- * dxdt(jl)
    !                       SUM(ABS(xtte(jl,jk)*dpg(jl,jk)))
    !
    ! Resulting in a corrective tendency [kg kg-1 s-1]:
    !
    !    zxttefix(jl,jk) = dxtfix(jl,jk) / dpg(jl,jk)
    !
    !
    ! Arguments:
    ! ----------
    !
    ! pxtte     = cumulative tendency for all tracers [kg kg-1 s-1]
    ! cxtbound  = name of stream holding the boundary condition, e.g.
    !             deposition flux [kg m-2 s-1]
    !
    ! Usage:
    ! ------
    ! Call twice: first with loini=.TRUE. to store the old tendency
    !             then to fix the mass of processes that have modified
    !             pxtte meanwhile with the current pxtte and loini=.FALSE. 

    USE mo_constants,   ONLY: g
    USE mo_aero_mem,    ONLY: aero
    USE mo_memory_base, ONLY: get_stream_element
    USE mo_tracer,      ONLY: trlist, ntrac ! uh: ntrac added

!uh+:
    USE mo_be10,        ONLY: be10_wdep_corr, BE_FLAG
    USE mo_aero_m7,     ONLY: nmod
    USE mo_control,     ONLY: lcn
! uh-


    IMPLICIT NONE


    !--- Arguments:

    INTEGER,      INTENT(IN)    :: kproma, kbdim, klev, klevp1, ktrac, krow

    LOGICAL,      INTENT(IN)    :: loini

    CHARACTER(*), INTENT(IN)    :: cxtbound

    REAL(dp),     INTENT(IN)    :: papp1(kbdim,klev), paphp1(kbdim,klevp1)
    REAL(dp),     INTENT(INOUT) :: pxtte(kbdim,klev,ktrac)

    !--- Local Variables:

    INTEGER :: jl, jk, jt

    REAL(dp):: zeps, zxttefix

!uh+:
    REAL(dp)    :: zfix(kbdim,klev,ntrac)

! uh-


    REAL(dp):: zdxtdt(kbdim), zdxtdtsum(kbdim) 

    REAL(dp):: zxtte(kbdim,klev), zdpg(kbdim,klev)

    REAL(dp), POINTER :: xtbound_p(:,:)


    !--- 1) Initialization mode:


    IF (loini) THEN

       ALLOCATE (pxtte_old(kbdim,klev,ktrac))

       pxtte_old=pxtte

       RETURN

    END IF

    !--- 2) Mass fix mode:

    zeps=EPSILON(1.0_dp) 

    !--- 1.1) Calculate auxiliary variable dp/g :

    !--- Uppermost level:

    zdpg(1:kproma,1)=2._dp*(paphp1(1:kproma,2)-papp1(1:kproma,1))/g

    !--- Other levels:

    DO jk=2, klev
     
       zdpg(1:kproma,jk)=(paphp1(1:kproma,jk+1)-paphp1(1:kproma,jk))/g

    END DO

    DO jt=1, ktrac
!       IF(trlist%ti(jt)%nwetdep >0) THEN
! uh: changed -> no mass fixing for be10

       IF(trlist%ti(jt)%nwetdep >0.AND. trlist%ti(jt)%nwetdep < BE_FLAG) THEN


          zdxtdt(1:kproma)   =0.0_dp
          zdxtdtsum(1:kproma)=0.0_dp

          CALL get_stream_element(aero,TRIM(cxtbound)//'_'//TRIM(trlist%ti(jt)%fullname), xtbound_p)

          !--- Accumulated tendency since initialization:

          zxtte(1:kproma,:)=pxtte(1:kproma,:,jt)-pxtte_old(1:kproma,:,jt)

          !--- Calculate vertically integrated mass error tendency [kg m-2 s-1]:

          DO jk=1, klev
             DO jl=1, kproma

                zdxtdt(jl)=zdxtdt(jl)      +zxtte(jl,jk)*zdpg(jl,jk)

                zdxtdtsum(jl)=zdxtdtsum(jl)+ABS(zxtte(jl,jk)*zdpg(jl,jk))

             END DO
          END DO

          zdxtdt(1:kproma)=zdxtdt(1:kproma)+xtbound_p(1:kproma,krow)

          DO jk=1, klev
             DO jl=1, kproma
                IF (ABS(zdxtdtsum(jl)) > zeps) THEN

                   zxttefix=-((ABS(zxtte(jl,jk)*zdpg(jl,jk))/zdxtdtsum(jl))*zdxtdt(jl))/zdpg(jl,jk)

                   pxtte(jl,jk,jt)=pxtte(jl,jk,jt)+zxttefix
                   
!uh
!                   IF (trlist%ti(jt)%nphase==AEROSOLMASS) THEN

                   zfix(jl,jk,jt)=zxttefix

!                   END IF
                ELSE

!                   IF (trlist%ti(jt)%nphase==AEROSOLMASS) THEN
                   zfix(jl,jk,jt)=0._dp
!
!                   END IF
!uh-


                END IF
             END DO ! kproma
          END DO ! klev

       END IF ! nwetdep
    END DO ! ktrac

    DEALLOCATE(pxtte_old)

! uh+
    IF(lcn.AND..NOT.loini) CALL be10_wdep_corr(kproma,kbdim,klev,krow,zfix)
! uh-


  END SUBROUTINE xt_massfix


END MODULE mo_aero_massfix
