 MODULE mo_bgc_dust_emis
 
  ! Description:
  ! ------------
  ! *mo_bgc_dust_emis* handles all routines for MPI BGC dust emission scheme.
  ! The public parameters and variables of the scheme are found in *mo_bgc_dust_mem*.
  !
  ! Author:
  ! -------
  ! Martin Werner, MPI Biogeochemistry (MPI BGC), Jena, October 2002.
  ! Martin Werner, MPI Biogeochemistry (MPI BGC), Jena, October 2004:
  !                changed scheme from 0.5x0.5 degree grid to default
  !                Echam5-HAM grid (to save computational costs)
  !
  ! Interface:
  ! ----------
  ! The routines for reading in necessary input files and for calculating the emissions 
  ! are called from the routine *call_read_forcing* in file *call_submodels.f90*.
  !
  ! Method:
  ! -------
  ! *mo_bgc_dust_emis* holds the various routines for the MPI BGC dust emission scheme, while 
  ! public parameters and variables are stored in *mo_bgc_dust_mem*.
  !
  ! The stream *emi_du_src* of the HAM model is used to store the calculated dust emissions after
  ! interpolation to the defined ECHAM resolution is performed. The routine *dust_emissions_bgc*
  ! (called by routine *call_read_forcing* in file *call_submodels.f90*) links the MPI BGC emissions
  ! to the HAM aerosol model developed by Philip Stier et al., MPI Meteorology, Hamburg.
  
  USE mo_kind, ONLY: dp
  USE mo_mpi,  ONLY: p_pe, p_io

  IMPLICIT NONE

CONTAINS

  SUBROUTINE bgc_read_annual_fields

    ! Description:
    ! ------------
    ! *bgc_read_annual_fields* reads the prescribed annual field of preferential dust source area
    !
    ! Authors:
    ! --------
    ! M. Werner, MPI BGC, November 2002: original source
    ! P. Stier,  MPI Met, January  2004: replaced by fast NetCDF read-in
    ! M. Werner, MPI BGC, October  2004: changed emission cheme from 0.5x0.5 degree grid to 
    !                                    default Echam5-HAM grid (to save computational costs)
    !                                 => annual fields of different biome types are not read
    !                                    in this version any longer but may be added again in future releases
    !
    ! Interface:
    ! ----------
    ! *bgc_read_annual_fields* is called from *call_read_forcing* in file *call_submodels.f90*.
    !
    ! Method:
    ! -------
    ! Reads biome and soil fields specified in the three different namlists 
    ! *biome_fields.nml*, *soil_types.nml*, *pot_sources.nml*
    ! and stores them in the corresponding variables defined in the module
    ! *mo_bgc_dust_mem*. 
    !
    ! This routine is based on *xt_read_emiss_fields* by Philip Stier,  MPI-Met, HH.

    USE mo_exception,         ONLY: finish
    USE mo_control,           ONLY: nlon, ngl
    USE mo_doctor,            ONLY: nout
    USE mo_mpi,               ONLY: p_parallel_io, p_io
    USE mo_bgc_dust_mem,      ONLY: biome,                    &
                                    mat_s1, mat_s2, mat_s3,   &
                                    mat_s4, mat_s6, mat_psrc
    USE mo_read_netcdf77,     ONLY: read_var_nf77_2d
    USE mo_decomposition,     ONLY: lc => local_decomposition, global_decomposition
    USE mo_transpose,         ONLY: scatter_gp

    IMPLICIT NONE

    LOGICAL        :: lex           ! flag to test if file exists
    INTEGER        :: ierr
    
    CHARACTER(512) :: cfile         ! filename

    REAL(dp), ALLOCATABLE, TARGET :: zin(:,:)
    REAL(dp), POINTER :: gl_data(:,:)


    IF (.NOT. ALLOCATED(biome))    ALLOCATE (biome(nlon, ngl))
    IF (.NOT. ALLOCATED(mat_s1))   ALLOCATE (mat_s1(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(mat_s2))   ALLOCATE (mat_s2(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(mat_s3))   ALLOCATE (mat_s3(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(mat_s4))   ALLOCATE (mat_s4(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(mat_s6))   ALLOCATE (mat_s6(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(mat_psrc)) ALLOCATE (mat_psrc(lc%nproma, lc%ngpblks))
    
    IF (p_parallel_io) THEN
      ALLOCATE (zin(lc%nlon,lc%nlat))
    END IF

!    IF (p_parallel_io) THEN
!
!       !--- Read biome-fields on 720x360 degree grid:
!
!       cfile='biome4.nc'
!
!       WRITE(nout,*) ' reading biome-fields ', TRIM(cfile)
!
!       INQUIRE (file=TRIM(cfile) ,exist=lex)
!
!       IF (lex) THEN
!
!          CALL read_var_nf77_2d (cfile, "longitude", "latitude",  &
!                                 "biome",       biome,      ierr  )
!         
!       ELSE
!
!          CALL finish('bgc_read_annual_fields','missing input file '//TRIM(cfile))
!
!       END IF
!
!   END IF

    !--- Set biome type distribution:

    biome(:,:)=21._dp      ! Biome type is set to dessert (= biome type 21) for all grid points

    !--- Set initial soil type distribution:

    mat_s1(:,:)=1._dp      ! Soil type is initialized to coarse texture (= soil type 1) for all grid points

    !--- Read other soil type distributions:

    IF (p_parallel_io) THEN
       cfile='soil_type2.nc'
       WRITE(nout,*) ' reading soil type #2 distribution ', TRIM(cfile)
       INQUIRE (file=TRIM(cfile) ,exist=lex)
       IF (lex) THEN
          CALL read_var_nf77_2d (cfile,  "lon", "lat", "type", zin, ierr)
       ELSE
          CALL finish('bgc_read_annual_fields','missing input file '//TRIM(cfile))
       END IF
    END IF

    NULLIFY (gl_data)
    IF (p_pe == p_io) gl_data => zin(:,:)
    CALL scatter_gp (gl_data, mat_s2(:,:), global_decomposition)

    IF (p_parallel_io) THEN
       cfile='soil_type3.nc'
       WRITE(nout,*) ' reading soil type #3 distribution ', TRIM(cfile)
       INQUIRE (file=TRIM(cfile) ,exist=lex)
       IF (lex) THEN
          CALL read_var_nf77_2d (cfile,  "lon", "lat", "type", zin, ierr)
       ELSE
          CALL finish('bgc_read_annual_fields','missing input file '//TRIM(cfile))
       END IF
    END IF

    NULLIFY (gl_data)
    IF (p_pe == p_io) gl_data => zin(:,:)
    CALL scatter_gp (gl_data, mat_s3(:,:), global_decomposition)

    IF (p_parallel_io) THEN
       cfile='soil_type4.nc'
       WRITE(nout,*) ' reading soil type #4 distribution ', TRIM(cfile)
       INQUIRE (file=TRIM(cfile) ,exist=lex)
       IF (lex) THEN
          CALL read_var_nf77_2d (cfile,  "lon", "lat", "type", zin, ierr)
       ELSE
          CALL finish('bgc_read_annual_fields','missing input file '//TRIM(cfile))
       END IF
    END IF

    NULLIFY (gl_data)
    IF (p_pe == p_io) gl_data => zin(:,:)
    CALL scatter_gp (gl_data, mat_s4(:,:), global_decomposition)

    IF (p_parallel_io) THEN
       cfile='soil_type6.nc'
       WRITE(nout,*) ' reading soil type #6 distribution ', TRIM(cfile)
       INQUIRE (file=TRIM(cfile) ,exist=lex)
       IF (lex) THEN
          CALL read_var_nf77_2d (cfile,  "lon", "lat", "type", zin, ierr)
       ELSE
          CALL finish('bgc_read_annual_fields','missing input file '//TRIM(cfile))
       END IF
    END IF

    NULLIFY (gl_data)
    IF (p_pe == p_io) gl_data => zin(:,:)
    CALL scatter_gp (gl_data, mat_s6(:,:), global_decomposition)

       !--- Read potential sources distribution:

    IF (p_parallel_io) THEN
       cfile='pot_sources.nc'
       WRITE(nout,*) ' reading potential sources distribution ', TRIM(cfile)
       INQUIRE (file=TRIM(cfile) ,exist=lex)
       IF (lex) THEN
          CALL read_var_nf77_2d (cfile,    "lon", "lat", "source", zin, ierr)
       ELSE
          CALL finish('bgc_read_annual_fields','missing input file '//TRIM(cfile))
       END IF
    END IF

    NULLIFY (gl_data)
    IF (p_pe == p_io) gl_data => zin(:,:)
    CALL scatter_gp (gl_data, mat_psrc(:,:), global_decomposition)

    IF (p_parallel_io) THEN
       DEALLOCATE (zin)
    ENDIF

  END SUBROUTINE bgc_read_annual_fields


  SUBROUTINE bgc_read_fpar_field(kdate)

    ! Description:
    ! ------------
    ! *bgc_read_fpar_field* reads the effective FPAR field (NDVI satellite data or KERNEL output)
    !
    ! Authors:
    ! --------
    ! M. Werner, MPI BGC, November 2002: original source
    ! P. Stier,  MPI Met, January  2004: replaced by fast NetCDF read-in
    ! M. Werner, MPI BGC, October  2004: changed emission cheme from 0.5x0.5 degree grid to 
    !                                    default Echam5-HAM grid (to save computational costs)
    !
    ! Interface:
    ! ----------
    ! *bgc_read_fpar_field* is called from *call_read_forcing* in file *call_submodels.f90*.
    !
    ! Method:
    ! -------
    ! Reads daily (or monthly) effective FPAR fields specified in the namlist 
    ! *fpar_eff.nml* and stores it in the variable *k_fpar_eff* defined in the module
    ! *mo_bgc_dust_mem*. 
    !

    USE mo_doctor,       ONLY: nout
    USE mo_exception,    ONLY: finish
    USE mo_mpi,          ONLY: p_parallel_io, p_io
    USE mo_bgc_dust_mem, ONLY: k_fpar_eff
    USE mo_read_netcdf77,   ONLY: read_var_hs_nf77_2d
    USE mo_decomposition,     ONLY: lc => local_decomposition, global_decomposition
    USE mo_transpose,         ONLY: scatter_gp

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: kdate
    INTEGER             :: ierr
    LOGICAL             :: lex    ! flag to test if file exists
    CHARACTER(512)      :: cfile

    REAL(dp), ALLOCATABLE, TARGET :: zin(:,:)
    REAL(dp), POINTER :: gl_data(:,:)


    IF (.NOT. ALLOCATED(k_fpar_eff)) ALLOCATE (k_fpar_eff(lc%nproma, lc%ngpblks))

    IF (p_parallel_io) THEN
      ALLOCATE (zin(lc%nlon,lc%nlat))
    END IF

    IF (p_parallel_io) THEN

       !--- Read biome-fields on 720x360 degree grid:

       cfile='ndvi_lai_eff.12m.nc'

       WRITE(nout,*) ' reading LAI fields ', TRIM(cfile)

       INQUIRE (file=TRIM(cfile) ,exist=lex)

       IF (lex) THEN

          CALL read_var_hs_nf77_2d (cfile, "lon", "lat", "time", &
                                    kdate, "laieff",    zin, ierr    )

       ELSE

          CALL finish('bgc_read_fpar_field','missing input file '//TRIM(cfile))

       END IF

    END IF

    NULLIFY (gl_data)
    IF (p_pe == p_io) gl_data => zin(:,:)
    CALL scatter_gp (gl_data, k_fpar_eff(:,:), global_decomposition)

    IF (p_parallel_io) THEN
       DEALLOCATE (zin)
    ENDIF

  END SUBROUTINE bgc_read_fpar_field
  

  SUBROUTINE bgc_dust_initialize

    ! Description:
    ! ------------
    ! *bgc_dust_initialize* initializes several parameters and variables of the MPI BGC dust emission scheme.
    ! It also performs the wind independent calculations of threshold friction velocity and soil particle
    ! distribution. The effective soil fraction for dust deflation is determined, too.
    !
    ! Authors:
    ! --------
    ! M. Werner, MPI BGC, November 2002
    ! P. Stier, MPI MET,  June     2004: added scale factor for wind stress threshold
    ! M. Werner, MPI BGC, October  2004: changed emission cheme from 0.5x0.5 degree grid to 
    !                                    default Echam5-HAM grid (to save computational costs)
    !
    ! Interface:
    ! ----------
    ! *bgc_dust_initialize* is called from *call_read_forcing* in file *call_submodels.f90*.
    !
    ! This routine is based on the offline dust emission scheme by Ina Tegen, MPI BGC.
    
    USE mo_constants,     ONLY: api, g
    USE mo_control,       ONLY: nlon, ngl
    USE mo_decomposition, ONLY: lc => local_decomposition
    USE mo_bgc_dust_data, ONLY: active, solspe
    USE mo_bgc_dust_mem,  ONLY: biome,                            &
                                c_eff, idust,                     &
                                nats, nclass, dmin, dmax,         &
                                a_rnolds, b_rnolds, x_rnolds,     &
                                roa, rop, d_thrsld, dstep,        &
                                nmode, z01, z02, z0s, aeff,       &
                                xeff,                             &
                                flux_ann,                         &
                                flux_a1,flux_a2,flux_a10,         &
                                srel, srelV, su_srelV, Uth,       &
!                                duststat, duststat2,              &
                                cuscale

    IMPLICIT NONE
    
    REAL(dp), PARAMETER :: gravi=g*100._dp              ! gravity acceleration in cm/s2

    REAL(dp)            :: AAA, BB, CCC, DDD, EE, FF ! tool variables
    
    REAL(dp)            :: d1                        ! distance between obstacles
    REAL(dp)            :: dlastj(nclass)
    REAL(dp)            :: rdp

    REAL(dp)            :: feff
      
    INTEGER         :: i, j                      ! do indices
    INTEGER         :: kk                        ! do index
    INTEGER         :: nd, nsi, np
    INTEGER         :: ns                        ! do index

    REAL(dp)            :: rsize(Nclass)              ! granulometric class size
    REAL(dp)            :: stotal                    ! total surface
    REAL(dp)            :: stotalV
    REAL(dp)            :: su                        ! surface 
    REAL(dp)            :: suV
    REAL(dp)            :: su_loc                    ! surface
    REAL(dp)            :: su_locV
    REAL(dp)            :: su_class(nclass)          ! surface occupied by each granulometric class
    REAL(dp)            :: su_classV(nclass)
    REAL(dp)            :: sum_srel(nats,nclass)     ! srel accumulated

    REAL(dp)            :: utest(nats)

    INTEGER         :: vgtp        

    REAL(dp)            :: xnV
    REAL(dp)            :: xk,xl,xm,xn               ! tool variables


!    IF (.NOT. ALLOCATED(duststat))  ALLOCATE (duststat(nlon, ngl, 2))
!    IF (.NOT. ALLOCATED(duststat2)) ALLOCATE (duststat2(nlon, ngl, 2))
    IF (.NOT. ALLOCATED(flux_ann))  ALLOCATE (flux_ann(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(flux_a1))   ALLOCATE (flux_a1(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(flux_a2))   ALLOCATE (flux_a2(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(flux_a10))  ALLOCATE (flux_a10(lc%nproma, lc%ngpblks))
    IF (.NOT. ALLOCATED(idust))     ALLOCATE (idust(nlon, ngl))


!    duststat = 0._dp                      ! initializations
!    duststat2 = 0._dp
    flux_ann(:,:) = 0._dp
    flux_a1(:,:) = 0._dp
    flux_a2(:,:) = 0._dp
    flux_a10(:,:) = 0._dp
    rsize(:) = 0._dp
    srel(:,:) = 0._dp
    srelV(:,:) = 0._dp
    sum_srel(:,:) = 0._dp
    su_srelV(:,:) = 0._dp
    Uth(:) = 0._dp
    utest(:) = 0._dp

    i = 0                              ! calculation of the threshold friction velocity Uth
    rdp = Dmin
    DO WHILE(rdp.lE.Dmax + 1.E-5_dp)
     i = i + 1
     rsize(i) = rdp
     BB = a_rnolds * (rdp ** x_rnolds) + b_rnolds
     XK = SQRT(rop * gravi * rdp / roa)
     CCC = SQRT(1._dp + d_thrsld /(rdp ** 2.5_dp))
     IF (BB.LT.10._dp) THEN
      DDD=SQRT(1.928_dp * (BB ** 0.092_dp) - 1._dp)
      Uth(i) = 0.129_dp * XK * CCC / DDD
     ELSE
      EE = -0.0617_dp * (BB - 10._dp) 
      FF = 1._dp -0.0858_dp * EXP(EE)
      Uth(i) = 0.12_dp * XK * CCC * FF       
     ENDIF
     rdp = rdp * EXP(Dstep)   
    END DO      

    Uth(:)=Uth(:)*cuscale ! scale wind stress threshold

    DO ns = 1,nats ! loop over all soil types
       
     rdp = Dmin     ! calculation of the soil particle distribution and related surfaces
     kk = 0
     Stotal = 0._dp
     StotalV = 0._dp
     su_class(:) = 0._dp
     su_classV(:) = 0._dp
     DO WHILE (rdp.LE.Dmax+1.E-5_dp)        ! surface calculations
      kk = kk + 1
      su = 0._dp
      suV = 0._dp
      DO i = 1, Nmode
       nd  = ((i - 1) *3 ) + 1
       nsi = nd + 1
       np  = nd + 2
       IF (solspe(ns,nd).EQ.0._dp) THEN            
        su_loc = 0._dp
        su_locV=0._dp
       ELSE
        xk = solspe(ns,np)/(sqrt(2._dp* api)*log(solspe(ns,nsi)))
        xl = ((log(rdp)-log(solspe(ns,nd)))**2)/(2._dp*(log(solspe(ns,nsi)))**2)
        xm = xk * exp(-xl)
        xn =  rop*(2._dp/3._dp)*(rdp/2._dp) !surface
        xnV =  1._dp !volume
        su_loc = (xm*Dstep/xn)       
        su_locV = (xm*Dstep/xnV)     
       ENDIF !
       su = su + su_loc
       suV = suV + su_locV
      END DO !Nmode
      su_class(kk) = su
      su_classV(kk) = suV
      Stotal = Stotal + su
      StotalV = StotalV + suV
      dlastj(kk)=rdp*5000._dp
      rdp = rdp * exp(Dstep)
     END DO !rdp
         
     do j = 1,Nclass
      if (Stotal.eq.0._dp)then
       srel(ns,j) = 0._dp
       srelV(ns,j) = 0._dp
      else
       srel(ns,j) = su_class(j)/Stotal
       srelV(ns,j) = su_classV(j)/StotalV
       utest(ns)=utest(ns)+srelV(ns,j)
       su_srelV(ns,j)=utest(ns) 
      endif
     end do !j=1,nclass
         
    end do !ns (soil type)


    d1   = 0._dp      ! initializations                 
    feff = 0._dp
    AAA=0._dp
    BB=0._dp
    CCC=0._dp
    DDD=0._dp
    EE=0._dp
    FF=0._dp

    if (Z01.EQ.0._dp) then                ! calculation of the effective fraction
     feff = 0._dp
    else
! partition of energy between the surface and the elements of rugosity (Marticorena thesis, pp 111-112)
     AAA = log(Z01/Z0S)
     BB = log(aeff*(xeff/Z0S)**0.8_dp)
     CCC = 1._dp- AAA/BB
     if (d1.eq.0._dp) then                ! partition between Z01 and Z02
      FF = 1._dp
     else
      DDD = log(Z02/Z01)
      EE = log(aeff * (d1/Z01)**0.8_dp)
      FF = 1._dp- DDD/EE
     endif
     feff = FF*CCC                     ! total effective fraction   
     if (feff.lt.0._dp) feff=0._dp
     if (feff.gt.1._dp) feff=1._dp
    endif       
          
    C_eff = feff
      
    do j = 1,ngl                       ! loop over all latitudes
     do i = 1,nlon                     ! loop over all longitudes
      idust(i,j) = 0                   ! check if vegetation type allows dust deflation
      vgtp = int(biome(i,j))           ! (idust=0: no deflation allowed; idust=1: deflation allowed)
      if (vgtp.GT.0) idust(i,j) = active(vgtp)
     enddo !i
    enddo !j
    
  END SUBROUTINE bgc_dust_initialize


  SUBROUTINE bgc_dust_calc_emis(kproma, krow)

    ! Description:
    ! ------------
    ! *bgc_dust_calc_emis* calculates online global dust emission fields
    ! using the MPI BGC dust emissions scheme.
    !
    ! Authors:
    ! --------
    ! M. Werner, MPI BGC, October 2002
    ! P. Stier, MPI MET,  June    2004: added scale factor for wind stress threshold
    ! M. Werner, MPI BGC, October 2004: changed emission cheme from 0.5x0.5 degree grid to 
    !                                   default Echam5-HAM grid (to save computational costs)
    ! L. Kornblueh, MPI MET, July 2006: fix bug with respect to subscriptin in flux_6h
    !
    ! Interface:
    ! ----------
    ! The routine is called from *dust_emissions_bgc* in module *mo_bgc_dust_emis.f90*.
    !
    ! Method:
    ! -------
    ! The emissions are calculated every model time step using ECHAM 10m wind speeds.
    !
    ! This version of the emission scheme calculates dust emssissions for different bins, controlled by 
    ! dmin:   minimum particle diameter, 
    ! nbin:   number of binned size fractions, 
    ! ntrace: maximum number of bins, 
    ! dbmin(ntrace) & dbmax(ntrace): diameter limits of each bin
    !
    ! This routine is based on the offline dust emission scheme by Ina Tegen, MPI BGC.

    USE mo_constants,     ONLY: g
    USE mo_memory_g3b,    ONLY: slm, glac, vlt, wl, ws, wsmx, sn, snc, orostd
    USE mo_physc2,        ONLY: cqsncr, cwlmax
    USE mo_time_control,  ONLY: delta_time
    USE mo_test_trans,    ONLY: test_gridpoint
    USE mo_decomposition, ONLY: lc => local_decomposition

    USE mo_aero_mem,      ONLY: velo10m, emi_du_mai, emi_du_mci, ustar_acrit
  
    USE mo_bgc_dust_data, ONLY: solspe
    USE mo_bgc_dust_mem,  ONLY: k_fpar_eff,                          &
                                mat_s1, mat_s2, mat_s3,              &
                                mat_s4, mat_s6, mat_psrc,            &
                                c_eff, idust,                        &
                                flux_6h, flux_ann,                   &
                                flux_a1,flux_a2,flux_a10,            &
                                srel, srelV, su_srelV, Uth,          &
                                ntrace, nclass,                      &
                                vk, zz, z02, umin, dmin,             &
                                nmode, dstep, nbin, w0, roa,         &
!                                duststat, duststat2,                 &
                                min_ai, max_ai, min_ci, max_ci,      &
                                cuscale

    IMPLICIT NONE

! Parameter
    INTEGER, INTENT(in)      :: kproma, krow
    REAL(dp), PARAMETER          :: cd=1.00_dp*roa/(g*100._dp) ! flux dimensioning parameter

    REAL(dp), PARAMETER          :: zepsec=1.E-12_dp    ! ECHAM parameter for snow cover calculation
    REAL(dp), PARAMETER          :: zsigfac=0.15_dp     ! ECHAM parameter for snow cover calculation
      
! Variables
    REAL(dp)           :: alpha

    REAL(dp)           :: dbmin(ntrace)              ! bin size limits
    REAL(dp)           :: dbmax(ntrace)              ! bin size limits
    REAL(dp)           :: dlast                      ! bin size limits
    REAL(dp)           :: rdp
    REAL(dp)           :: dpk(ntrace)
    INTEGER        :: dn(kproma)               ! number of dislocated particle classes
    INTEGER        :: dk(kproma,nclass)        ! dislocated particle classes
    
    REAL(dp)           :: du_snow(kproma)          ! ECHAM snow coverage interpolated to 0.5x0.5 grid resolution
    REAL(dp)           :: du_W1r(kproma)           ! ECHAM relative soil humidity interpolated to 0.5x0.5 grid resolution
    REAL(dp)           :: du_wind(kproma)          ! ECHAM wind field interpolated to 0.5x0.5 grid resolution

    INTEGER        :: dust_mask(kproma)        ! grid point mask for dust emissions (0: no emission)

    REAL(dp)           :: fluxtyp(kproma,nclass)
    REAL(dp)           :: fluxbin(kproma,ntrace)
    REAL(dp)           :: flux_ai(kproma),    &    ! dust flux sum for accumulation mode
                      flux_ci(kproma)          ! dust flux sum for coarse mode
    REAL(dp)           :: fdp1, fdp2
    
    REAL(dp)           :: fluxdiam1(kproma,nclass)  ! flux for soil type #1
    REAL(dp)           :: fluxdiam2(kproma,nclass)  ! flux for soil type #2
    REAL(dp)           :: fluxdiam3(kproma,nclass)  ! flux for soil type #3
    REAL(dp)           :: fluxdiam4(kproma,nclass)  ! flux for soil type #4
    REAL(dp)           :: fluxdiam6(kproma,nclass)  ! flux for soil type #6
    REAL(dp)           :: fluxdiam_pf(kproma,nclass)! flux for preferential sources soil type
    
    INTEGER        :: i                          ! loop index
    INTEGER        :: i_soil                     ! soil type
    
    INTEGER        :: kk,kkk                     ! loop index
    INTEGER        :: kkmin
        
    INTEGER        :: n,nn                       ! loop index
    
    REAL(dp)           :: Ustar,Ustar_d(kproma)    ! threshold friction velocity for saltation
    REAL(dp)           :: uthp
    
    REAL(dp)           :: zw1r(kproma)             ! ECHAM relative soil moisture
    REAL(dp)           :: zcvs(kproma)             ! ECHAM snow cover fraction
    REAL(dp)           :: zcvsc(kproma)            ! ECHAM snow cover fraction at the canopy
    REAL(dp)           :: zwlmx(kproma)            ! auxiliary variable for ECHAM snow cover calculation
    REAL(dp)           :: zcvw(kproma)             ! auxiliary variable for ECHAM snow cover calculation
    REAL(dp)           :: zsn_mm, zsigh              ! auxiliary variable for ECHAM snow cover calculation


    IF (.NOT. ALLOCATED(flux_6h)) ALLOCATE (flux_6h(lc%nproma, ntrace, lc%ngpblks))

    du_wind(:) = velo10m(1:kproma,krow)

    WHERE (du_wind(:) .LT. 0._dp) du_wind(:) = 0._dp         ! set missing values to zero

    ! calculate ECHAM snow cover fraction *zcvs* (algorithm taken from routine *physc.f90*)
    ! and ECHAM relative soil moisture *zw1r* (ratio *wl* to *wlmax*)
    DO i=1,kproma
     
      IF ((slm(i,krow).GT.0.5_dp).AND.(glac(i,krow).LT.0.5_dp)) THEN    ! land surface, but not a glacier 
        zw1r(i)= MIN(ws(i,krow)/wsmx(i,krow),1._dp)
        zwlmx(i)=cwlmax*(1._dp+vlt(i,krow))
        zcvw(i)=MIN(wl(i,krow)/zwlmx(i),1._dp)
        zsn_mm=1000._dp*sn(i,krow)
        zsigh=SQRT(zsn_mm/(zsn_mm+zepsec+zsigfac*orostd(i,krow)))
        zcvs(i)=cqsncr*TANH(zsn_mm/10._dp)*zsigh
        zcvsc(i)=MIN(1._dp,snc(i,krow)/(zwlmx(i)-cwlmax+EPSILON(1._dp)))
        IF (zcvs(i).LT.EPSILON(1._dp) .AND. zcvsc(i).GE.EPSILON(1._dp)) THEN
           zcvs(i)=zcvsc(i)
        END IF
      ELSEIF((slm(i,krow).GT.0.5_dp).AND.(glac(i,krow).GT.0.5_dp)) THEN ! glacier on land surface
        zw1r(i)= 1._dp
        zcvs(i)=1._dp
      ELSE                                                        ! ocean grid point
        zw1r(i)= -9.e9_dp
        zcvs(i)=-9.e9_dp      
      END IF
      
    END DO !i

    du_W1r(:) = zw1r(:)
    du_snow(:) = zcvs(:)

    WHERE (du_W1r(:)  .LT. 0._dp) du_W1r(:)  = 0._dp ! set missing values to zero
    WHERE (du_snow(:) .LT. 0._dp) du_snow(:) = 0._dp

    IF (c_eff.GT.0._dp) THEN

     rdp=dmin  
     dpk(:)=0._dp
     dlast=dmin
     nn=1     
     dbmin(:)=0._dp
     dbmax(:)=0._dp      
     DO kk=1,nclass                              ! assign fluxes to bins
      IF (mod(kk,nbin).eq.0) THEN
       dbmax(nn)=rdp*10000._dp*0.5_dp                   ! calculate bin minimum/maximum radius in um
       dbmin(nn)=dlast*10000._dp*0.5_dp     
       dpk(nn)=sqrt(dbmax(nn)*dbmin(nn))
       nn=nn+1
       dlast=rdp
      ENDIF
      rdp = rdp * exp(Dstep)
     ENDDO !kk      

     fluxbin(:,:)=0._dp
     flux_6h(:,:,krow)=0._dp
        
     dust_mask(:)=0
     
     DO i = 1,kproma
       Ustar = (VK * du_wind(i) *100._dp)/(log(ZZ/Z02))  ! set wind speed
       IF (Ustar.gt.0.and.(Ustar.GE. umin*cuscale/c_eff)) THEN  ! check critical wind speed (wind stress threshold scaled)
         IF (k_fpar_eff(i,krow).GT.1.E-10_dp) THEN   ! check if the grid cell is a potential dust source
           dust_mask(i)=1                       ! set grid point as a potential dust source point
           Ustar_d(i)=Ustar                     ! store wind speed of dust grid
         ENDIF ! k_fpar_eff.gt.0.
         ustar_acrit(i,krow)=ustar_acrit(i,krow)+delta_time
       ENDIF   ! Ustar
     END DO !i
     
     fluxtyp(:,:)=0._dp
     fluxdiam1(:,:)=0._dp
     fluxdiam2(:,:)=0._dp
     fluxdiam3(:,:)=0._dp
     fluxdiam4(:,:)=0._dp
     fluxdiam6(:,:)=0._dp
     fluxdiam_pf(:,:)=0._dp
     dn(:)=0
         
     DO kk=1,nclass
      Uthp=Uth(kk)
      DO i = 1,kproma
        IF (dust_mask(i).EQ.1) THEN
        ! Marticorena:               
        fdp1 = (1._dp-(Uthp/(c_eff * UStar_d(i))))
        fdp2 = (1._dp+(Uthp/(c_eff * UStar_d(i))))**2
        ! Shao:
        ! fdp1 = (1.-(Uthp/(c_eff * UStar_d(i)))**2)
        ! fdp2 = 1.
         IF (fdp1.gt.0._dp) THEN
          i_soil=1
          alpha= solspe(i_soil,nmode*3+1)
          fluxdiam1(i,kk) = srel(i_soil,kk) * fdp1 * fdp2 * cd * UStar_d(i)**3 *alpha ! flux for soil type #1
          i_soil=2
          alpha= solspe(i_soil,nmode*3+1)
          fluxdiam2(i,kk) = srel(i_soil,kk) * fdp1 * fdp2 * cd * UStar_d(i)**3 *alpha ! flux for soil type #2
          i_soil=3
          alpha= solspe(i_soil,nmode*3+1)
          fluxdiam3(i,kk) = srel(i_soil,kk) * fdp1 * fdp2 * cd * UStar_d(i)**3 *alpha ! flux for soil type #3
          i_soil=4
          alpha= solspe(i_soil,nmode*3+1)
          fluxdiam4(i,kk) = srel(i_soil,kk) * fdp1 * fdp2 * cd * UStar_d(i)**3 *alpha ! flux for soil type #4
          i_soil=6
          alpha= solspe(i_soil,nmode*3+1)
          fluxdiam6(i,kk) = srel(i_soil,kk) * fdp1 * fdp2 * cd * UStar_d(i)**3 *alpha ! flux for soil type #6
          i_soil=10
          alpha= solspe(i_soil,nmode*3+1)
          fluxdiam_pf(i,kk)= srel(i_soil,kk) * fdp1 * fdp2 * cd * UStar_d(i)**3 *alpha ! fluxdiam_pf = potential flux if soil type would be a preferential source
          IF (kk.eq.1) THEN 
           fluxtyp(i,kk) = fluxtyp(i,kk)                                                                      &
             + fluxdiam1(i,kk)*(1._dp-mat_psrc(i,krow))*(mat_s1(i,krow)-mat_s2(i,krow)-mat_s3(i,krow)         &
                                                                             -mat_s4(i,krow)-mat_s6(i,krow))  &
             + fluxdiam2(i,kk)*(1._dp-mat_psrc(i,krow))*mat_s2(i,krow)                                        &
             + fluxdiam3(i,kk)*(1._dp-mat_psrc(i,krow))*mat_s3(i,krow)                                        &
             + fluxdiam4(i,kk)*(1._dp-mat_psrc(i,krow))*mat_s4(i,krow)                                        &
             + fluxdiam6(i,kk)*(1._dp-mat_psrc(i,krow))*mat_s6(i,krow)                                        &
             + fluxdiam_pf(i,kk)*mat_psrc(i,krow)
          ELSE
           dn(i)=dn(i)+1           ! increase number of dislocated particle classes
           dk(i,dn(i))=kk          ! store dislocated particle class
          ENDIF ! kk.eq.1
         ENDIF ! fdp1.gt.0.
        ENDIF ! dust_mask.eq.1
      ENDDO ! i
     ENDDO ! kk
                                              
     kkmin=1
     DO i = 1,kproma
       IF (dust_mask(i).EQ.1) THEN
        DO n=1,dn(i)                 ! loop over dislocated dust particle classes
         DO kkk=1,dk(i,n)            ! scaling with relative contribution of dust size  fraction
          i_soil=1
          fluxtyp(i,kkk) = fluxtyp(i,kkk) + &
        (1._dp-mat_psrc(i,krow))*(mat_s1(i,krow)-mat_s2(i,krow)-mat_s3(i,krow)-mat_s4(i,krow)-mat_s6(i,krow))  &
            * fluxdiam1(i,dk(i,n))*srelV(i_soil,kkk)/((su_srelV(i_soil,dk(i,n))-su_srelV(i_soil,kkmin)))
          i_soil=2
          fluxtyp(i,kkk) = fluxtyp(i,kkk) + (1._dp-mat_psrc(i,krow))*mat_s2(i,krow)           &
            * fluxdiam2(i,dk(i,n))*srelV(i_soil,kkk)/((su_srelV(i_soil,dk(i,n))-su_srelV(i_soil,kkmin)))
          i_soil=3
          fluxtyp(i,kkk) = fluxtyp(i,kkk) + (1._dp-mat_psrc(i,krow))*mat_s3(i,krow)            &
            * fluxdiam3(i,dk(i,n))*srelV(i_soil,kkk)/((su_srelV(i_soil,dk(i,n))-su_srelV(i_soil,kkmin)))
          i_soil=4
          fluxtyp(i,kkk) = fluxtyp(i,kkk) + (1._dp-mat_psrc(i,krow))*mat_s4(i,krow)             &
            * fluxdiam4(i,dk(i,n))*srelV(i_soil,kkk)/((su_srelV(i_soil,dk(i,n))-su_srelV(i_soil,kkmin)))
          i_soil=6
          fluxtyp(i,kkk) = fluxtyp(i,kkk) + (1._dp-mat_psrc(i,krow))*mat_s6(i,krow)              &
            * fluxdiam6(i,dk(i,n))*srelV(i_soil,kkk)/((su_srelV(i_soil,dk(i,n))-su_srelV(i_soil,kkmin)))
          IF(du_wind(i).gt.10._dp) THEN
           i_soil=11 ! flux from preferential source at high wind speeds
          ELSE
           i_soil=10 ! flux from preferential source at low wind speeds
          ENDIF
          fluxtyp(i,kkk) = fluxtyp(i,kkk) + mat_psrc(i,krow)                          &
            * fluxdiam_pf(i,dk(i,n))*srelV(i_soil,kkk)/((su_srelV(i_soil,dk(i,n))-su_srelV(i_soil,kkmin)))
         ENDDO ! kkk
        ENDDO ! n
       ENDIF ! dust_mask.eq.1
     ENDDO ! i
       
     DO nn=1,ntrace
      DO i = 1,kproma
        IF (dust_mask(i).EQ.1) THEN
         fluxbin(i,nn) = fluxbin(i,nn)+SUM(fluxtyp(i,(nn-1)*nbin+1:nn*nbin))
         IF (du_W1r(i).gt.w0) fluxbin(i,nn)=0._dp        ! mask out dust fluxes, where soil moisture threshold is reached
         flux_6h(i,nn,krow)=fluxbin(i,nn)*10000._dp* &        ! fluxbin: g/cm2/sec; flux_6h: g/m2/sec
                            (1._dp-du_snow(i))*k_fpar_eff(i,krow)
        ENDIF ! dust_mask.eq.1
      ENDDO ! i
     ENDDO ! nn

     DO nn=1,ntrace
      DO i=1,kproma
       flux_ann(i,krow)=flux_ann(i,krow)+flux_6h(i,nn,krow)*delta_time                       ! flux_ann: g/m2/year
       IF (dpk(nn).le.10._dp) flux_a10(i,krow)=flux_a10(i,krow)+flux_6h(i,nn,krow)*delta_time   ! flux_a10: g/m2/year
       IF (dpk(nn).le. 2._dp) flux_a2(i,krow) = flux_a2(i,krow)+flux_6h(i,nn,krow)*delta_time   ! flux_a2:  g/m2/year
       IF (dpk(nn).le. 1._dp) flux_a1(i,krow) = flux_a1(i,krow)+flux_6h(i,nn,krow)*delta_time   ! flux_a1:  g/m2/year
      ENDDO ! i                           
     ENDDO ! nn

    ENDIF   !c_eff            

    !--- Sum over internal tracer classes for HAM modal scheme:

    flux_ai(:) = 0._dp
    flux_ci(:) = 0._dp

    ! Accumulation mode:

    DO nn=min_ai, max_ai
       flux_ai(:) = flux_ai(:) + flux_6h(:kproma,nn,krow)
    END DO

    ! Coarse mode:

    DO nn=min_ci, max_ci
     flux_ci(:)=flux_ci(:)+flux_6h(:kproma,nn,krow)
    ENDDO !nn

    WHERE (flux_ai.le.0) flux_ai = 0._dp  ! clean up fluxes
    WHERE (flux_ci.le.0) flux_ci = 0._dp  ! clean up fluxes

! not used
!    duststat(:,:,2)=duststat(:,:,2)+(flux_ai(:)+flux_ci(:))*delta_time    ! for diagnostic output
          
    WHERE (glac(1:kproma,krow) > 0.5_dp .OR. slm(1:kproma,krow) < 0.5_dp)      ! mask out glacier and ocean values on ECHAM grid
       flux_ai(:) = 0._dp     
       flux_ci(:) = 0._dp 
    ENDWHERE

! not used
!    duststat2(1:nlon,1:ngl,2)=duststat2(1:nlon,1:ngl,2)+(flux_ai(:)+flux_ci(:))*delta_time

    emi_du_mai(:kproma,krow) = flux_ai(:kproma)
    emi_du_mci(:kproma,krow) = flux_ci(:kproma)

  END SUBROUTINE bgc_dust_calc_emis


  SUBROUTINE dust_emissions_bgc(pxtems, kproma, kbdim, krow)

    ! Description:
    ! ------------
    ! *dust_emissions_bgc* is the interface between the MPI BGC dust emission scheme
    ! and the M7 aerosol module. Dust number flux is calculated from the mass flux and 
    ! both fluxes are stored as ECHAM5 tracers.
    !
    ! Author:
    ! -------
    ! M. Werner, MPI BGC, November 2002
    !
    ! Interface:
    ! ----------
    ! *dust_emissions_bgc* is called from *call_chem_bcond* in file *call_submodels.f90*.
    !
    ! This routine is based on *dust_emissions* by Philip Stier, MPI-Met, 
    ! and *dust_emissions_lsce* by Michael Schulz, LSCE.

    USE mo_tracer,        ONLY: ntrac, flag
    USE mo_constants,     ONLY: api
    USE mo_exception,     ONLY: finish
    USE mo_aero_trac,     ONLY: idt_mduai,   idt_mduci,   &
                                idt_nai,     idt_nci
    USE mo_aero_m7,       ONLY: iacci, icoai, sigma
    USE mo_aero_mem,      ONLY: d_emi_mdu,                &
                                d_emi_mduai, d_emi_mduci, &
                                d_emi_nduai, d_emi_nduci, &
                                emi_du_mai,  emi_du_mci
    USE mo_time_control,  ONLY: delta_time
    USE mo_bgc_dust_mem,  ONLY: mmr_ai,      mmr_ci,      &
                                sig_ai,      sig_ci

    IMPLICIT NONE

    !--- 0) Initialisation:

    !--- Parameters:
    
    INTEGER, INTENT(in)     :: kproma, kbdim, krow
    REAL(dp),    INTENT(inout)  :: pxtems(kbdim,ntrac)
    
    !--- Local Variables:

    REAL(dp)          :: fcon_ai, fcon_ci ! Mass to number conversion factor for number flux
    REAL(dp)          :: densdust         ! Particle density in kg/m3


    !--- Consistency check:

    IF( ABS(sig_ai-sigma(iacci))>1.E-3_dp .OR. ABS(sig_ci-sigma(icoai))>1.E-3_dp) THEN
       CALL finish('bgc_dust_emis','inconsistent source standard deviation')
    END IF


    !--- Get density of dust from the tr_list:

    densdust=flag('density',idt_mduci)


    !--- 1) Calculate BGC dust emissions:

    CALL bgc_dust_calc_emis(kproma, krow)

    
    !--- 2) Apply BGC dust emissions:

    !--- Computation of the mass flux [kg m-2 s-1]: 

    pxtems(1:kproma,idt_mduai)=MAX(emi_du_mai(1:kproma,krow), 0._dp)/1000._dp
    pxtems(1:kproma,idt_mduci)=MAX(emi_du_mci(1:kproma,krow), 0._dp)/1000._dp

    !--- Computation of the emission number flux [m-2 s-1]:

    !--- Calculate conversion factor from mass flux to number flux for 
    !    log-normal distributions:

    fcon_ai=3._dp/(4._dp*api*densdust*mmr_ai**3) * EXP(4.5_dp*LOG(sig_ai)**2)
    fcon_ci=3._dp/(4._dp*api*densdust*mmr_ci**3) * EXP(4.5_dp*LOG(sig_ci)**2)

    !--- Convert mass flux into number fluxes:

    pxtems(1:kproma,idt_nai) = pxtems(1:kproma,idt_mduai)*fcon_ai
    pxtems(1:kproma,idt_nci) = pxtems(1:kproma,idt_mduci)*fcon_ci

    !--- 5) Store the emission fluxes for diagnostics:

    d_emi_mdu(1:kproma,krow)   = d_emi_mdu(1:kproma,krow)  + &
                                 (pxtems(1:kproma,idt_mduai)+pxtems(1:kproma,idt_mduci))*delta_time

    d_emi_mduai(1:kproma,krow) = d_emi_mduai(1:kproma,krow)+pxtems(1:kproma,idt_mduai)        *delta_time
    d_emi_nduai(1:kproma,krow) = d_emi_nduai(1:kproma,krow)+pxtems(1:kproma,idt_mduai)*fcon_ai*delta_time

    d_emi_mduci(1:kproma,krow) = d_emi_mduci(1:kproma,krow)+pxtems(1:kproma,idt_mduci)        *delta_time
    d_emi_nduci(1:kproma,krow) = d_emi_nduci(1:kproma,krow)+pxtems(1:kproma,idt_mduci)*fcon_ci*delta_time

  END SUBROUTINE dust_emissions_bgc


  SUBROUTINE bgc_prestat

    ! Description:
    ! ------------
    ! *bgc_prestat* reads total global dust emission numbers from the text file "dust.diagnostics.dat"
    ! The purpose is to correctly continue monitoring total dust emissions after a (re)start of the model
    !
    ! Authors:
    ! --------
    ! M. Werner, MPI BGC, January 2003
    !
    ! Interface:
    ! ----------
    ! *bgc_prestat* is called from *call_submodels.f90*.
    !
    ! Method:
    ! -------
    ! Reads diagnostic dust sums from text file "dust.diagnostics.dat".
    ! If the text file does not exist, the dust sums are reset to zero (=initial start).

    USE mo_mpi,               ONLY: p_parallel_io, p_io, p_bcast
    USE mo_filename,          ONLY: find_next_free_unit
    USE mo_bgc_dust_mem,      ONLY: dustsum

    IMPLICIT NONE

    INTEGER                           :: iunit   ! fortran unit for input file
    LOGICAL                           :: lex     ! flag to test if file exists
    INTEGER                           :: kyear, kmonth, kday, khour, kminute, ksecond

    INQUIRE(file='./dust.diagnostics.dat', exist=lex)  ! read dust diagnostic file
    IF (lex) THEN
     IF (p_parallel_io) THEN
      iunit = find_next_free_unit(10,99)
      OPEN(iunit,file='./dust.diagnostics.dat',status='old')      
20    READ(iunit,*,END=25) kyear, kmonth, kday, khour, kminute, ksecond, dustsum(1), dustsum(2), dustsum(3), dustsum(4) 
      GOTO 20
25    CLOSE(iunit)
      dustsum(:) = dustsum(:)*1.e12_dp              ! convert back from Mt/yr to g/yr
     ENDIF
    ELSE
     dustsum(:) = 0._dp                             ! if diagnostic file does not exist: initialize dust diagnostics
    ENDIF     

    !--- Broadcast field to all CPUs:

    CALL p_bcast(dustsum, p_io) 

  END SUBROUTINE bgc_prestat

  
  SUBROUTINE bgc_trastat(kyear, kmonth, kday, khour, kminute, ksecond)

    ! Description:
    ! ------------
    ! *bgc_trastat* writes total global dust emission numbers to the text file "dust.diagnostics.dat"
    ! The purpose is to correctly continue monitoring total dust emissions after a (re)start of the model
    !
    ! Authors:
    ! --------
    ! M. Werner, MPI BGC, January 2003
    ! M. Werner, MPI BGC, October 2004: changed emission cheme from 0.5x0.5 degree grid to 
    !                                    default Echam5-HAM grid (to save computational costs)
    !
    ! Interface:
    ! ----------
    ! *bgc_trastat* is called from *call_submodels.f90*.
    !
    ! Method:
    ! -------
    ! Writes diagnostic dust sums to text file "dust.diagnostics.dat".
    ! If the text file does not exist, it is newly created (=cold start), 
    ! otherwise the output is appended (=restart).

    USE mo_constants,         ONLY: api
    USE mo_control,           ONLY: nlon, ngl
    USE mo_doctor,            ONLY: nout
    USE mo_mpi,               ONLY: p_parallel_io
    USE mo_filename,          ONLY: find_next_free_unit
    USE mo_bgc_dust_mem,      ONLY: idust,k_fpar_eff,                 &
                                    dustsum, flux_ann,                &
                                    flux_a1,flux_a2,flux_a10
    USE mo_decomposition,     ONLY: gl_dc=> global_decomposition
    USE mo_transpose,         ONLY: gather_gp
                                    

    IMPLICIT NONE

    REAL(dp), POINTER :: zflux_ann(:,:)
    REAL(dp), POINTER :: zflux_a1(:,:)
    REAL(dp), POINTER :: zflux_a2(:,:)
    REAL(dp), POINTER :: zflux_a10(:,:)
    REAL(dp), POINTER :: zk_fpar_eff(:,:)

    INTEGER                           :: iunit   ! fortran unit for input file
    INTEGER, INTENT(IN)               :: kyear, kmonth, kday, khour, kminute, ksecond

    REAL(dp)           :: dustsumt, dustsum10, dustsum2, dustsum1  ! dummy values for diagnostics

    REAL(dp)           :: latitude    
    REAL(dp)           :: S
    REAL(dp)           :: surf(ngl)                  ! grid surface area
    REAL(dp)           :: tool
        
    INTEGER        :: x, y                       ! longitude / latitude index

    REAL(dp)             :: resol
    REAL(dp), PARAMETER  :: radius = 6371000._dp        ! radius of Earth


    resol = 180._dp/ngl

    dustsumt = dustsum(1)                        ! initialize dustsums with values from last (re)start
    dustsum10 = dustsum(2)                       ! initialize dustsums with values from last (re)start
    dustsum2 = dustsum(3)
    dustsum1 = dustsum(4)

    DO y = 1,ngl                                 ! calculate grid surface area for different latitudes
      latitude =(90._dp + resol * (0.5_dp - REAL(y,dp)) )
      S = radius * radius * api / (180._dp/resol)
      tool = COS(latitude * api/180._dp)
      surf(y) = S * abs(tool) * 2._dp*api/(360._dp/resol)
    ENDDO

    IF (p_pe == p_io) THEN
       ALLOCATE (zflux_ann(nlon,ngl))
       ALLOCATE (zflux_a10(nlon,ngl))
       ALLOCATE (zflux_a2(nlon,ngl))
       ALLOCATE (zflux_a1(nlon,ngl))
       ALLOCATE (zk_fpar_eff(nlon,ngl))
    END IF

    CALL gather_gp (zflux_ann, flux_ann, gl_dc)
    CALL gather_gp (zflux_a10, flux_a10, gl_dc)
    CALL gather_gp (zflux_a2, flux_a2, gl_dc)
    CALL gather_gp (zflux_a1, flux_a1, gl_dc)
    CALL gather_gp (zk_fpar_eff, k_fpar_eff, gl_dc)

    IF (p_parallel_io) THEN                      ! write out diagnostics to standard output   
      DO y = 1,ngl
        DO x = 1,nlon
          IF (zk_fpar_eff(x,y).GT.1.E-10_dp) THEN        ! if the area is a potential dust source: sum up for diagnostic output
            dustsumt=dustsumt+zflux_ann(x,y)*surf(y)   ! total grams of dust per year
            dustsum10=dustsum10+zflux_a10(x,y)*surf(y) ! small dust particles < 10um
            dustsum2=dustsum2+zflux_a2(x,y)*surf(y)    ! small dust particles < 2um
            dustsum1=dustsum1+zflux_a1(x,y)*surf(y)    ! small dust particles < 1um
          ENDIF !! (k_fpar_eff.gt.1.e-10)
        END DO !! x
      END DO !! y

      DEALLOCATE (zflux_ann)
      DEALLOCATE (zflux_a10)
      DEALLOCATE (zflux_a2)
      DEALLOCATE (zflux_a1)
      DEALLOCATE (zk_fpar_eff)

     WRITE(nout,'(4(a,e10.4))')                                 &
                   ' DUST (Mt/yr): total: ',dustsumt*1.e-12_dp,    &
                   '     < 10um: ',dustsum10*1.e-12_dp,            &
                   '     < 2um: ',dustsum2*1.e-12_dp,              &
                   '     < 1um: ',dustsum1*1.e-12_dp
     iunit = find_next_free_unit(55,99)
     OPEN(iunit,file='dust.diagnostics.dat',status='unknown',position='append')
     WRITE(iunit,*) kyear, kmonth, kday, khour, kminute, ksecond, &
                    dustsumt*1.e-12_dp, dustsum10*1.e-12_dp, dustsum2*1.e-12_dp, dustsum1*1.e-12_dp
     CLOSE(iunit)

    ENDIF

  END SUBROUTINE bgc_trastat

  
  SUBROUTINE bgc_dust_cleanup

    ! Description:
    ! ------------
    ! *bgc_dust_cleanup* deallocates memory needed for the dust scheme.
    !
    ! Authors:
    ! --------
    ! M. Werner, MPI BGC, November 2004
    !
    ! Interface:
    ! ----------
    ! *bgc_dust_cleanup* is called from *call_free_submodel_memory*
    
    USE mo_bgc_dust_mem,      ONLY: biome,                    &
                                    mat_s1, mat_s2, mat_s3,   &
                                    mat_s4, mat_s6, mat_psrc, &
                                    k_fpar_eff,               &
!                                    duststat, duststat2,      &
                                    flux_ann, flux_a1,        &
                                    flux_a2, flux_a10,        &
                                    idust, flux_6h
    IMPLICIT NONE

    IF (ALLOCATED(biome))      DEALLOCATE (biome)
    IF (ALLOCATED(mat_s1))     DEALLOCATE (mat_s1)
    IF (ALLOCATED(mat_s2))     DEALLOCATE (mat_s2)
    IF (ALLOCATED(mat_s3))     DEALLOCATE (mat_s3)
    IF (ALLOCATED(mat_s4))     DEALLOCATE (mat_s4)
    IF (ALLOCATED(mat_s6))     DEALLOCATE (mat_s6)
    IF (ALLOCATED(mat_psrc))   DEALLOCATE (mat_psrc)
    IF (ALLOCATED(k_fpar_eff)) DEALLOCATE (k_fpar_eff)
!    IF (ALLOCATED(duststat))   DEALLOCATE (duststat)
!    IF (ALLOCATED(duststat2))  DEALLOCATE (duststat2)
    IF (ALLOCATED(flux_ann))   DEALLOCATE (flux_ann)
    IF (ALLOCATED(flux_a1))    DEALLOCATE (flux_a1)
    IF (ALLOCATED(flux_a2))    DEALLOCATE (flux_a2)
    IF (ALLOCATED(flux_a10))   DEALLOCATE (flux_a10)
    IF (ALLOCATED(idust))      DEALLOCATE (idust)
    IF (ALLOCATED(flux_6h))    DEALLOCATE (flux_6h)

  END SUBROUTINE bgc_dust_cleanup


END MODULE mo_bgc_dust_emis
