 MODULE mo_chem_bcond
  !
  ! *mo_chem_bcond* handles all emissions for chemical tracers.
  !
  ! Authors:
  ! --------
  ! Philip Stier, MPI-MET                               2001-2005
  !
  ! Method:
  ! -------
  ! mo_chem_bcond holds the emission routines for the various
  ! tracer emissions. These are called by the subroutine
  ! chem_bcond ("chemical boundary conditions") which is called 
  ! from vdiff. 
  ! The routines xt_read... read necessary input files and are 
  ! called from read_chem_files in which is called from call_submodels. 
  ! 
  ! The emission routines themselfes are from various authors
  ! and are described in their header. 

  USE mo_kind,             ONLY: dp
  

IMPLICIT NONE

CONTAINS

  SUBROUTINE xt_volc_emiss_ini

    ! *xt_volc_emiss_ini* reads and prepares the volcanic emissions
    !
    ! Method:
    ! -------
    ! The input data is read from a file following the conventions of 
    ! Spiro et al. (1992) provided by the run script at the unit nemi.
    ! The emissions are first geo-located to the next neighbor model
    ! gridpoint, independent of the respective horizontal model
    ! resolution. The vertical emission level is estimated hard-coded
    ! for the L19 vertical resolution.
    ! Global emissions are normalized to 1 Tg of S to allow for easy
    ! scaling in the respective emission routine. 
    ! The normalized emission fluxes [kg(S) m-2 s-1]  and the emission
    ! model-levels are stored in the streams emi_so2_vol and 
    ! emi_so2_vol_lev, respectively. 
    !
    ! Interface:
    ! ----------
    ! *xt__volc_emiss_ini* is called from *read_chem_files*
    !
    ! Authors:
    ! --------
    ! Philip Stier, MPI-MET                      2002
    !
    ! Hans Feichter, MPI-MET 
    ! (original offline interpolation routine)   

    USE mo_exception,     ONLY: finish
    USE mo_control,       ONLY: nlon, ngl
    USE mo_doctor,        ONLY: nout
    USE mo_filename,      ONLY: find_next_free_unit
    USE mo_aero_mem,      ONLY: emi_so2_vol, emi_so2_vol_alt
    USE mo_mpi,           ONLY: p_pe, p_io
    USE mo_transpose,     ONLY: scatter_gp
    USE mo_decomposition, ONLY: gl_dc => global_decomposition
    USE mo_gaussgrid,     ONLY: gridarea
    USE mo_aero_tools,    ONLY: geoindex

    !--- Local variables: 
    
    LOGICAL          :: lex        ! .true. when input file exists
    INTEGER          :: klon,     klat,     ierr,    inumtot, &
                        inum,     ilon,     ilat,    istart,  &
                        idurat,   jvol,     iheight, iunit
    REAL(dp)             :: zxemis,   zsumemi,  znorm
    REAL(dp)             :: lon(72),  lat(46)
    REAL(dp),    POINTER :: ztmp1(:,:), ztmp2(:,:)

    INTEGER, POINTER :: zheights(:), ilats(:), ilons(:)
    REAL(dp),    POINTER :: zemiss(:)

    CHARACTER(len=256) :: cfile ='emiss_fields_volc.txt'

    !--- 0) Initializations:
    
    !--- Grid as defined by the Spiro et al. dataset:

    DATA LON/-180._dp,-175._dp,-170._dp,-165._dp,-160._dp,-155._dp,-150._dp,-145._dp,-140._dp,   &
             -135._dp,-130._dp,-125._dp,-120._dp,-115._dp,-110._dp,-105._dp,-100._dp, -95._dp,   &
              -90._dp, -85._dp, -80._dp, -75._dp, -70._dp, -65._dp, -60._dp, -55._dp, -50._dp,   &
              -45._dp, -40._dp, -35._dp, -30._dp, -25._dp, -20._dp, -15._dp, -10._dp,  -5._dp,   &
                0._dp,   5._dp,  10._dp,  15._dp,  20._dp,  25._dp,  30._dp,  35._dp,  40._dp,   &
               45._dp,  50._dp,  55._dp,  60._dp,  65._dp,  70._dp,  75._dp,  80._dp,  85._dp,   &
               90._dp,  95._dp, 100._dp, 105._dp, 110._dp, 115._dp, 120._dp, 125._dp, 130._dp,   &
              135._dp, 140._dp, 145._dp, 150._dp, 155._dp, 160._dp, 165._dp, 170._dp, 175._dp/

     DATA LAT/-89._dp,-86._dp,-82._dp,-78._dp,-74._dp,-70._dp,-66._dp,-62._dp,-58._dp,-54._dp,   &
              -50._dp,-46._dp,-42._dp,-38._dp,-34._dp,-30._dp,-26._dp,-22._dp,-18._dp,-14._dp,   &
              -10._dp, -6._dp, -2._dp,  2._dp,  6._dp, 10._dp, 14._dp, 18._dp, 22._dp, 26._dp,   &
               30._dp, 34._dp, 38._dp, 42._dp, 46._dp, 50._dp, 54._dp, 58._dp, 62._dp, 66._dp,   &
               70._dp, 74._dp, 78._dp, 82._dp, 86._dp,89._dp/


    !--- 1) Read fields:

    !--- 1.1) Inquire: 

    IF (p_pe == p_io) THEN

       INQUIRE (file=cfile,exist=lex)

       IF (lex) THEN
          iunit = find_next_free_unit (20, 99)
          ALLOCATE (ztmp1(nlon,ngl)) 
          ALLOCATE (ztmp2(nlon,ngl)) 
          OPEN (iunit,file=cfile,form='FORMATTED')
          REWIND(iunit)
          WRITE (nout,*) '------------------------------------------------------------------------------'
          WRITE (nout,*) '--- Initialization of volcanic SO2 emissions ---------------------------------'
          WRITE (nout,*) 
          WRITE (nout,*) '   xt_volc_emiss_ini: Reading volcanic SO2 emissions from unit ', iunit
       ELSE
          WRITE (nout,*) ' Trying to read from unit ', iunit
          WRITE (nout,*) ' File for volcanic SO2 emissions does not exist.'
          CALL finish('xt_read_volc_emis','Error reading input file '//TRIM(cfile))
       END IF

       inumtot=0
       inum=0
       zsumemi=0._dp

       !--- 1.2) Determine number of fields:

       DO
          READ(iunit,*,IOSTAT=ierr)

          IF(ierr < 0 ) THEN
             REWIND(iunit)
             EXIT
          ELSE IF(ierr > 0) THEN
             CALL finish('xt_read_volc_emis','Error reading input file (2)')
          END IF

          inumtot=inumtot+1

       END DO

       !--- 1.3) Allocate fields and initialize:

       ALLOCATE(zheights(inumtot))
       ALLOCATE(zemiss(inumtot))
       ALLOCATE(ilats(inumtot))
       ALLOCATE(ilons(inumtot))

       zheights=0
       ilats=0
       ilons=0
       zemiss=0._dp
       ztmp1=0._dp
       ztmp2=0._dp

       !--- 1.4) Read emissions:

       DO
          READ(iunit,fmt='(3x,2i4,e11.3,2x,i5,2x,2i4)',IOSTAT=ierr) ilon, ilat, zxemis, iheight, istart, idurat
             
          IF(ierr < 0 ) THEN
             EXIT
          ELSE IF(ierr > 0) THEN
             CALL finish('xt_read_volc_emis','Error reading input file (3)')
          END IF

          !--- Consider continuously degassing volcanoes only (365 d duration):

          IF(idurat.EQ.365) THEN
                
             ! Store the read data in vectors:

             inum=inum+1
             zheights(inum)=iheight
             zemiss(inum)=zxemis
             IF(LON(ilon)<0)THEN
                ilons(inum)=LON(ilon)+360
             ELSE IF(LON(ilon)>=0)THEN
                ilons(inum)=LON(ilon)
             END IF
             ilats(inum)=LAT(ilat)
                
             ! Sum up the total emissions:

             zsumemi=zsumemi+zxemis

          ENDIF

       END DO
          
       WRITE(nout,*) '   Consider ', inum, ' continuously degasing volcanoes'

       !--- 2) Calculate factor to normalize to 1 Tg S per year:

       !     1Tg [kg]  1 year [s]
       znorm=1.E+09_dp / (86400._dp*365._dp)

       WRITE(nout,*) '   Normalize emissions of ',zsumemi, '[t(S) day-1]'
       WRITE(nout,*) '   to', (86400._dp*365._dp)*znorm*1.E-9_dp, '[Tg S per year]'

       !--- 3) Store on 2D field and normalize:

       DO jvol=1, inum

          !--- Find corresponding lat-lon indices:

          CALL geoindex(REAL(ilons(jvol),dp),REAL(ilats(jvol),dp),klon,klat)

          !--- Calculate the normalized emissions and store in 2D buffer:
             
          ztmp1(klon,klat)=ztmp1(klon,klat)+(zemiss(jvol)/zsumemi)*znorm /gridarea(klat)

          !--- Store corresponding emission height:

          ztmp2(klon,klat)=zheights(jvol)

       END DO

    END IF ! p_pe == p_io

    !--- 4) Scatter to streams over the processors:

    CALL scatter_gp(ztmp1, emi_so2_vol, gl_dc)

    CALL scatter_gp(ztmp2, emi_so2_vol_alt, gl_dc)
       
    !--- 5) Deallocate temporal field and close input file

    IF (p_pe == p_io) THEN

       DEALLOCATE(ztmp1,ztmp2)
       DEALLOCATE(zheights,zemiss,ilats,ilons)

       CLOSE (iunit)

       WRITE (nout,*) '--- End of the initialization of volcanic SO2 emissions ----------------------'
       WRITE (nout,*) '------------------------------------------------------------------------------'

    END IF

  END SUBROUTINE xt_volc_emiss_ini

  SUBROUTINE xt_read_dust_emiss

    ! Description:
    !
    ! *reademi* reads global surface emission fields
    ! to be used in the emission subroutines called from chem_forcing.
    !
    ! Method:
    !
    ! *xt_read_dust_emiss* is called from *read_chem_files*
    !
    ! Authors:
    !
    ! U. Schlese, DKRZ, January 1995, original source
    ! L. Kornblueh, MPI, May 1998, f90 rewrite
    ! U. Schulzweida, MPI, May 1998, f90 rewrite
    ! P. Stier, MPI, January 2002, adopted to dust emissions
    ! 
    ! for more details see file AUTHORS
    !

    USE mo_exception,     ONLY: finish
    USE mo_control,       ONLY: nlon, ngl
    USE mo_doctor,        ONLY: nout
    USE mo_aero_mem,      ONLY: emi_du_thr, emi_du_src, emi_du_cla
    USE mo_mpi,           ONLY: p_parallel_io
    USE mo_transpose,     ONLY: scatter_gp
    USE mo_decomposition, ONLY: gl_dc => global_decomposition
    USE mo_filename,      ONLY: find_next_free_unit
    USE mo_test_trans,    ONLY: test_gridpoint

    !  Local variables: 
    LOGICAL        :: lex        ! .true. when input file exists
    REAL(dp), POINTER  :: ztmp(:,:)
    INTEGER        :: jlon, jlat, iunit


    !  Executable statements 

    IF (p_parallel_io) THEN

       !-- 1) Read fields

       !-- 1.1) Open files and read:

       WRITE (nout,*) '------------------------------------------------------------------------------'
       WRITE (nout,*) '--- Read input data for dust emissions ---------------------------------------'
       WRITE (nout,*)

       !--- Threshold velocity map:

       INQUIRE (file='ThrVelMap.dat',exist=lex)

       IF (lex) THEN

          iunit=find_next_free_unit(55,100)

          WRITE (nout,*) '   Reading THRESHOLD VELOCITY MAP for dust emissions from file ThrVelMap.dat'

          ALLOCATE (ztmp(nlon,ngl)) 

          OPEN(iunit,file='ThrVelMap.dat',status='unknown',form='formatted')
          READ(iunit,'(G18.10)') ((ztmp(jlon,jlat),jlon=1,nlon),jlat=1,ngl)
          CLOSE(iunit)

       ELSE

          CALL finish('xt_read_dust_emiss','Missing input file ThrVelMap.dat')

       END IF

    END IF ! (p_parallel_io)

    CALL scatter_gp(ztmp, emi_du_thr, gl_dc)

    CALL test_gridpoint(emi_du_thr, 'xt_read_dust_emiss: emi_du_thr')

    IF(p_parallel_io) DEALLOCATE (ztmp) 


     !--- Source strength factor:

    IF (p_parallel_io) THEN

       INQUIRE (file='SrcStrFac.dat',exist=lex)

       IF (lex) THEN

          iunit=find_next_free_unit(55,100)

          WRITE (nout,*) '   Reading SOURCE STRENGTH MAP    for dust emissions from file SrcStrFac.dat'

          ALLOCATE (ztmp(nlon,ngl)) 

          OPEN(iunit,file='SrcStrFac.dat',status='unknown',form='formatted')
          READ(iunit,'(G18.10)') ((ztmp(jlon,jlat),jlon=1,nlon),jlat=1,ngl)
          CLOSE(iunit)

       ELSE

          CALL finish('xt_read_dust_emiss','Missing input file SrcStrFac.dat')

       END IF

    END IF ! (p_parallel_io)

    CALL scatter_gp(ztmp, emi_du_src, gl_dc)

    CALL test_gridpoint(emi_du_src, 'xt_read_dust_emiss: emi_du_src')

    IF(p_parallel_io) DEALLOCATE (ztmp) 


    !--- Clay content:

    IF (p_parallel_io) THEN

       INQUIRE (file='ClayCont.dat',exist=lex)

       IF (lex) THEN

          iunit=find_next_free_unit(55,100)

          WRITE (nout,*) '   Reading CLAY CONTENT MAP       for dust emissions from file ClayCont.dat'

          ALLOCATE (ztmp(nlon,ngl)) 
       
          OPEN(iunit,file='ClayCont.dat',status='unknown',form='formatted')
          READ(iunit,'(G18.10)') ((ztmp(jlon,jlat),jlon=1,nlon),jlat=1,ngl)
          CLOSE(iunit)

       ELSE

          CALL finish('xt_read_dust_emiss','Missing input file ClayCont.dat')

       END IF

    END IF ! (p_parallel_io)

    CALL scatter_gp(ztmp, emi_du_cla, gl_dc)

    CALL test_gridpoint(emi_du_cla, 'xt_read_dust_emiss: emi_du_cla')

    IF(p_parallel_io) THEN

       DEALLOCATE (ztmp) 

       WRITE (nout,*) '--- End of reading input data for dust emissions -----------------------------'
       WRITE (nout,*) '------------------------------------------------------------------------------'

    END IF

  END SUBROUTINE xt_read_dust_emiss

  SUBROUTINE xt_read_chem_fields(kmonth)
    
    USE mo_doctor,            ONLY: nout
    USE mo_mpi,               ONLY: p_parallel_io, p_io, p_bcast
    USE mo_decomposition,     ONLY: gc => global_decomposition,   &
                                    dc => local_decomposition

    USE mo_control,           ONLY: ngl, nlon, nlev
    USE mo_transpose,         ONLY: scatter_gp
    USE mo_read_netcdf77,     ONLY: read_var_hs_nf77_3d
    USE mo_tracer,            ONLY: trlist, INITIAL, get_tracer
    USE mo_memory_gl,         ONLY: xt
    USE mo_memory_g1a,        ONLY: xtm1
    USE mo_exception,         ONLY: finish, message

    INTEGER, INTENT(IN)             :: kmonth
    INTEGER, PARAMETER              :: nfield=4
    CHARACTER(len=256)              :: yvar_name(nfield)

    INTEGER                         :: i, ierr, jt

    REAL(dp) ,POINTER               :: field(:,:,:,:)
    CHARACTER(len=15)               :: cevar(nfield)  
    REAL(dp), POINTER               :: zin_3d(:,:,:), zptr(:,:,:) ! for scatter

    IF (p_parallel_io) THEN
       ALLOCATE(zin_3d(nlon,nlev,ngl))
       ALLOCATE (field (nlon,nlev,nfield,ngl))
    ELSE
       NULLIFY(zin_3d)
    ENDIF


    !--- 1) Read annual emission fields:

    yvar_name(1)   = 'OX_VMR_avrg'
    yvar_name(2)   = 'H2O2_VMR_inst'
    yvar_name(3)   = 'OH_VMR_avrg'
    yvar_name(4)   = 'NO2_VMR_avrg'
    cevar(1)        = 'O3'
    cevar(2)        = 'H2O2'
    cevar(3)        = 'OH'
    cevar(4)        = 'NO2'
    IF (p_parallel_io) THEN
       DO i=1, nfield
         CALL read_var_hs_nf77_3d ('chem_fields_monthly.nc', 'lon', 'lev', 'lat', 'time', kmonth,         &
                   TRIM(yvar_name(i)), field(:,:,i,:), ierr)
       END DO
    WRITE (nout,*) '------------------------------------------------------------------------------'
    WRITE (nout,*) '--- Initialization of chem emissions -------------------------------------'
    WRITE (nout,*)
    WRITE (nout,*) '   read chem_fields_monthly.nc: '
    WRITE (nout,*)
    WRITE (nout,*) '------------------------------------------------------------------------------'
    END IF
      DO i = 1, nfield ! loop over species that were read in

          CALL p_bcast(cevar(i), p_io)

          CALL get_tracer(cevar(i),idx=jt,ierr=ierr)

          IF (ierr == 1) THEN
             CALL message('xt_read_chem_fields',&
                          'WARNING: Tracer '//cevar(i)//' not defined')
          ELSE
             IF (p_parallel_io) THEN
                WRITE(nout,'(a,a)') 'reading tracer field for ',cevar(i)
                zin_3d(:,:,:) = field(:,:,i,:)
             ENDIF
             ! scatter the tracer over the processors
             zptr => xt(:,:,jt,:)
             CALL scatter_gp (zin_3d, zptr, gc)

             IF (ASSOCIATED(xtm1)) xtm1(:,:,jt,:) = xt(:,:,jt,:)

             ! set flag that shows that tracer is already initialized
             trlist%ti(jt)%init = INITIAL
          END IF
       END DO
   IF (ASSOCIATED(zin_3d)) DEALLOCATE(zin_3d)
  IF (p_parallel_io) DEALLOCATE (field)
  RETURN

  END SUBROUTINE xt_read_chem_fields

  SUBROUTINE seasalt_emissions_monahan(kproma, kbdim, krow, loland, pseaice, pxtems)
    
    ! Description:
    ! ------------
    ! Calculates the emitted sea salt flux from the 10m wind speed following
    ! Monahan et al., 1986 using the sea salt accumulation and coarse mode
    ! definitions of the GADS dataset at a relative humidity of 80%. 
    !
    ! Authors:
    ! Hans Feichter, MPI
    ! Philip Stier,  MPI
    !
    !--- Paramters:
    !
    !--- Local Variables:
    !
    ! zdr = radius interval in micro meter))
    
 
    USE mo_aero_mem,     ONLY: velo10m,   d_emi_mss 
    USE mo_aero_trac,    ONLY: idt_mssas, idt_msscs, idt_nas, idt_ncs 
    USE mo_tracer,       ONLY: ntrac
    USE mo_doctor,       ONLY: nout
    USE mo_constants,    ONLY: api
    USE mo_time_control, ONLY: delta_time

    IMPLICIT NONE

    !--- Parameters:

    INTEGER, INTENT(in)     :: kproma, kbdim, krow
    LOGICAL, INTENT(in)     :: loland(kbdim)
    REAL(dp),    INTENT(in)     :: pseaice(kbdim)

    REAL(dp), INTENT(inout)     :: pxtems(kbdim,ntrac)

    !--- Local Variables:

    INTEGER                 :: jl

    REAL(dp)                    :: zr1, zr2, zb1, zb2, zx1, zx2, zdr1, zdr2
    REAL(dp)                    :: zzzspeed, zfact

    REAL(dp), SAVE              :: zss1,      zss2 

    LOGICAL, SAVE           :: lofirst=.TRUE.

    !--- Constants:

    REAL(dp), PARAMETER         :: zssmassa= 0.37E-15_dp,   & ! Mass of accu mode ss particle [kg]
                               zssmassc= 0.37E-12_dp      ! Mass of coar mode ss particle [kg]


    !--- 1) Computations of coefficients for the sea-salt emissions: -----------------
    !  
    !     (Sea-salt flux calculation after Monahan et al. 1986 using
    !      ssam and sscm from GADS at 80% r.h. 
    !
    

    IF (lofirst) THEN

       lofirst=.FALSE.

       zr1=0.416_dp
       zr2=3.49_dp
       zb1=0.58_dp-1.54_dp*LOG10(zr1)
       zb2=0.58_dp-1.54_dp*LOG10(zr2)
       zx1=10._dp**(1.19_dp*EXP(-zb1*zb1))
       zx2=10._dp**(1.19_dp*EXP(-zb2*zb2))
       zdr1=0.5_dp
       zdr2=4.5_dp
       zfact=1.373_dp*4._dp/3._dp*api*1.15e3_dp
       zss1=zfact*(1._dp+0.057_dp*zr1**1.05_dp)*zx1*zdr1*1.e-18_dp
       zss2=zfact*(1._dp+0.057_dp*zr2**1.05_dp)*zx2*zdr2*1.e-18_dp

       WRITE(nout,*) "----------------------------------------------------------------"
       WRITE(nout,*) "--- Subroutine seasalt_emissions: "
       WRITE(nout,*) "---"
       WRITE(nout,*) "--- Factors for sea salt fluxes: 1st = ",zss1, "2nd = ", zss2
       WRITE(nout,*) "---"
       WRITE(nout,*) "----------------------------------------------------------------"

    END IF

    !--- 2) Calculate the emitted mass-flux: ------------------------------------------

    DO jl=1,kproma

       zzzspeed=MIN(velo10m(jl,krow),20._dp)

       IF (loland(jl)) THEN
          pxtems(jl,idt_mssas)=0._dp
          pxtems(jl,idt_msscs)=0._dp
       ELSE
          pxtems(jl,idt_mssas)=zss1*zzzspeed**3.41_dp*(1._dp-pseaice(jl))
          pxtems(jl,idt_msscs)=zss2*zzzspeed**3.41_dp*(1._dp-pseaice(jl))
       END IF

       !--- 3) Calculate the emitted number-flux: ----------------------------------------

       ! Accumulation soluble:
       pxtems(jl,idt_nas)= pxtems(jl,idt_nas) + pxtems(jl,idt_mssas)/zssmassa
       ! Coarse soluble:
       pxtems(jl,idt_ncs)= pxtems(jl,idt_ncs) + pxtems(jl,idt_msscs)/zssmassc


       !--- 4) Store the emitted mass for diagnostics: -----------------------------------

       d_emi_mss(jl,krow) = d_emi_mss(jl,krow)+                             &
                            (pxtems(jl,idt_mssas)+pxtems(jl,idt_msscs))*delta_time

    END DO

  END SUBROUTINE seasalt_emissions_monahan


  SUBROUTINE seasalt_emissions_lsce(kproma, kbdim, krow, pseaice, pxtems)
    !     -----------------------------------------------------------------------
    !     
    !   Author:
    !   -------
    !   Michael Schulz 
    !   Laboratoire des Sciences du Climat et de l'Environnement / Saclay
    !   10.1.2002
    !
    !   Modifications:
    !   --------------
    !   Philip Stier, MPI-MET  (Adaption to the ECHAM/HAM structure)       2002
    !   Michael Schulz, LSCE   (Modified source coefficients)        02/08/2002
    !
    !   Purpose:
    !   --------
    !   Describe source flux of sea salt aerosol mass and number flux 
    !   as a function of wind speed
    !   for two aerosol modes: coarse soluble and accumulation soluble
    !
    !   Interface:
    !   ----------
    !      input 
    !       delt           time step duration       [sec]
    !       zspeed         wind speed at 10 m       [m/s]
    !       slf            sea land fraction        [ % 0.0 (sea) - 1. (land)]
    !       alake          lake fraction            [%]
    !
    !      output 
    !       pxtems         emission flux of sea salt mass and number [kg m-2 s-1] / [# m-2 s-1]
    !
    !
    !   Method:
    !   -------
    !     Tabulated mass and number fluxes following Monahan 86 and Smith&Harrison 98
    !     for two aerosol modes; interpolated according to actual wind speed;
    !     sea salt mass and number fluxes are fitted to two lognormal distributions
    !     following work published by Guelle et al. (2001) JGR 106, pp. 27509-27524
    !
    !

    USE mo_aero_mem,     ONLY : velo10m, &  ! 10m windspeed [m s-1]
                                d_emi_mss, &
                                d_emi_mssas, d_emi_msscs, d_emi_nssas, d_emi_nsscs
    USE mo_aero_trac,    ONLY : idt_mssas, idt_msscs, idt_nas,   idt_ncs
    USE mo_memory_g3b,   ONLY : slf, alake
    USE mo_tracer,       ONLY : ntrac
    USE mo_time_control, ONLY : delta_time


    IMPLICIT NONE

    !--- Parameters:

    INTEGER, INTENT(in)    :: kproma, kbdim, krow
    REAL(dp),    INTENT(in)    :: pseaice(kbdim)
    REAL(dp),    INTENT(inout) :: pxtems(kbdim,ntrac)

    !--- Local Variables:

    INTEGER :: wcl(kbdim)                  ! windclass in 1 m/s steps
    REAL(dp)    :: dzspeed(kbdim)
    REAL(dp)    :: zseafrac(kbdim)             ! fraction of the gridcell covered by
                                           ! non-iced sea water [0.-1.]


    !--- Precalculated mass fluxes per wind class [kg m-2 s-1]:

    REAL(dp) , PARAMETER :: mass1flux(0:40) = (/                           &
         0.000E+00_dp, 2.483E-15_dp, 2.591E-14_dp, 1.022E-13_dp, 2.707E-13_dp, 5.761E-13_dp, &
         1.068E-12_dp, 1.800E-12_dp, 2.829E-12_dp, 4.215E-12_dp, 6.023E-12_dp, 8.317E-12_dp, &
         1.117E-11_dp, 1.464E-11_dp, 1.882E-11_dp, 2.378E-11_dp, 2.959E-11_dp, 3.633E-11_dp, &
         4.409E-11_dp, 5.296E-11_dp, 6.301E-11_dp, 7.433E-11_dp, 8.693E-11_dp, 1.012E-10_dp, &
         1.168E-10_dp, 1.342E-10_dp, 1.532E-10_dp, 1.741E-10_dp, 1.970E-10_dp, 2.219E-10_dp, &
         2.489E-10_dp, 2.781E-10_dp, 3.097E-10_dp, 3.437E-10_dp, 3.803E-10_dp, 4.195E-10_dp, &
         4.616E-10_dp, 5.065E-10_dp, 5.544E-10_dp, 6.054E-10_dp, 6.711E-10_dp             /)

    REAL(dp) , PARAMETER :: mass2flux(0:40) = (/                               &
         0.000E+00_dp, 2.319E-13_dp, 2.411E-12_dp, 9.481E-12_dp, 2.505E-11_dp, 5.321E-11_dp, &
         9.850E-11_dp, 1.658E-10_dp, 2.602E-10_dp, 3.874E-10_dp, 5.529E-10_dp, 7.628E-10_dp, &
         1.023E-09_dp, 1.341E-09_dp, 1.722E-09_dp, 2.175E-09_dp, 2.704E-09_dp, 3.319E-09_dp, &
         4.026E-09_dp, 4.832E-09_dp, 5.746E-09_dp, 6.776E-09_dp, 7.925E-09_dp, 9.214E-09_dp, &
         1.064E-08_dp, 1.221E-08_dp, 1.394E-08_dp, 1.584E-08_dp, 1.791E-08_dp, 2.016E-08_dp, &
         2.261E-08_dp, 2.526E-08_dp, 2.812E-08_dp, 3.120E-08_dp, 3.451E-08_dp, 3.806E-08_dp, &
         4.186E-08_dp, 4.592E-08_dp, 5.025E-08_dp, 5.486E-08_dp, 6.014E-08_dp             /)

    !--- Precalculated number fluxes per wind class [m-2 s-1]:

    REAL(dp) , PARAMETER :: numb1flux(0:40) = (/                               &
         0.000E+00_dp, 3.004E+01_dp, 3.245E+02_dp, 1.306E+03_dp, 3.505E+03_dp, 7.542E+03_dp, &
         1.410E+04_dp, 2.394E+04_dp, 3.787E+04_dp, 5.674E+04_dp, 8.147E+04_dp, 1.130E+05_dp, &
         1.523E+05_dp, 2.005E+05_dp, 2.586E+05_dp, 3.278E+05_dp, 4.091E+05_dp, 5.037E+05_dp, &
         6.129E+05_dp, 7.379E+05_dp, 8.800E+05_dp, 1.041E+06_dp, 1.220E+06_dp, 1.422E+06_dp, &
         1.646E+06_dp, 1.893E+06_dp, 2.166E+06_dp, 2.466E+06_dp, 2.794E+06_dp, 3.152E+06_dp, &
         3.541E+06_dp, 3.962E+06_dp, 4.419E+06_dp, 4.911E+06_dp, 5.441E+06_dp, 6.011E+06_dp, &
         6.621E+06_dp, 7.274E+06_dp, 7.972E+06_dp, 8.716E+06_dp, 8.801E+06_dp             /)

    REAL(dp) , PARAMETER :: numb2flux(0:40) = (/                               &
         0.000E+00_dp, 1.934E+01_dp, 2.068E+02_dp, 8.271E+02_dp, 2.211E+03_dp, 4.741E+03_dp, &
         8.841E+03_dp, 1.497E+04_dp, 2.363E+04_dp, 3.534E+04_dp, 5.066E+04_dp, 7.017E+04_dp, &
         9.447E+04_dp, 1.242E+05_dp, 1.600E+05_dp, 2.025E+05_dp, 2.525E+05_dp, 3.106E+05_dp, &
         3.776E+05_dp, 4.542E+05_dp, 5.413E+05_dp, 6.395E+05_dp, 7.501E+05_dp, 8.726E+05_dp, &
         1.009E+06_dp, 1.160E+06_dp, 1.327E+06_dp, 1.509E+06_dp, 1.709E+06_dp, 1.927E+06_dp, &
         2.163E+06_dp, 2.420E+06_dp, 2.697E+06_dp, 2.996E+06_dp, 3.318E+06_dp, 3.664E+06_dp, &
         4.034E+06_dp, 4.430E+06_dp, 4.852E+06_dp, 5.303E+06_dp, 5.740E+06_dp             /)

    !--- Precalculated mass flux gradient for each wind class for interpolation [kg m-3]:
    !    (dm/dv(i) where m = m(i) + dm/dv(i) * dv(i) )

    REAL(dp) , PARAMETER :: dmass1flux(0:40) = (/                              &
         2.483E-15_dp, 2.343E-14_dp, 7.630E-14_dp, 1.684E-13_dp, 3.054E-13_dp, 4.919E-13_dp, &
         7.319E-13_dp, 1.029E-12_dp, 1.386E-12_dp, 1.807E-12_dp, 2.294E-12_dp, 2.850E-12_dp, &
         3.477E-12_dp, 4.174E-12_dp, 4.960E-12_dp, 5.810E-12_dp, 6.745E-12_dp, 7.762E-12_dp, &
         8.863E-12_dp, 1.005E-11_dp, 1.132E-11_dp, 1.260E-11_dp, 1.423E-11_dp, 1.569E-11_dp, &
         1.733E-11_dp, 1.907E-11_dp, 2.090E-11_dp, 2.284E-11_dp, 2.487E-11_dp, 2.700E-11_dp, &
         2.924E-11_dp, 3.158E-11_dp, 3.403E-11_dp, 3.659E-11_dp, 3.925E-11_dp, 4.202E-11_dp, &
         4.490E-11_dp, 4.790E-11_dp, 5.100E-11_dp, 6.578E-11_dp, 6.578E-11_dp             /)

    REAL(dp) , PARAMETER :: dmass2flux(0:40) = (/                              &
         2.319E-13_dp, 2.179E-12_dp, 7.070E-12_dp, 1.557E-11_dp, 2.817E-11_dp, 4.528E-11_dp, &
         6.727E-11_dp, 9.446E-11_dp, 1.271E-10_dp, 1.655E-10_dp, 2.100E-10_dp, 2.606E-10_dp, &
         3.177E-10_dp, 3.812E-10_dp, 4.522E-10_dp, 5.297E-10_dp, 6.145E-10_dp, 7.068E-10_dp, &
         8.066E-10_dp, 9.142E-10_dp, 1.030E-09_dp, 1.149E-09_dp, 1.289E-09_dp, 1.425E-09_dp, &
         1.573E-09_dp, 1.730E-09_dp, 1.896E-09_dp, 2.070E-09_dp, 2.254E-09_dp, 2.447E-09_dp, &
         2.649E-09_dp, 2.860E-09_dp, 3.080E-09_dp, 3.311E-09_dp, 3.550E-09_dp, 3.800E-09_dp, &
         4.060E-09_dp, 4.329E-09_dp, 4.609E-09_dp, 5.279E-09_dp, 5.279E-09_dp             /)

    !--- Precalculated number flux gradient for each wind class for interpolation [m-3]:
    !    (dn/dv(i) where n = n(i) + dn/dv(i) * dv(i) )

    REAL(dp) , PARAMETER :: dnumb1flux(0:40) = (/                              &
         3.004E+01_dp, 2.945E+02_dp, 9.811E+02_dp, 2.200E+03_dp, 4.036E+03_dp, 6.562E+03_dp, &
         9.839E+03_dp, 1.393E+04_dp, 1.887E+04_dp, 2.473E+04_dp, 3.153E+04_dp, 3.935E+04_dp, &
         4.818E+04_dp, 5.808E+04_dp, 6.914E+04_dp, 8.130E+04_dp, 9.465E+04_dp, 1.092E+05_dp, &
         1.250E+05_dp, 1.421E+05_dp, 1.605E+05_dp, 1.798E+05_dp, 2.017E+05_dp, 2.237E+05_dp, &
         2.476E+05_dp, 2.729E+05_dp, 2.997E+05_dp, 3.280E+05_dp, 3.577E+05_dp, 3.890E+05_dp, &
         4.219E+05_dp, 4.564E+05_dp, 4.923E+05_dp, 5.301E+05_dp, 5.695E+05_dp, 6.105E+05_dp, &
         6.531E+05_dp, 6.976E+05_dp, 7.437E+05_dp, 8.550E+04_dp, 8.550E+04_dp             /)

    REAL(dp) , PARAMETER :: dnumb2flux(0:40) = (/                              &
         1.934E+01_dp, 1.875E+02_dp, 6.203E+02_dp, 1.384E+03_dp, 2.530E+03_dp, 4.100E+03_dp, &
         6.132E+03_dp, 8.659E+03_dp, 1.171E+04_dp, 1.532E+04_dp, 1.951E+04_dp, 2.430E+04_dp, &
         2.972E+04_dp, 3.582E+04_dp, 4.251E+04_dp, 4.997E+04_dp, 5.812E+04_dp, 6.700E+04_dp, &
         7.663E+04_dp, 8.702E+04_dp, 9.820E+04_dp, 1.106E+05_dp, 1.225E+05_dp, 1.366E+05_dp, &
         1.511E+05_dp, 1.664E+05_dp, 1.826E+05_dp, 1.997E+05_dp, 2.177E+05_dp, 2.366E+05_dp, &
         2.565E+05_dp, 2.773E+05_dp, 2.990E+05_dp, 3.218E+05_dp, 3.455E+05_dp, 3.702E+05_dp, &
         3.959E+05_dp, 4.226E+05_dp, 4.504E+05_dp, 4.376E+05_dp, 4.376E+05_dp             /)


    !--- Windclass as a function of wind speed (limited to 40 m/s):

    wcl(1:kproma)=INT(velo10m(1:kproma,krow))
    wcl(1:kproma)=MAX(0,MIN(wcl(1:kproma),40))

    !--- Calculate dv for interpolation in the windclas interval:

    dzspeed(1:kproma)=velo10m(1:kproma,krow)-REAL(wcl(1:kproma),dp)

    !--- Calculate fraction of the gridcell of non ice-covered water:

    zseafrac(1:kproma)=(1._dp-slf(1:kproma,krow)-alake(1:kproma,krow))*(1-pseaice(1:kproma))
    zseafrac(1:kproma)=MAX(0._dp,MIN(zseafrac(1:kproma),1._dp))

    !--- Accumulation mode soluble:

    !    Mass flux [kg m-2 s-1]:
    pxtems(1:kproma,idt_mssas)= (mass1flux(wcl(1:kproma))+dmass1flux(wcl(1:kproma)) * &
                                dzspeed(1:kproma))*zseafrac(1:kproma)

    !    Number flux [m-2 s-1]:
    pxtems(1:kproma,idt_nas)  = (numb1flux(wcl(1:kproma))+dnumb1flux(wcl(1:kproma)) * &
                                dzspeed(1:kproma))*zseafrac(1:kproma)

    !--- Coarse Mode Soluble:

    !    Mass flux [kg m-2 s-1]:
    pxtems(1:kproma,idt_msscs)= (mass2flux(wcl(1:kproma))+dmass2flux(wcl(1:kproma)) * &
                                dzspeed(1:kproma))*zseafrac(1:kproma)

    !    Number flux [m-2 s-1]:
    pxtems(1:kproma,idt_ncs)  = (numb2flux(wcl(1:kproma))+dnumb2flux(wcl(1:kproma)) * &
                                dzspeed(1:kproma))*zseafrac(1:kproma)

    !--- Accumulated emitted masses per area for diagnostics [kg m-2]:

    d_emi_mss(1:kproma,krow)   = d_emi_mss(1:kproma,krow) + &
                                 (pxtems(1:kproma,idt_mssas)+pxtems(1:kproma,idt_msscs))*delta_time

    d_emi_mssas(1:kproma,krow) = d_emi_mssas(1:kproma,krow)+pxtems(1:kproma,idt_mssas)*delta_time
    d_emi_msscs(1:kproma,krow) = d_emi_msscs(1:kproma,krow)+pxtems(1:kproma,idt_msscs)*delta_time
    d_emi_nssas(1:kproma,krow) = d_emi_nssas(1:kproma,krow)+pxtems(1:kproma,idt_nas)  *delta_time
    d_emi_nsscs(1:kproma,krow) = d_emi_nsscs(1:kproma,krow)+pxtems(1:kproma,idt_ncs)  *delta_time
    

  END SUBROUTINE seasalt_emissions_lsce


  SUBROUTINE piston_velocity(kproma, kbdim, krow, loland, ptsw)
    
    ! *piston_velocity* calculation of the piston velocity
    !                   for the tracers DMS, CO2 and NH3
    !
    ! Authors:
    ! --------
    ! S. Kloster MPI-MET 28/10/02
    ! 

    USE mo_aero_mem,      ONLY: velo10m, vdms
    USE mo_aero,          ONLY: npist              ! flag for piston velocity calculation   
    USE mo_constants,     ONLY: tmelt

    IMPLICIT NONE
 
    !--- Parameters:

    INTEGER, INTENT(IN)    :: kproma, kbdim, krow
    LOGICAL, INTENT(IN)    :: loland(kbdim)
    REAL(dp),    INTENT(IN)    :: ptsw(kbdim)
    
    !--- Local Variables:

    INTEGER                :: jl
    
    REAL(dp)                   :: zsst, zzspeed
    REAL(dp)                   :: zschmidt_dms,zschmidt_co2,zschmidt_nh3
    REAL(dp)                   :: zkw


    DO jl=1,kproma

       !--- 1.) Schmidt number

       zsst=ptsw(jl)-tmelt

       !--- 1.1.)  DMS Schmidt number after Andreae

       !--- Limit SST to valid range as the polynomial gets negative 
       !    for T > 35 C:

       zsst=MIN( 35.0_dp , zsst )

       zschmidt_dms=3652.047271_dp-246.99_dp*zsst+8.536397_dp*zsst*zsst     &
         &          -0.124397_dp*zsst*zsst*zsst
       
       !--- 1.2.) CO2 Schmidt number after Jaehne et al. (1987b)
!      zschmidt_co2=2073.1     -125.62*zsst+3.6276  *zsst*zsst     &
!                   -0.043219*zsst*zsst*zsst

       !--- 1.2.) NH3 Schmidt number 
!      zschmidt_nh3=2073.1     -125.62*zsst+3.6276  *zsst*zsst     &
!                   -0.043219*zsst*zsst*zsst
       
       !--- 2.)  piston velocity in [cm/hr]

       IF (loland(jl)) THEN   ! land
          zkw=0._dp
       ELSE                    ! ocean  
          zzspeed=velo10m(jl,krow)
          IF (npist == 1) THEN ! Calculate piston velocity (Liss&Merlivat, 1986) 
             IF (zzspeed.GT.3.6_dp.AND.zzspeed.LE.13._dp) THEN
                zkw=2.85_dp*zzspeed-9.65_dp
                vdms(jl,krow)=zkw*(zschmidt_dms/600._dp)**(-0.5_dp)
!               vco2(jl,krow)=zkw*(zschmidt_co2/600.)**(-0.5)
!               vnh3(jl,krow)=zkw*(zschmidt_nh3/600.)**(-0.5)
             ELSE IF(zzspeed.LE.3.6_dp) THEN
                zkw=0.17_dp*zzspeed
                vdms(jl,krow)=zkw*(zschmidt_dms/600._dp)**(-2._dp/3._dp)
!               vco2(jl,krow)=zkw*(zschmidt_co2/600.)**(-2./3.)
!               vnh3(jl,krow)=zkw*(zschmidt_nh3/600.)**(-2./3.)
             ELSE
                zkw=5.9_dp*zzspeed-49.3_dp
                vdms(jl,krow)=zkw*(zschmidt_dms/600._dp)**(-0.5_dp)
!               vco2(jl,krow)=zkw*(zschmidt_co2/600.)**(-0.5)
!               vnh3(jl,krow)=zkw*(zschmidt_nh3/600.)**(-0.5)
             END IF
          ELSE IF (npist == 2) THEN ! Calculate piston velocity (Wanninkhof, 1992)
             zkw=0.31_dp*zzspeed*zzspeed
             vdms(jl,krow)=zkw*(zschmidt_dms/660._dp)**(-0.5_dp)
!            vco2(jl,krow)=zkw*(zschmidt_co2/660.)**(-0.5)
!            vnh3(jl,krow)=zkw*(zschmidt_nh3/660.)**(-0.5)
          ELSE IF (npist == 3) THEN  ! Calculate piston velocity (Nightingale, 2000)
             zkw=(0.222_dp*zzspeed*zzspeed+0.333_dp*zzspeed)
             vdms(jl,krow)=zkw*(zschmidt_dms/660._dp)**(-0.5_dp)
!            vco2(jl,krow)=zkw*(zschmidt_co2/660.)**(-0.5)
!            vnh3(jl,krow)=zkw*(zschmidt_nh3/660.)**(-0.5)
          END IF
       END IF
    END DO

  END SUBROUTINE piston_velocity
 
    
  SUBROUTINE dms_emissions(kproma, kbdim, krow, loland, pseaice, pxtems)

    ! *dms_emissions* calculated DMS emission fluxes from the 
    !                 precalculated piston velocities
    !
    !    Authors:
    !    --------
    !    Johann Feichter, MPI-Met
    !    Philip Stier,    MPI-Met        06/2002
    !    Silvia Kloster,  MPI-Met        10/2002

    USE mo_aero_mem,      ONLY: d_emi_ms,                     &
                                d_emi_dms_sea, d_emi_dms_ter, &
                                emi_dms_sea,   emi_dms_ter,   &
                                vdms
    USE mo_aero_trac,     ONLY: idt_dms
    USE mo_time_control,  ONLY: delta_time
    USE mo_tracer,        ONLY: ntrac

    IMPLICIT NONE

    !--- Parameters:
    
    INTEGER, INTENT(In)     :: kproma, kbdim, krow
    LOGICAL, INTENT(in)     :: loland(kbdim)
    REAL(dp),    INTENT(in)     :: pseaice(kbdim)
    REAL(dp),    INTENT(inout)  :: pxtems(kbdim,ntrac)

    !--- Local Variables:

    INTEGER                 :: jl

    REAL(dp)                    :: zdmscon, zdmsemiss


    DO jl=1,kproma

       !--- 1) Calculate DMS emissions 

       IF (loland(jl)) THEN

          zdmscon=0._dp

       ELSE
          
          zdmscon=emi_dms_sea(jl,krow) ! no coupled mode use DMS seawater concentration from input stream
          
          !--- DMS emissions
          
          zdmsemiss=zdmscon*vdms(jl,krow)*32.064e-11_dp/3600._dp

          pxtems(jl,idt_dms)=pxtems(jl,idt_dms)+zdmsemiss*(1._dp-pseaice(jl))

       END IF

       !--- Store applied emission fluxes for diagnostics:

       d_emi_dms_sea(jl,krow) = d_emi_dms_sea(jl,krow) + pxtems(jl,idt_dms)*delta_time


       !--- 2) Add terrestrial DMS emission flux from stream:

       pxtems(jl,idt_dms)=pxtems(jl,idt_dms)+emi_dms_ter(jl,krow)

       !--- Store applied emission fluxes for diagnostics:

       d_emi_dms_ter(jl,krow) = d_emi_dms_ter(jl,krow) + emi_dms_ter(jl,krow)*delta_time


       !--- 3) Store total emission fluxes for diagnostics:

       d_emi_ms(jl,krow)      = d_emi_ms(jl,krow)      + pxtems(jl,idt_dms)*delta_time

    END DO

  END SUBROUTINE dms_emissions


  SUBROUTINE dust_emissions(kproma, kbdim, krow, pxtems)

    USE mo_tracer,        ONLY: ntrac
    USE mo_aero_trac,     ONLY: idt_mduai, idt_mduci, idt_nai,  idt_nci
    USE mo_aero_mem,      ONLY: d_emi_mdu
    USE mo_time_control,  ONLY: delta_time
    USE mo_exception,     ONLY: finish

    IMPLICIT NONE

    !--- Parameters:
    
    INTEGER, INTENT(in)     :: kproma, kbdim, krow
    REAL(dp),    INTENT(inout)  :: pxtems(kbdim,ntrac)
    
    !--- Local Variables:

    INTEGER                 :: jl

    !--- Constants:

    REAL(dp), PARAMETER         :: zdumassa= 0.37E-15_dp,   & ! Mass of accu mode du particle [kg]
                               zdumassc= 0.37E-12_dp      ! Mass of coar mode du particle [kg]


    !--- Initialization:

    DO jl=1,kproma

       !  dust
       !pxtems(jl,idt_mduai)= emi_du_acc(jl,krow)
       !pxtems(jl,idt_mduci)= emi_du_coa(jl,krow)
       CALL finish('xt_dust_emissions','Offline dust emission fields currently not implemented')

       ! Accumulation insoluble:
       pxtems(jl,idt_nai) = pxtems(jl,idt_mduai) / zdumassa
       
       ! Coarse insoluble:
       pxtems(jl,idt_nci) = pxtems(jl,idt_mduci) / zdumassc      
       
       d_emi_mdu(jl,krow) = d_emi_mdu(jl,krow)+                                  &
                            (pxtems(jl,idt_mduai)+pxtems(jl,idt_mduci))*delta_time

    END DO

  END SUBROUTINE dust_emissions


  SUBROUTINE dust_emissions_lsce(pxtems, kproma, kbdim, krow)
    !
    !     Authors:
    !     --------
    !     Michael Schulz; LSCE,Saclay (original source)   9. June 2002
    !     Philip Stier;   MPI-Met, Hamburg, (stier@dkrz.de)
    !                     (adaption to ECHAM5/HAM)           2002/2003
    !
    !     purpose
    !     -------
    !     compute interactively emission flux of mineral dust
    !
    !     interface
    !     ---------
    !      from aerosol modules
    !       zspeed         wind speed at 10 m                 [m/s]
    !       slm            sea land mask                      []
    !       ztemp          surface air temperature            [C]
    !       zprecip        precipitation amount               [kg m-2 s-1]
    !       zprecipinsoil  proxy for soil humidity            [mm]
    !       densdust       particle density
    !       rhv            source strength factor             [kg s2 m-5]
    !       wth            threshold velocity                 [m/s]
    !       cly            clay content                       [%]
    !
    !       srcsigma       sigma of aerosol mode 
    !       srcsigmaln
    !
    !      variables saved
    !       entered        input fields read flag
    !       mask           dust emission area mask            [] 
    !
    !      output 
    !       pxtems         emission of dust per tracer        [kg m-2 s-1]
    !       
    !     method
    !     ------
    !     
    !     The dust aerosol emitted is described as a lognormal size
    !     distribution with a mass median radius of 2.5 um and a standard
    !     deviation of s = 2. Such a source size distribution was found to
    !     provide a consistent mineral aerosol transport model, which
    !     provides enough mineral dust for long range transport from the
    !     Sahara to the Atlantic and the Mediterranean. Dust plumes
    !     simulated with such a model provide aerosol optical thickness
    !     values very close to those observed with satellites [Schulz et
    !     al., 1998] [Guelle et al., 2000]. Using a fixed lognormal
    !     distribution at the source implies that number concentrations can
    !     be derived from mass emission flux and imposed source size
    !     distribution [Schulz et al., 1998].
    !
    !     While surface wind speed is produced interactively or read from
    !     ECMWF winds (see routine aerosol_meteo_calc), the source function
    !     incorporates also spatial information on three parameters a) the
    !     source strength factor (rhv); b) the threshold velocity (wth) and
    !     c) the clay fraction of the uppermost soil layer (cly). 
    !
    !         eflux = max(rhv*(zspeed-wth)*zspeed**2,0.)*landmask
    !
    !     This work is based on an offline dust source developped by
    !     [Claquin, 1999].  He used a global TM3 dust simulation with a
    !     dustsource in all arid regions to obtain an approximate dust
    !     field. In a second step he compared regional averages of TOMS
    !     observations of aerosol index to this simulation, to obtain a
    !     source strength factor for a given region [see also Balkanski et
    !     al. [submitted , 2002]. The threshold velocity was obtained by
    !     comparing the FAO soil database to the maps of Marticorena [1995]
    !     to derive by this procedure a soil specific threshold velocity for
    !     desert soils. Combined with the high resolution FAO dataset this
    !     results in a global map of threshold velocities for dust rise in
    !     arid and semi-arid regions. The offline source formulation was
    !     converted to an interactive dust source for this work [Schulz und
    !     Timmreck, in prep] including prognostic variables produced by the
    !     GCM, such as wind speed, precipitation and surface temperature.
    !
    !     Dust emission is thought to be inhibited, when the surface soil is
    !     wettened or frozen. Drying of desert soils can be parameterised as
    !     a function of recent accumulated precipitation, surface
    !     temperature. The clay content of the soil is a key parameter for
    !     this process, since clay retains water for a longer period. While
    !     Claquin [1999] used an offline calculation to define the periods
    !     when the soil is available for emission of dust, we prefer an
    !     interactive approach here. Soil wettening and drying rate are
    !     computed from GCM calculated preceipitation and surface
    !     temperatures. The clay content is assembled from the FAO soil
    !     data. During freezing conditions, no drying of the soil and no
    !     emissions are assumed.  The parameterisation uses the conditions
    !     given by Claquin [1999] and computes an evaporation rate acting on
    !     the accumulated precipitation amount [Schulz und Timmreck, in
    !     prep]. This might overestimate the time which is needed to dry off
    !     the top soil, since no run-off or losses to the ground water
    !     reservoirs are assumed.
    !     (parameterised M. Schulz LSCE 13.2.2001)
    !     -----------------------------------------------------------------------

    USE mo_aero_mem,     ONLY: emi_du_thr, emi_du_src,  emi_du_cla,  velo10m, &
                               d_emi_mdu,  d_emi_mduci, d_emi_nduci, zprecip, &
                               zprecipinsoil
    USE mo_aero_trac,    ONLY: idt_mduci,  idt_nci
    USE mo_aero_m7,      ONLY: sigma,      icoai
    USE mo_tracer,       ONLY: ntrac, flag
    USE mo_time_control, ONLY: delta_time, time_step_len
    USE mo_memory_g3b,   ONLY: tslm1, slm
    USE mo_exception,    ONLY: finish
    USE mo_doctor,       ONLY: nout
    USE mo_constants,    ONLY: tmelt, api

    IMPLICIT NONE

    INTEGER :: kproma, kbdim, krow

    ! temp variables

    REAL(dp)          :: densdust       ! Particle density in kg/m3
    REAL(dp)          :: fdist          ! Number/Mass factor to compute number flux
    REAL(dp)          :: avgdryrate     ! Average drying rate [mm dt-1]

    REAL(dp)          :: drying(kbdim)  !
    REAL(dp)          :: clyfac(kbdim)

    REAL(dp)          :: ztempc(kbdim)
    REAL(dp)          :: pxtems(kbdim,ntrac)
    LOGICAL       :: lomask(kbdim)


    LOGICAL, SAVE :: entered = .FALSE.

    !--- Source mass median diameter [m]:

    REAL(dp), PARAMETER :: srcmmd= 2.5e-6_dp

    !--- Source standard deviation:

    REAL(dp), PARAMETER :: srcsigma= 2._dp

    
    !--- 1) Initializations:

    IF( .NOT. entered ) THEN

       !--- Ensure that standard deviations of the source and
       !    of the aerosol model M7 are consistent:
       !    Dust source for coarse mode insoluble only:

       IF ( ABS(sigma(icoai) - srcsigma) > 0.01_dp ) THEN

          WRITE (nout,*) 'xt_dustsource: M7 standard deviation and emission '
          WRITE (nout,*) '               standard deviation do not match!'
          CALL finish('xt_dustsource','Run terminated.')

       END IF

       entered = .TRUE.

    END IF

    !--- Get density of dust from the tr_list:

    densdust=flag('density',idt_mduci)

    !--- Temperature in C:

    ztempc(1:kproma)=MAX(-10._dp,MIN(50._dp,tslm1(1:kproma,krow)-tmelt))

    !--- Average drying rate [mm dt-1]:

    avgdryrate=300._dp/365._dp*time_step_len/86400._dp

    lomask(1:kproma)=(emi_du_cla(1:kproma,krow) < 9990._dp .AND. &
                      emi_du_thr(1:kproma,krow) < 9990._dp .AND. &
                      slm(1:kproma,krow)        > 0.5_dp         )


    !--- 2) Determine soil moisture:

    WHERE (lomask(1:kproma))

       !--- Accumulate precipitation:

       zprecipinsoil(1:kproma,krow)=zprecipinsoil(1:kproma,krow)+zprecip(1:kproma,krow)

       !--- Wet time of the soil depends on clay content:
       !
       !    (New soil wettening formulation f(clay content and soil temperature)
       !     Dependent on clay content and interpolated to obtain 
       !     clyfac, which is the maximim number of dry days after a rain event
       !     and corresponds to a maximum amount of water hold in top soil [mm] 
       !     at the lowest temperatures [5C]. The avg drying rate scales the
       !     drying to keep the area an arid area with an aridity limit assumed
       !     to be 300 mm precipitation per year.
       !     More rapid drying only depends on temperature and fits the 
       !     'dry days table' of Claquin to a function with two fitting parameters.)

       !--- Maximal amount of water hold in top soil [mm]:

       clyfac(1:kproma)=MIN(16._dp,emi_du_cla(1:kproma,krow)*0.4_dp+8._dp)

       !--- Amount of water drying in top soil during one timestep [mm]:

       drying(1:kproma)=avgdryrate*EXP(0.03905491_dp*EXP(0.17446_dp*ztempc(1:kproma)))

       !--- Remaining water in top soil [mm]:

       zprecipinsoil(1:kproma,krow)=MIN(MAX(0._dp,zprecipinsoil(1:kproma,krow)-drying(1:kproma)), &
                                        clyfac(1:kproma)                                        )

    ENDWHERE


    !--- 3) Computation of the mass flux [kg m-2 s-1]: 

    WHERE ((ztempc(1:kproma) > 0._dp) .AND. lomask(1:kproma) .AND. (zprecipinsoil(1:kproma,krow).LE.0._dp)) 

       pxtems(1:kproma,idt_mduci)=MAX((emi_du_src(1:kproma,krow)                          * &
                                       (velo10m(1:kproma,krow)-emi_du_thr(1:kproma,krow)) * &
                                       velo10m(1:kproma,krow)**2)                         , &
                                      0._dp)
    ELSEWHERE

       pxtems(1:kproma,idt_mduci)=0._dp

    ENDWHERE

    !--- 4) Computation of the emission number flux [m-2 s-1]:

    !--- Calculate conversion factor from mass flux to number flux for 
    !    log-normal distributions:

    fdist=1._dp/api*6._dp/densdust/srcmmd**3 *EXP(4.5_dp*LOG(srcsigma)**2)

    !--- Convert mass flux into number flux:

    pxtems(1:kproma,idt_nci) = pxtems(1:kproma,idt_mduci)*fdist

    !--- 5) Store the emission fluxes for diagnostics:

    d_emi_mdu(1:kproma,krow)   = d_emi_mdu(1:kproma,krow)  +pxtems(1:kproma,idt_mduci)      *delta_time

    d_emi_mduci(1:kproma,krow) = d_emi_mduci(1:kproma,krow)+pxtems(1:kproma,idt_mduci)      *delta_time
    d_emi_nduci(1:kproma,krow) = d_emi_nduci(1:kproma,krow)+pxtems(1:kproma,idt_mduci)*fdist*delta_time

  END SUBROUTINE dust_emissions_lsce

  SUBROUTINE so2_emissions(kproma, kbdim, klev, klevp1, krow, paphp1, pxtems, pxtte)

    USE mo_tracer,        ONLY: ntrac
    USE mo_aero_trac,     ONLY: idt_so2
    USE mo_aero_mem,      ONLY: d_emi_ms,                        &
                                d_emi_so2_wf,    d_emi_so2_ant,  &
                                emi_so2_wf,                      &
                                emi_so2_ant_low, emi_so2_ant_high
    USE mo_control,       ONLY: nlev
    USE mo_time_control,  ONLY: delta_time
    USE mo_constants,     ONLY: g
    USE mo_exception,     ONLY: finish

    IMPLICIT NONE

    !--- Parameters:

    INTEGER, INTENT(IN)      :: kproma, kbdim, klev, klevp1, krow

    REAL(dp),    INTENT(in)      :: paphp1(kbdim,klevp1)
    REAL(dp),    INTENT(inout)   :: pxtems(kbdim,ntrac)
    REAL(dp),    INTENT(inout)   :: pxtte(kbdim,klev,ntrac)

    !--- Local Variables:

    INTEGER                  :: jl, ilev


    !--- 0) Initializations: ----------------------------------------------

    IF (nlev/=19 .AND. nlev/=31) CALL finish('so2_emissions','missmatch with number of vertical levels')


    !--- 1) Surface emission------------------------------------------------        
    !
    !
    DO jl=1,kproma

       !--- Biomass Burning:

       pxtems(jl,idt_so2)=emi_so2_wf(jl,krow)

       !--- Add anthropogenic emissions:

       pxtems(jl,idt_so2)= pxtems(jl,idt_so2)+emi_so2_ant_low(jl,krow)

       !--- Store for diagnostics:

       d_emi_so2_wf(jl,krow) = d_emi_so2_wf(jl,krow)  + emi_so2_wf(jl,krow)     *delta_time
       d_emi_so2_ant(jl,krow)= d_emi_so2_ant(jl,krow) + emi_so2_ant_low(jl,krow)*delta_time
       d_emi_ms(jl,krow)     = d_emi_ms(jl,krow)      + pxtems(jl,idt_so2)      *delta_time  
    
    END DO

    !--- 2) Emission of anthropogenic SO2 in the second lowest model level as this 
    !       corresponds to the next higher level above 100m as given by the GEIA
    !       inventory. 
    !       Note: While surface emissions are specified as fluxes, these
    !       emissions are applied as tracer tendencies in the 
    !       respective model layer
    !    

    ilev=klev-1

    DO jl=1, kproma

       pxtte(jl,ilev,idt_so2)=pxtte(jl,ilev,idt_so2) +                                         &
                              emi_so2_ant_high(jl,krow)*g/(paphp1(jl,ilev+1)-paphp1(jl,ilev))

       !--- Add emissions to total sulfur emissions:

       d_emi_ms(jl,krow)     = d_emi_ms(jl,krow)      + emi_so2_ant_high(jl,krow)*delta_time 
       d_emi_so2_ant(jl,krow)= d_emi_so2_ant(jl,krow) + emi_so2_ant_high(jl,krow)*delta_time

    END DO

  END SUBROUTINE so2_emissions

  SUBROUTINE volcanic_so2_emissions(kproma,kbdim,klev,klevp1,krow,paphp1,pxtte,pgeom1)

    !
    !    Volcanic background emissions 
    !
    !   3 Emission levels:
    !
    !    1. pre-intra eruption in level ivolc-height (=top of volcano)
    !    2. post-extra eruption in level 15 -16 (ca 550-1736m)
    !    3. explosive eruption in level 10 - 11 (ca 5000-7900m)
    !
    !       zvolcemi  total emission from volcanoes in tg/yr
    !
    !    Author: H. Feichter,    MI, August 1991, original source
    !            U. Schlese,     DKRZ, January 1995, changed
    !            L. Kornblueh,   MPI, May 1998, f90 rewrite
    !            U. Schulzweida, MPI, May 1998, f90 rewrite
    !            H.-S. Bauer and M. Schultz, MPI, November 2000, 
    !               (inclusion into new chemistry structure) 
    !            P. Stier,       MPI, 2001-2003
    !               (rewritten, removed resolution dependencies)
    !
    
    USE mo_time_control,  ONLY: delta_time
    USE mo_aero_trac,     ONLY: idt_so2
    USE mo_aero_mem,      ONLY: emi_so2_vol, emi_so2_vol_alt, emi_so2_vol_lev, &
                                d_emi_ms,    d_emi_so2_vol
    USE mo_constants,     ONLY: g
    USE mo_tracer,        ONLY: ntrac

    ! Scalar arguments
    INTEGER, INTENT(in) :: kproma, kbdim, klev, klevp1, krow

    ! Array arguments
    REAL(dp), INTENT(in)    :: pgeom1(kbdim,klev)
    REAL(dp), INTENT(in)    :: paphp1(kbdim,klevp1)
    REAL(dp), INTENT(inout) :: pxtte(kbdim,klev,ntrac)
    
    ! Local arrays
    REAL(dp) :: zvolcano(kbdim)
    
    ! Local scalars
    REAL(dp) :: zvolcemi1,   & !
            zvolcemi2,   & !
            zheight,     & !
            zscale,      &
            zgdp
    
    INTEGER :: jl, jk, ilev, ibelow
    
    REAL(dp),    PARAMETER :: zvolcemi       = 8.0_dp ! Total emission from volcanos [Tg/yr]
                                               ! Low range of the estimate of 
                                               ! Graf et al. (1997)

    !--- 0) Initialisations:
    
    zvolcano(1:kproma)=emi_so2_vol(1:kproma,krow)

    !--- Split emissions: 50% at volcano height and 50% at fixed levels:

    zvolcemi1=zvolcemi*0.5_dp
    zvolcemi2=zvolcemi*0.5_dp

    DO jl=1,kproma

       IF (emi_so2_vol(jl,krow) > 0._dp ) THEN

          !--- Find corresponding model level to emission height:

          DO jk=1, klev

             zheight=pgeom1(jl,jk)/g
                
             IF(emi_so2_vol_alt(jl,krow)<zheight) THEN

                CYCLE

             ELSE IF(emi_so2_vol_alt(jl,krow)>=zheight) THEN

                !--- Select level that is one above the emission height:

                ilev=jk-1

                emi_so2_vol_lev(jl,krow)=REAL(ilev,dp)

                EXIT

             END IF

          END DO

          !--- Emit 50% of the emissions in the emission level:

          zgdp=g/(paphp1(jl,ilev+1)-paphp1(jl,ilev))

          pxtte(jl,ilev,idt_so2)=pxtte(jl,ilev,idt_so2)+             &
                                 zvolcemi1*zvolcano(jl)*zgdp

          !--- Distribute 50% of the emissions to all levels below the emission level:

          !--- Number of levels below emission level:

          ibelow=klev-ilev

          !--- Fraction of the 50% emissions for each layer below emission level:

          zscale=1._dp/REAL(ibelow,dp)

          !---  Emission from bottom to level below emission level:

          DO jk=klev, ilev+1, -1

             zgdp=g/(paphp1(jl,jk+1)-paphp1(jl,jk))

             pxtte(jl,jk,idt_so2)=pxtte(jl,jk,idt_so2)+              &
                                  zscale*zvolcemi2*zvolcano(jl)*zgdp

          END DO

          !--- Store volcanic emissions for diagnostics:

          d_emi_ms(jl,krow)=d_emi_ms(jl,krow)+zvolcemi*zvolcano(jl)          *delta_time

          d_emi_so2_vol(jl,krow)=d_emi_so2_vol(jl,krow)+zvolcemi*zvolcano(jl)*delta_time

          
       END IF
    END DO

  END SUBROUTINE volcanic_so2_emissions

  SUBROUTINE carbon_emissions(kproma, kbdim, krow, pxtems)

    USE mo_tracer,        ONLY: ntrac
    USE mo_time_control,  ONLY: delta_time,  get_date_components, &
                                current_date
    USE mo_geoloc,        ONLY: philat_2d
    USE mo_aero,          ONLY: zm2n_bcki_ff, zm2n_bcki_bb, &
                                zm2n_ocki_ff,               &
                                zm2n_ocki_bb, zm2n_ocki_bg, &
                                zm2n_ocks_bb, zm2n_ocks_bg, &
                                zom2oc,                     &
                                zbb_wsoc_perc,zbge_wsoc_perc
    USE mo_aero_trac,     ONLY: idt_mbcki,                         &
                                idt_mocki,    idt_mocks,           &
                                idt_nki,      idt_nks
    USE mo_aero_mem,      ONLY: d_emi_mbc,    d_emi_moc,           &
                                d_emi_bc_ant, d_emi_bc_wf,         &
                                d_emi_bc_ag,  d_emi_oc_ant,        &
                                d_emi_oc_bge, d_emi_oc_wf,         &
                                d_emi_oc_ag,                       &
                                emi_bc_ant,   emi_bc_wf,           &
                                emi_bc_ag,    emi_oc_ant,          &
                                emi_oc_bge,   emi_oc_wf,           &
                                emi_oc_ag

    IMPLICIT NONE

    !--- Parameters:
    
    INTEGER, INTENT(in)     :: kproma, kbdim, krow
    REAL(dp),    INTENT(inout)  :: pxtems(kbdim,ntrac)

    !--- Local Variables:

    INTEGER                 :: jl, imonth

    REAL(dp)                    :: zcarbonseas

    !--- Define paramters:

    !    Seasonality of fossil fuel carbon emissions after Rotty (1987)

    LOGICAL, PARAMETER      :: lseason=.TRUE.

    REAL(dp), PARAMETER     :: zcarbonseas_nh(12)=(/1.146_dp, 1.139_dp, 1.081_dp, 0.995_dp, 0.916_dp, 0.920_dp,&
                                                    0.910_dp, 0.907_dp, 0.934_dp, 0.962_dp, 1.019_dp, 1.072_dp/)


    !--- 0) Initialisation:

    CALL get_date_components (current_date,  month=imonth)


    !--- 1) Apply the emissions:

    DO jl=1, kproma

       !--- 1.1) Calculate seasonality for anthropogenic emissions:

       !--- Northern Hemisphere:

       IF(lseason .AND. philat_2d(jl,krow)>0._dp) THEN

          zcarbonseas=zcarbonseas_nh(imonth)

       !--- Southern Hemisphere (no seasonaltiy):

       ELSE 

          zcarbonseas=1._dp

       END IF

       !--- 1.2) Carbon mass flux:

       !--- Black Carbon:

       !    Anthropogenic:
 
       pxtems(jl,idt_mbcki)=emi_bc_ant(jl,krow)*zcarbonseas

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            emi_bc_ant(jl,krow)*zcarbonseas * zm2n_bcki_ff

       !    Add Biomass Burning, agricultural waste and fuelwood:

       pxtems(jl,idt_mbcki)=pxtems(jl,idt_mbcki)+ &
                            emi_bc_wf(jl,krow)+emi_bc_ag(jl,krow)

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            (emi_bc_wf(jl,krow)+emi_bc_ag(jl,krow)) * zm2n_bcki_bb

       !    Store emitted mass for diagnostics:

       d_emi_bc_ant(jl,krow)=d_emi_bc_ant(jl,krow)+ &
                             emi_bc_ant(jl,krow)*zcarbonseas*delta_time
       d_emi_bc_wf(jl,krow) =d_emi_bc_wf(jl,krow) + &
                             emi_bc_wf(jl,krow)*delta_time
       d_emi_bc_ag(jl,krow) =d_emi_bc_ag(jl,krow) + &
                             emi_bc_ag(jl,krow)*delta_time
       d_emi_mbc(jl,krow)   =d_emi_mbc(jl,krow)   + &
                             pxtems(jl,idt_mbcki)*delta_time        
       
       !--- Organic Carbon:

       !    Anthropogenic:
 
       pxtems(jl,idt_mocki)=emi_oc_ant(jl,krow)*zcarbonseas*zom2oc

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            emi_oc_ant(jl,krow)*zcarbonseas*zom2oc * zm2n_ocki_ff

       !    Add Biomass Burning, agricultural waste and fuelwood - insoluble fraction:

       pxtems(jl,idt_mocki)=pxtems(jl,idt_mocki)+ &
                            (emi_oc_wf(jl,krow)+emi_oc_ag(jl,krow))*zom2oc*(1._dp-zbb_wsoc_perc)

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            (emi_oc_wf(jl,krow)+emi_oc_ag(jl,krow))*zom2oc*(1._dp-zbb_wsoc_perc)*zm2n_ocki_bb

       !    Add Biomass Burning, agricultural waste and fuelwood - soluble fraction:

       pxtems(jl,idt_mocks)=pxtems(jl,idt_mocks)+ &
                            (emi_oc_wf(jl,krow)+emi_oc_ag(jl,krow))*zom2oc*zbb_wsoc_perc

       pxtems(jl,idt_nks)  =pxtems(jl,idt_nks)+ &
                            (emi_oc_wf(jl,krow)+emi_oc_ag(jl,krow))*zom2oc*zbb_wsoc_perc * zm2n_ocks_bb

       !    Add Biogenic - insoluble fraction:

       pxtems(jl,idt_mocki)=pxtems(jl,idt_mocki)+ &
                            emi_oc_bge(jl,krow)*zom2oc*(1._dp-zbge_wsoc_perc)

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            emi_oc_bge(jl,krow)*zom2oc*(1._dp-zbge_wsoc_perc) * zm2n_ocki_bg

       !    Add Biogenic - soluble fraction:

       pxtems(jl,idt_mocks)=pxtems(jl,idt_mocks)+ &
                            emi_oc_bge(jl,krow)*zom2oc*zbge_wsoc_perc

       pxtems(jl,idt_nks)  =pxtems(jl,idt_nks)+ &
                            emi_oc_bge(jl,krow)*zom2oc*zbge_wsoc_perc * zm2n_ocks_bg

       !    Store emitted mass for diagnostics:

       d_emi_oc_ant(jl,krow)=d_emi_oc_ant(jl,krow)+ &
                             emi_oc_ant(jl,krow)*zcarbonseas*zom2oc * delta_time
       d_emi_oc_wf(jl,krow) =d_emi_oc_wf(jl,krow) + &
                             emi_oc_wf(jl,krow)*zom2oc              * delta_time
       d_emi_oc_ag(jl,krow) =d_emi_oc_ag(jl,krow) + &
                             emi_oc_ag(jl,krow)*zom2oc              * delta_time
       d_emi_oc_bge(jl,krow)=d_emi_oc_bge(jl,krow)+ &
                             emi_oc_bge(jl,krow)*zom2oc                     * delta_time
       d_emi_moc(jl,krow)   =d_emi_moc(jl,krow)   + &
                             (pxtems(jl,idt_mocki)+pxtems(jl,idt_mocks))    * delta_time

    END DO

  END SUBROUTINE carbon_emissions

END MODULE mo_chem_bcond
