MODULE mo_chem_bcond_aerocom
  !
  ! *mo_chem_bcond_aerocom* handles all emissions tracers specific for aerocom settings
  !
  !

  USE mo_kind,             ONLY: dp

  IMPLICIT NONE
  
CONTAINS

    SUBROUTINE xt_cont_volc_emiss_ini_aerocom

    ! *xt_cont_volc_emiss_ini_aerocom* reads and prepares the volcanic emissions
    !   
    ! adopted from xt_volc_emiss_ini for AEROCOM emissions continuous volcanos
    !  
    ! Interface:
    ! ----------
    ! *xt_cont_volc_emiss_ini_aerocom* is called from *read_chem_files*
    !
    ! Authors:
    ! --------
    ! Philip Stier, MPI-MET                      2002  
    !  
    !   
    ! Hans Feichter, MPI-MET 
    ! (original offline interpolation routine)   

    USE mo_exception,     ONLY: finish
    USE mo_control,       ONLY: nlon, ngl
    USE mo_filename,      ONLY: find_next_free_unit
    USE mo_doctor,        ONLY: nout
    USE mo_aero_mem,      ONLY: emi_so2_cvol,&
                                emi_so2_cvol_alt_high, emi_so2_cvol_alt_low
    USE mo_mpi,           ONLY: p_pe, p_io, p_bcast, p_parallel_io
    USE mo_transpose,     ONLY: scatter_gp
    USE mo_decomposition, ONLY: gl_dc => global_decomposition
    USE mo_gaussgrid,     ONLY: gridarea
    USE mo_aero_tools,    ONLY: geoindex

    !--- Local variables: 
    
    LOGICAL          :: lex        ! .true. when input file exists
    CHARACTER(len=256) :: ccfile ='continuous_volcanos.dat'     ! continuous errupting volcanos
    INTEGER          :: klon,     klat,     ierr,    inumtot, &
                        inum,     ilon,     ilat,             &
                        jvol,     iunit
    REAL(dp)             :: zheight_low, zheight_high
    REAL(dp)             :: zxemis,   zsumemi
    REAL(dp),    POINTER :: ztmp1(:,:), ztmp2(:,:), ztmp3(:,:)

    INTEGER, POINTER :: zheights_low(:), zheights_high(:), &
                        ilats(:), ilons(:)
    REAL(dp),    POINTER :: zemiss(:)

     !--- 1) Read fields:
     
     IF (p_parallel_io) iunit = find_next_free_unit (20, 99)
     
     !--- 1.1) Inquire: 
     IF (p_pe == p_io) THEN
        INQUIRE (file=ccfile,exist=lex) !if IO processor, check file existence
     END IF
     
     CALL p_bcast(lex,p_io)
     
     IF (p_pe == p_io) THEN
        IF (lex) THEN
           ALLOCATE (ztmp1(nlon,ngl)) 
           ALLOCATE (ztmp2(nlon,ngl)) 
           ALLOCATE (ztmp3(nlon,ngl)) 
           OPEN (iunit,file=ccfile,form='FORMATTED')
           REWIND(iunit)
           WRITE (nout,*) '------------------------------------------------------------------------------'
           WRITE (nout,*) '--- Initialization of volcanic SO2 emissions ---------------------------------'
           WRITE (nout,*) 
           WRITE (nout,*) '   xt_volc_emiss_ini: ' 
           WRITE (nout,*) 
           WRITE (nout,*) '      reading volcanic SO2 explosive emissions from file ', ccfile
        ELSE
           WRITE (nout,*) ' Trying to read from unit ', iunit
           WRITE (nout,*) ' Trying to read file ', ccfile
           WRITE (nout,*) ' File for continuous volcanic SO2 emissions does not exist.'
           CALL finish('xt_read_volc_emis','Error reading input file (1)')
        END IF
     END IF

    IF (lex) THEN
       IF (p_pe == p_io) THEN
          inumtot=0
          inum=0
          zsumemi=0._dp

          !--- 1.2) Determine number of fields:
          DO
             READ(iunit,*,IOSTAT=ierr)
             
             IF(ierr < 0 ) THEN
                REWIND(iunit)
                EXIT
             ELSE IF(ierr > 0) THEN
                CALL finish('xt_read_volc_emis','Error reading input file (2)')
             END IF

             inumtot=inumtot+1

          END DO

          !--- 1.3) Allocate fields and initialize:

          ALLOCATE(zheights_low(inumtot))
          ALLOCATE(zheights_high(inumtot))
          ALLOCATE(zemiss(inumtot))
          ALLOCATE(ilats(inumtot))
          ALLOCATE(ilons(inumtot))

          zheights_high=0
          zheights_low=0
          ilats=0
          ilons=0
          zemiss=0._dp
          ztmp1=0._dp
          ztmp2=0._dp
          ztmp3=0._dp

          !--- 1.4) Read emissions:

          DO
             READ(iunit,fmt='(2(1x,i4),3(1x,e10.4))',IOSTAT=ierr) ilon, ilat, zxemis, zheight_low, zheight_high         
             IF(ierr < 0 ) THEN
                EXIT
             ELSE IF(ierr > 0) THEN
                CALL finish('xt_read_volc_emis','Error reading input file (3)')
             END IF
                
             ! Store the read data in vectors:

             inum=inum+1
             zheights_low(inum)=zheight_low
             zheights_high(inum)=zheight_high
             zemiss(inum)=zxemis
             IF(ilon<0)THEN
                ilons(inum)=ilon+360
             ELSE IF(ilon>=0)THEN
                ilons(inum)=ilon
             END IF
             ilats(inum)=ilat
             
             ! Sum up the total emissions:

             zsumemi=zsumemi+zxemis
             
          END DO

          !--- 2) Store on 2D field

          DO jvol=1, inum

             !--- Find corresponding lat-lon indices:

             CALL geoindex(REAL(ilons(jvol),dp),REAL(ilats(jvol),dp),klon,klat)

             !--- Calculate the emissions flux [kg(SO2)/m**2/s] and store in 2D buffer:
             
             ztmp1(klon,klat)=ztmp1(klon,klat)+zemiss(jvol)/(gridarea(klat)*86400._dp*365._dp)

             !--- Store corresponding emission heights:
 
             ztmp2(klon,klat)=zheights_low(jvol)

             ztmp3(klon,klat)=zheights_high(jvol) 

          END DO

       END IF

       !--- 4) Scatter to streams over the processors:

       CALL scatter_gp(ztmp1, emi_so2_cvol, gl_dc)

       CALL scatter_gp(ztmp2, emi_so2_cvol_alt_low, gl_dc)

       CALL scatter_gp(ztmp3, emi_so2_cvol_alt_high, gl_dc)
       
       !--- 5) Deallocate temporal field and close input file

       IF (p_pe == p_io) THEN

          DEALLOCATE(ztmp1,ztmp2,ztmp3)
          DEALLOCATE(zheights_low,zheights_high,zemiss,ilats,ilons)

          CLOSE (iunit)    

          WRITE (nout,*) '--- End of the initialization of continuoes volcanic SO2 emissions ------------'
          WRITE (nout,*) '-------------------------------------------------------------------------------'

       END IF

    END IF

  END SUBROUTINE xt_cont_volc_emiss_ini_aerocom

  
  SUBROUTINE xt_expl_volc_emiss_ini_aerocom
       
    ! *xt_volc_emiss_ini* reads and prepares the volcanic emissions
    !
    ! adopted from xt_volc_emiss_ini for AEROCOM emissions explosive volcanos
    !
    ! Interface:
    ! ----------
    ! *xt_expl_volc_emiss_ini_aerocom* is called from *read_chem_files*
    !
    ! Authors:
    ! --------
    ! Philip Stier, MPI-MET                      2002
    !
    ! Hans Feichter, MPI-MET 
    ! (original offline interpolation routine)   

    USE mo_exception,     ONLY: finish
    USE mo_control,       ONLY: nlon, ngl
    USE mo_filename,      ONLY: find_next_free_unit
    USE mo_doctor,        ONLY: nout
    USE mo_aero_mem,      ONLY: emi_so2_evol,&
                                emi_so2_evol_alt_high, emi_so2_evol_alt_low
    USE mo_mpi,           ONLY: p_pe, p_io, p_bcast, p_parallel_io
    USE mo_transpose,     ONLY: scatter_gp
    USE mo_decomposition, ONLY: gl_dc => global_decomposition
    USE mo_gaussgrid,     ONLY: gridarea
    USE mo_aero_tools,    ONLY: geoindex 
    !--- Local variables: 
    
    LOGICAL          :: lex        ! .true. when input file exists
    CHARACTER(len=256) :: ccfile='explosive_volcanos.dat'     ! explosive errupting volcanos
    INTEGER          :: klon,     klat,     ierr,    inumtot, &
                        inum,     ilon,     ilat,             &
                        jvol,     iunit
    REAL(dp)             :: zheight_low, zheight_high
    REAL(dp)             :: zxemis,   zsumemi
    REAL(dp),    POINTER :: ztmp1(:,:), ztmp2(:,:), ztmp3(:,:)

    INTEGER, POINTER :: zheights_low(:), zheights_high(:), &
                        ilats(:), ilons(:)
    REAL(dp),    POINTER :: zemiss(:)


     !--- 1) Read fields:
     
     IF (p_parallel_io) iunit = find_next_free_unit (20, 99)
     
     !--- 1.1) Inquire: 

     IF (p_pe == p_io) THEN
        INQUIRE (file=ccfile,exist=lex) !if IO processor, check file existence
     END IF
     
     CALL p_bcast(lex,p_io)
     
     IF (p_pe == p_io) THEN
        IF (lex) THEN
           ALLOCATE (ztmp1(nlon,ngl)) 
           ALLOCATE (ztmp2(nlon,ngl)) 
           ALLOCATE (ztmp3(nlon,ngl)) 
           OPEN (iunit,file=ccfile,form='FORMATTED')
           REWIND(iunit)
           WRITE (nout,*) '------------------------------------------------------------------------------'
           WRITE (nout,*) '--- Initialization of volcanic SO2 emissions ---------------------------------'
           WRITE (nout,*) 
           WRITE (nout,*) '   xt_volc_emiss_ini: ' 
           WRITE (nout,*) 
           WRITE (nout,*) '      reading volcanic SO2 explosive emissions from file ', ccfile
        ELSE
           WRITE (nout,*) ' Trying to read from unit ', iunit
           WRITE (nout,*) ' Trying to read file ', ccfile
           WRITE (nout,*) ' File for explosive volcanic SO2 emissions does not exist.'
           CALL finish('xt_read_volc_emis','Error reading input file (1)')
        END IF
     END IF

    IF (lex) THEN
       IF (p_pe == p_io) THEN
          inumtot=0
          inum=0
          zsumemi=0._dp

          !--- 1.2) Determine number of fields:
          DO
             READ(iunit,*,IOSTAT=ierr)
             
             IF(ierr < 0 ) THEN
                REWIND(iunit)
                EXIT
             ELSE IF(ierr > 0) THEN
                CALL finish('xt_read_volc_emis','Error reading input file (2)')
             END IF

             inumtot=inumtot+1

          END DO

          !--- 1.3) Allocate fields and initialize:

          ALLOCATE(zheights_low(inumtot))
          ALLOCATE(zheights_high(inumtot))
          ALLOCATE(zemiss(inumtot))
          ALLOCATE(ilats(inumtot))
          ALLOCATE(ilons(inumtot))

          zheights_high=0
          zheights_low=0
          ilats=0
          ilons=0
          zemiss=0._dp
          ztmp1=0._dp
          ztmp2=0._dp
          ztmp3=0._dp

          !--- 1.4) Read emissions:

          DO
             READ(iunit,fmt='(2(1x,i4),3(1x,e10.4))',IOSTAT=ierr) ilon, ilat, zxemis, zheight_low, zheight_high         
             IF(ierr < 0 ) THEN
                EXIT
             ELSE IF(ierr > 0) THEN
                CALL finish('xt_read_volc_emis','Error reading input file (3)')
             END IF
                
             ! Store the read data in vectors:

             inum=inum+1
             zheights_low(inum)=zheight_low
             zheights_high(inum)=zheight_high
             zemiss(inum)=zxemis
             IF(ilon<0)THEN
                ilons(inum)=ilon+360
             ELSE IF(ilon >= 0)THEN
                ilons(inum)=ilon
             END IF
             ilats(inum)=ilat
             
             ! Sum up the total emissions:

             zsumemi=zsumemi+zxemis
             
          END DO

          !--- 2) Store on 2D field

          DO jvol=1, inum

             !--- Find corresponding lat-lon indices:

             CALL geoindex(REAL(ilons(jvol),dp),REAL(ilats(jvol),dp),klon,klat)

             !--- Calculate the emissions flux [kg(SO2/m**2/s)]and store in 2D buffer:
             
             ztmp1(klon,klat)=ztmp1(klon,klat)+zemiss(jvol)/(gridarea(klat)*86400._dp*365._dp)

             !--- Store corresponding emission heights:

             ztmp2(klon,klat)=zheights_low(jvol)

             ztmp3(klon,klat)=zheights_high(jvol) 

          END DO
       END IF

       !--- 4) Scatter to streams over the processors:

       CALL scatter_gp(ztmp1, emi_so2_evol, gl_dc)

       CALL scatter_gp(ztmp2, emi_so2_evol_alt_low, gl_dc)

       CALL scatter_gp(ztmp3, emi_so2_evol_alt_high, gl_dc)
       
       !--- 5) Deallocate temporal field and close input file

       IF (p_pe == p_io) THEN

          DEALLOCATE(ztmp1,ztmp2,ztmp3)
          DEALLOCATE(zheights_low,zheights_high,zemiss,ilats,ilons)

          CLOSE (iunit)    

          WRITE (nout,*) '--- End of the initialization of explosive volcanic SO2 emissions ------------'
          WRITE (nout,*) '------------------------------------------------------------------------------'

       END IF

    END IF

  END SUBROUTINE xt_expl_volc_emiss_ini_aerocom

 SUBROUTINE cont_volc_so2_emissions_aerocom(kproma,kbdim,klev,klevp1,krow,paphp1,pxtte,pgeom1)

    !
    !    Volcanic background emissions 
    !
    !    volcanic_so2_emissions adopted for AEROCOM continuous volcanos emissions
    !      
    !    Author: H. Feichter,    MI, August 1991, original source
    !            U. Schlese,     DKRZ, January 1995, changed
    !            L. Kornblueh,   MPI, May 1998, f90 rewrite
    !            U. Schulzweida, MPI, May 1998, f90 rewrite
    !            H.-S. Bauer and M. Schultz, MPI, November 2000, 
    !               (inclusion into new chemistry structure) 
    !            P. Stier,       MPI, 2001-2003
    !               (rewritten, removed resolution dependencies)
    !
    
    USE mo_time_control,  ONLY: delta_time
    USE mo_aero_trac,     ONLY: idt_so2,                           &
                                idt_ms4ks,    idt_ms4as,           &
                                idt_nas   ,   idt_nks
    USE mo_aero_mem,      ONLY: emi_so2_cvol,                                &
                                emi_so2_cvol_alt_high, emi_so2_cvol_alt_low ,&
                                emi_so2_cvol_lev1,     emi_so2_cvol_lev2,    &
                                d_emi_ms,                                    &
                                d_emi_so2_cvol,        d_emi_so4_cvol
    USE mo_constants,     ONLY: g
    USE mo_tracer,        ONLY: ntrac
    USE mo_exception,     ONLY: finish
    USE mo_aero,          ONLY: so2ts,facso2 ,&
                                zm2n_s4ks_sk, zm2n_s4as_sa

    ! Scalar arguments
    INTEGER, INTENT(in) :: kproma, kbdim, klev, klevp1, krow

    ! Array arguments
    REAL(dp), INTENT(in)    :: pgeom1(kbdim,klev)
    REAL(dp), INTENT(in)    :: paphp1(kbdim,klevp1)
    REAL(dp), INTENT(inout) :: pxtte(kbdim,klev,ntrac)
    
    
    ! Local scalars
    REAL(dp) :: zgdp
    
    INTEGER :: jl, jk, ilev_high, ilev_low, inumlev

    REAL(dp) :: zheight(klev)

    !--- 0) Initialisations:
  
    DO jl=1,kproma

       zheight(:)=pgeom1(jl,:)/g
       
       IF (emi_so2_cvol(jl,krow) > EPSILON(1.0_dp) ) THEN

          !--- Find corresponding model level to emission height high:

          IF (emi_so2_cvol_alt_high(jl,krow) <= zheight(klev)) THEN

             ilev_high=klev

          ELSE          
          
             DO jk=1, klev
             
                IF(emi_so2_cvol_alt_high(jl,krow)<zheight(jk)) THEN
                   
                   CYCLE
                
                ELSE IF(emi_so2_cvol_alt_high(jl,krow)>=zheight(jk)) THEN                             
                
                   ilev_high=jk
                
                   emi_so2_cvol_lev1(jl,krow)=REAL(ilev_high,dp)
                
                   EXIT
                
                END IF
             
             END DO

          END IF

          
          !--- Find corresponding model level to emission height low:

          IF (emi_so2_cvol_alt_low(jl,krow) <= zheight(klev)) THEN 

            ilev_low=klev

          ELSE  
          
             DO jk=1,klev
             
                IF(emi_so2_cvol_alt_low(jl,krow)<zheight(jk)) THEN
                   
                   CYCLE
                   
                ELSE IF(emi_so2_cvol_alt_low(jl,krow)>=zheight(jk)) THEN
                   
                   ilev_low=jk
                
                   emi_so2_cvol_lev2(jl,krow)=REAL(ilev_low,dp)

                   EXIT
                   
                END IF
                
             END DO

          END IF

          !--- Emit the emissions between the emission level ilev_low and ilev_high:

          inumlev=ilev_low-ilev_high+1

          IF (inumlev <= 0 ) CALL finish('cont_volcanic_so2_emissions_aerocom','Error in emission heights')                     

          DO jk=ilev_high,ilev_low

          zgdp=g/(paphp1(jl,jk+1)-paphp1(jl,jk))

          pxtte(jl,jk,idt_so2)=pxtte(jl,jk,idt_so2)+             &
                               (1._dp/REAL(inumlev,dp))*emi_so2_cvol(jl,krow)*so2ts*facso2*zgdp

          !--- Primary SO4 emissions

          pxtte(jl,jk,idt_ms4ks)=pxtte(jl,jk,idt_ms4ks) +  &
                                 0.5_dp*(1._dp/REAL(inumlev,dp))*emi_so2_cvol(jl,krow)*so2ts*(1._dp-facso2)*zgdp
       
          pxtte(jl,jk,idt_nks)=pxtte(jl,jk,idt_nks) + &
                               0.5_dp*(1._dp/REAL(inumlev,dp))*emi_so2_cvol(jl,krow)*so2ts*(1._dp-facso2)*zm2n_s4ks_sk*zgdp 

          !--- 50% Accumulation
       
          pxtte(jl,jk,idt_ms4as)=pxtte(jl,jk,idt_ms4as) +  &
                                 0.5_dp*(1._dp/REAL(inumlev,dp))*emi_so2_cvol(jl,krow)*so2ts*(1._dp-facso2)*zgdp
       
          pxtte(jl,jk,idt_nas)=pxtte(jl,jk,idt_nas) + &
                               0.5_dp*(1._dp/REAL(inumlev,dp))*emi_so2_cvol(jl,krow)*so2ts*(1._dp-facso2)*zm2n_s4as_sa*zgdp 

         
          END DO
 
          !--- Store volcanic emissions for diagnostics:

          d_emi_ms(jl,krow)=d_emi_ms(jl,krow)+emi_so2_cvol(jl,krow)            *so2ts       *delta_time

          d_emi_so2_cvol(jl,krow)=d_emi_so2_cvol(jl,krow)+emi_so2_cvol(jl,krow)*so2ts*facso2*delta_time
          
          d_emi_so4_cvol(jl,krow)=d_emi_so4_cvol(jl,krow)+emi_so2_cvol(jl,krow)*so2ts*(1._dp-facso2)*delta_time
          
       
       END IF
    END DO

  END SUBROUTINE cont_volc_so2_emissions_aerocom

  SUBROUTINE expl_volc_so2_emissions_aerocom(kproma,kbdim,klev,klevp1,krow,paphp1,pxtte,pgeom1)

    !
    !    Volcanic background emissions 
    !
    !    volcanic_so2_emissions adopted for AEROCOM explosive volcanos emissions
    !
    !    Author: H. Feichter,    MI, August 1991, original source
    !            U. Schlese,     DKRZ, January 1995, changed
    !            L. Kornblueh,   MPI, May 1998, f90 rewrite
    !            U. Schulzweida, MPI, May 1998, f90 rewrite
    !            H.-S. Bauer and M. Schultz, MPI, November 2000, 
    !               (inclusion into new chemistry structure) 
    !            P. Stier,       MPI, 2001-2003
    !               (rewritten, removed resolution dependencies)
    
    USE mo_time_control,  ONLY: delta_time
    USE mo_aero_trac,     ONLY: idt_so2,                           &
                                idt_ms4ks,    idt_ms4as,           &
                                idt_nas   ,   idt_nks
    USE mo_aero_mem,      ONLY: emi_so2_evol,                                &
                                emi_so2_evol_alt_high, emi_so2_evol_alt_low ,&
                                emi_so2_evol_lev1,     emi_so2_evol_lev2,    &
                                d_emi_ms,                                    &
                                d_emi_so2_evol,        d_emi_so4_evol
    USE mo_constants,     ONLY: g
    USE mo_tracer,        ONLY: ntrac
    USE mo_exception,     ONLY: finish
    USE mo_aero,          ONLY: so2ts,facso2,               &
                                zm2n_s4ks_sk, zm2n_s4as_sa

    ! Scalar arguments
    INTEGER, INTENT(in) :: kproma, kbdim, klev, klevp1, krow

    ! Array arguments
    REAL(dp), INTENT(in)    :: pgeom1(kbdim,klev)
    REAL(dp), INTENT(in)    :: paphp1(kbdim,klevp1)
    REAL(dp), INTENT(inout) :: pxtte(kbdim,klev,ntrac)
    
    
    ! Local scalars
    REAL(dp) :: zgdp
    
    INTEGER :: jl, jk, ilev_high, ilev_low, inumlev

    REAL(dp) :: zheight(klev)

    !--- 0) Initialisations:
  
    DO jl=1,kproma

       zheight(:)=pgeom1(jl,:)/g

       IF (emi_so2_evol(jl,krow) > EPSILON(1.0_dp) ) THEN

          !--- Find corresponding model level to emission height high:
          
          IF (emi_so2_evol_alt_high(jl,krow) <= zheight(klev)) THEN 

            ilev_high=klev

          ELSE  
          
             DO jk=1, klev
                
                IF(emi_so2_evol_alt_high(jl,krow)<zheight(jk)) THEN
                   
                   CYCLE
                   
                ELSE IF(emi_so2_evol_alt_high(jl,krow)>=zheight(jk)) THEN                             
                   
                   ilev_high=jk
                   
                   emi_so2_evol_lev1(jl,krow)=REAL(ilev_high,dp)
                   
                   EXIT
                   
                END IF
                
             END DO
             
          END IF
          
          !--- Find corresponding model level to emission height low:

          IF (emi_so2_evol_alt_low(jl,krow) <= zheight(klev)) THEN 
             
             ilev_low=klev
             
          ELSE  
             
             DO jk=1,klev
                
                IF(emi_so2_evol_alt_low(jl,krow)<zheight(jk)) THEN
                   
                   CYCLE
                   
                ELSE IF(emi_so2_evol_alt_low(jl,krow)>=zheight(jk)) THEN
                
                   ilev_low=jk
                   
                   emi_so2_evol_lev2(jl,krow)=REAL(ilev_low,dp)
                   
                   EXIT
                   
                END IF
                
             END DO

          END IF

          !--- Emit the emissions between the emission level ilev_low and ilev_high:

          inumlev=ilev_low-ilev_high+1

          IF (inumlev <= 0 ) CALL finish('expl_volcanic_so2_emissions_aerocom','Error in emission heights')                     

          DO jk=ilev_high,ilev_low

          zgdp=g/(paphp1(jl,jk+1)-paphp1(jl,jk))

          pxtte(jl,jk,idt_so2)=pxtte(jl,jk,idt_so2)+             &
                               (1._dp/REAL(inumlev,dp))*emi_so2_evol(jl,krow)*so2ts*facso2*zgdp

          !--- Primary SO4 emissions

          pxtte(jl,jk,idt_ms4ks)=pxtte(jl,jk,idt_ms4ks) +  &
                                 0.5_dp*(1._dp/REAL(inumlev,dp))*emi_so2_evol(jl,krow)*so2ts*(1._dp-facso2)*zgdp
       
          pxtte(jl,jk,idt_nks)=pxtte(jl,jk,idt_nks) + &
                               0.5_dp*(1._dp/REAL(inumlev,dp))*emi_so2_evol(jl,krow)*so2ts*(1._dp-facso2)*zm2n_s4ks_sk*zgdp 

          !--- 50% Accumulation
       
          pxtte(jl,jk,idt_ms4as)=pxtte(jl,jk,idt_ms4as) +  &
                                 0.5_dp*(1._dp/REAL(inumlev,dp))*emi_so2_evol(jl,krow)*so2ts*(1._dp-facso2)*zgdp
       
          pxtte(jl,jk,idt_nas)=pxtte(jl,jk,idt_nas) + &
                               0.5_dp*(1._dp/REAL(inumlev,dp))*emi_so2_evol(jl,krow)*so2ts*(1._dp-facso2)*zm2n_s4as_sa*zgdp 

       END DO

       !--- Store volcanic emissions for diagnostics:

       d_emi_ms(jl,krow)=d_emi_ms(jl,krow)+emi_so2_evol(jl,krow)            *so2ts       *delta_time
       
       d_emi_so2_evol(jl,krow)=d_emi_so2_evol(jl,krow)+emi_so2_evol(jl,krow)*so2ts*facso2*delta_time
       
       d_emi_so4_evol(jl,krow)=d_emi_so4_evol(jl,krow)+emi_so2_evol(jl,krow)*so2ts*(1._dp-facso2)*delta_time

       END IF
    END DO

  END SUBROUTINE expl_volc_so2_emissions_aerocom
  

 SUBROUTINE so2_emissions_aerocom(kproma, kbdim, klev, klevp1, krow, paphp1, pxtems, pxtte)

    USE mo_tracer,        ONLY: ntrac
    USE mo_aero_trac,     ONLY: idt_so2,                         &
                                idt_ms4ks,    idt_ms4as ,        &
                                idt_ms4cs,                       &
                                idt_nas   ,    idt_nks ,         &
                                idt_ncs  
    USE mo_aero_mem,      ONLY: d_emi_ms,                        &
                                d_emi_so2_wf_3d, d_emi_so2_ant,  &
                                emi_so2_wf_3d,                   &
                                d_emi_so4_wf_3d,  d_emi_so4_ant, &
                                emi_so2_ant_low, emi_so2_ant_high
    USE mo_control,       ONLY: nlev
    USE mo_time_control,  ONLY: delta_time
    USE mo_constants,     ONLY: g
    USE mo_exception,     ONLY: finish
    USE mo_aero,          ONLY: so2ts,facso2,                    &             
                                zm2n_s4ks_sk, zm2n_s4as_sa ,     &
                                zm2n_s4cs_sc

    IMPLICIT NONE

    !--- Parameters:

    INTEGER, INTENT(IN)      :: kproma, kbdim, klev, klevp1, krow

    REAL(dp),    INTENT(in)      :: paphp1(kbdim,klevp1)
    REAL(dp),    INTENT(inout)   :: pxtems(kbdim,ntrac)
    REAL(dp),    INTENT(inout)   :: pxtte(kbdim,klev,ntrac)

    !--- Local Variables:

    REAL(dp)                     :: zgdp

    INTEGER                  :: jl, jk, ilev


    !--- 0) Initializations: ----------------------------------------------

!    IF (nlev/=19 .AND. nlev/=31) CALL finish('so2_emissions','missmatch with number of vertical levels')

 !uh for ma-echam:    

    IF (nlev/=19 .AND. nlev/=31 .AND. nlev/=39 .AND. nlev/=47) CALL finish('so2_emissions','missmatch with number of vertical levels')


  
    DO jl=1,kproma
       !--- 1) Surface emission      
       !       anthropogenic emissions lowest model layer : Domestic + road + off_road

       pxtems(jl,idt_so2)= pxtems(jl,idt_so2)+emi_so2_ant_low(jl,krow)*so2ts*facso2
       
       !--- primary SO4 emissions 

       !--- 50% Aitken

       pxtems(jl,idt_ms4ks)=pxtems(jl,idt_ms4ks) + &
                            0.5_dp*emi_so2_ant_low(jl,krow)*so2ts*(1._dp-facso2)

       
       pxtems(jl,idt_nks)  =pxtems(jl,idt_nks)+ &
                            0.5_dp*emi_so2_ant_low(jl,krow)*so2ts*(1._dp-facso2)* zm2n_s4ks_sk

       !--- 50% Accumultaion

       pxtems(jl,idt_ms4as)=pxtems(jl,idt_ms4as) + &
                            0.5_dp*emi_so2_ant_low(jl,krow)*so2ts*(1._dp-facso2)

       pxtems(jl,idt_nas)  =pxtems(jl,idt_nas)+ &
                            0.5_dp*emi_so2_ant_low(jl,krow)*so2ts*(1._dp-facso2)* zm2n_s4as_sa

       !--- Store for diagnostics:

       d_emi_so2_ant(jl,krow)= d_emi_so2_ant(jl,krow) + emi_so2_ant_low(jl,krow)*so2ts*facso2    *delta_time
       d_emi_so4_ant(jl,krow)= d_emi_so4_ant(jl,krow) + emi_so2_ant_low(jl,krow)*so2ts*(1._dp-facso2)*delta_time
       d_emi_ms(jl,krow)     = d_emi_ms(jl,krow)      + emi_so2_ant_low(jl,krow)*so2ts           *delta_time  
    
    END DO

    !--- 2) Emission of anthropogenic SO2 in the second lowest model level as this 
    !       corresponds to the next higher level above 100m 
    !       Note: While surface emissions are specified as fluxes, these
    !       emissions are applied as tracer tendencies in the 
    !       respective model layer industrial
    !    

    ilev=nlev-1

    DO jl=1, kproma

       zgdp=g/(paphp1(jl,ilev+1)-paphp1(jl,ilev))

       pxtte(jl,ilev,idt_so2)=pxtte(jl,ilev,idt_so2) +  &
                              emi_so2_ant_high(jl,krow)*so2ts*facso2*zgdp

       !--- Primary SO4 emissions

       !--- 50% Accumulation
       
       pxtte(jl,ilev,idt_ms4as)=pxtte(jl,ilev,idt_ms4as) +  &
                             0.5_dp*emi_so2_ant_high(jl,krow)*so2ts*(1._dp-facso2)*zgdp
       
       pxtte(jl,ilev,idt_nas)=pxtte(jl,ilev,idt_nas) + &
                               0.5_dp*emi_so2_ant_high(jl,krow)*so2ts*(1._dp-facso2)*zm2n_s4as_sa*zgdp 
       
       !--- 50% Coarse  

       pxtte(jl,ilev,idt_ms4cs)=pxtte(jl,ilev,idt_ms4cs) +  &
                             0.5_dp*emi_so2_ant_high(jl,krow)*so2ts*(1._dp-facso2)*zgdp
       
       pxtte(jl,ilev,idt_ncs)=pxtte(jl,ilev,idt_ncs) + &
                               0.5_dp*emi_so2_ant_high(jl,krow)*so2ts*(1._dp-facso2)*zm2n_s4cs_sc*zgdp 

       !--- Add emissions to total sulfur emissions:

       d_emi_ms(jl,krow)     = d_emi_ms(jl,krow)      + emi_so2_ant_high(jl,krow)*so2ts           *delta_time 
       d_emi_so2_ant(jl,krow)= d_emi_so2_ant(jl,krow) + emi_so2_ant_high(jl,krow)*so2ts*facso2    *delta_time
       d_emi_so4_ant(jl,krow)= d_emi_so4_ant(jl,krow) + emi_so2_ant_high(jl,krow)*so2ts*(1._dp-facso2)*delta_time

    END DO

    !--- 3) Biomass burning : Emissions are applied as tracer tendencies in the respective model
    !                         layer
    DO jk=1, klev
       DO jl=1, kproma

          zgdp=g/(paphp1(jl,jk+1)-paphp1(jl,jk))

          pxtte(jl,jk,idt_so2)=pxtte(jl,jk,idt_so2) +  &
                               emi_so2_wf_3d(jl,jk,krow)*so2ts*facso2*zgdp

          !--- Primary SO4 emissions

          !--- 50% Aitken
       
          pxtte(jl,jk,idt_ms4ks)=pxtte(jl,jk,idt_ms4ks) +  &
                             0.5_dp*emi_so2_wf_3d(jl,jk,krow)*so2ts*(1._dp-facso2)*zgdp
       
          pxtte(jl,jk,idt_nks)=pxtte(jl,jk,idt_nks) + &
                               0.5_dp*emi_so2_wf_3d(jl,jk,krow)*so2ts*(1._dp-facso2)*zm2n_s4ks_sk*zgdp 

          !--- 50% Accumulation
       
          pxtte(jl,jk,idt_ms4as)=pxtte(jl,jk,idt_ms4as) +  &
                             0.5_dp*emi_so2_wf_3d(jl,jk,krow)*so2ts*(1._dp-facso2)*zgdp
       
          pxtte(jl,jk,idt_nas)=pxtte(jl,jk,idt_nas) + &
                               0.5_dp*emi_so2_wf_3d(jl,jk,krow)*so2ts*(1._dp-facso2)*zm2n_s4as_sa*zgdp 
          
          !--- Store for diagnostics:
          
          d_emi_so2_wf_3d(jl,jk,krow) = d_emi_so2_wf_3d(jl,jk,krow) +&
                                        emi_so2_wf_3d(jl,jk,krow)*so2ts*facso2*delta_time

          d_emi_so4_wf_3d(jl,jk,krow) = d_emi_so4_wf_3d(jl,jk,krow) +&
                                        emi_so2_wf_3d(jl,jk,krow)*so2ts*(1._dp-facso2)*delta_time
          
          !--- Add emissions to total sulfur emissions

          d_emi_ms(jl,krow)     = d_emi_ms(jl,krow)      + emi_so2_wf_3d(jl,jk,krow)*so2ts           *delta_time 
          d_emi_so2_ant(jl,krow)= d_emi_so2_ant(jl,krow) + emi_so2_wf_3d(jl,jk,krow)*so2ts*facso2    *delta_time
          d_emi_so4_ant(jl,krow)= d_emi_so4_ant(jl,krow) + emi_so2_wf_3d(jl,jk,krow)*so2ts*(1._dp-facso2)*delta_time

       END DO

       
    END DO

  END SUBROUTINE so2_emissions_aerocom

  SUBROUTINE carbon_emissions_aerocom(kproma, kbdim, klev, klevp1, krow, paphp1, pxtems, pxtte)

    USE mo_tracer,        ONLY: ntrac
    USE mo_time_control,  ONLY: delta_time
    USE mo_constants,     ONLY: g     
    USE mo_aero,          ONLY: zm2n_bcki_ff, zm2n_bcki_bb, &
                                zm2n_ocki_ff,               &
                                zm2n_ocki_bb, zm2n_ocki_bg, &
                                zm2n_ocks_bb, zom2oc,       &
                                zbb_wsoc_perc,zbge_wsoc_perc
    USE mo_aero_trac,     ONLY: idt_mbcki,                         &
                                idt_mocki,    idt_mocks,           &
                                idt_mocas,                         &
                                idt_nki,      idt_nks     
    USE mo_aero_mem,      ONLY: d_emi_mbc,    d_emi_moc,           &
                                d_emi_bc_ant, d_emi_bc_wf_3d,      &
                                d_emi_bc_ag,  d_emi_oc_ant,        &
                                d_emi_oc_bge, d_emi_oc_wf_3d,      &
                                d_emi_oc_ag,                       &
                                emi_bc_ant,   emi_bc_wf_3d,        &
                                emi_bc_ag,    emi_oc_ant,          &
                                emi_oc_bge,   emi_oc_wf_3d,        &
                                emi_oc_ag

    IMPLICIT NONE

    !--- Parameters:
    
    INTEGER, INTENT(in)      :: kproma,klev,klevp1,kbdim,krow
    REAL(dp),    INTENT(in)      :: paphp1(kbdim,klevp1)
    REAL(dp),    INTENT(inout)   :: pxtems(kbdim,ntrac)
    REAL(dp),    INTENT(inout)   :: pxtte(kbdim,klev,ntrac)

    !--- Local Variables:

    INTEGER                 :: jl, jk

    REAL(dp)                    ::  zgdp


    !--- 1) Apply the emissions:

    DO jl=1, kproma

       !--- 1.1) AEROCOM no seasonality !

       !--- 1.2) Carbon mass flux:

       !--- Black Carbon:

       !    Anthropogenic:
 
       pxtems(jl,idt_mbcki)=pxtems(jl,idt_mbcki)+ &
                            emi_bc_ant(jl,krow)

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            emi_bc_ant(jl,krow) * zm2n_bcki_ff

       !    Add agricultural waste and fuelwood:

       pxtems(jl,idt_mbcki)=pxtems(jl,idt_mbcki)+ &
                            emi_bc_ag(jl,krow)

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            (emi_bc_ag(jl,krow)) * zm2n_bcki_bb

       !    Store emitted mass for diagnostics:

       d_emi_bc_ant(jl,krow)=d_emi_bc_ant(jl,krow)+ &
                             emi_bc_ant(jl,krow)*delta_time     
       d_emi_bc_ag(jl,krow) =d_emi_bc_ag(jl,krow) + &
                             emi_bc_ag(jl,krow)*delta_time
       d_emi_mbc(jl,krow)   =d_emi_mbc(jl,krow)   + &
                             (emi_bc_ant(jl,krow)+emi_bc_ag(jl,krow))*delta_time        
       
       !--- Organic Carbon:

       !    Anthropogenic:
 
       pxtems(jl,idt_mocki)=emi_oc_ant(jl,krow)*zom2oc

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            emi_oc_ant(jl,krow)*zom2oc * zm2n_ocki_ff

       !    Add  agricultural waste and fuelwood - insoluble fraction:

       pxtems(jl,idt_mocki)=pxtems(jl,idt_mocki)+ &
                            (emi_oc_ag(jl,krow))*zom2oc*(1._dp-zbb_wsoc_perc)

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            (emi_oc_ag(jl,krow))*zom2oc*(1._dp-zbb_wsoc_perc)*zm2n_ocki_bb

       !    Add  agricultural waste and fuelwood - soluble fraction:

       pxtems(jl,idt_mocks)=pxtems(jl,idt_mocks)+ &
                            (emi_oc_ag(jl,krow))*zom2oc*zbb_wsoc_perc

       pxtems(jl,idt_nks)  =pxtems(jl,idt_nks)+ &
                            (emi_oc_ag(jl,krow))*zom2oc*zbb_wsoc_perc * zm2n_ocks_bb

       !    Add Biogenic - insoluble fraction:

       pxtems(jl,idt_mocki)=pxtems(jl,idt_mocki)+ &
                            emi_oc_bge(jl,krow)*(1._dp-zbge_wsoc_perc)

       pxtems(jl,idt_nki)  =pxtems(jl,idt_nki)+ &
                            emi_oc_bge(jl,krow)*(1._dp-zbge_wsoc_perc) * zm2n_ocki_bg

       !    Add Biogenic - soluble fraction:

       !--- 50% Aitken mode

       pxtems(jl,idt_mocks)=pxtems(jl,idt_mocks)+ &
                            0.5_dp*emi_oc_bge(jl,krow)*zbge_wsoc_perc

       !--- 50% Accumulation mode

       pxtems(jl,idt_mocas)=pxtems(jl,idt_mocas)+ &
                            0.5_dp*emi_oc_bge(jl,krow)*zbge_wsoc_perc

       !    Store emitted mass for diagnostics:

       d_emi_oc_ant(jl,krow)=d_emi_oc_ant(jl,krow)+ &
                             emi_oc_ant(jl,krow)*zom2oc             * delta_time    
       d_emi_oc_ag(jl,krow) =d_emi_oc_ag(jl,krow) + &
                             emi_oc_ag(jl,krow)*zom2oc              * delta_time
       d_emi_oc_bge(jl,krow)=d_emi_oc_bge(jl,krow)+ &
                             emi_oc_bge(jl,krow)                    * delta_time
       d_emi_moc(jl,krow)   =d_emi_moc(jl,krow)   + &
                             ( emi_oc_ant(jl,krow)*zom2oc + emi_oc_ag(jl,krow)*zom2oc + emi_oc_bge(jl,krow)  ) * delta_time

    END DO

    ! BIOMASS BURNING EMISSIONS
    
    DO jk=1, klev
       DO jl=1, kproma

          zgdp = g/(paphp1(jl,jk+1)-paphp1(jl,jk))
          
          ! BLACK CARBON
          
          pxtte(jl,jk,idt_mbcki)=pxtte(jl,jk,idt_mbcki) +  &
                                 emi_bc_wf_3d(jl,jk,krow)*zgdp
          
          pxtte(jl,jk,idt_nki)=pxtte(jl,jk,idt_nki) + &
                               emi_bc_wf_3d(jl,jk,krow) * zm2n_bcki_bb * zgdp
          
          !--- store for diagnostic
          d_emi_bc_wf_3d(jl,jk,krow)=d_emi_bc_wf_3d(jl,jk,krow) + &
                                     emi_bc_wf_3d(jl,jk,krow)*delta_time
          
          !--- update total mass
          d_emi_mbc(jl,krow)     =d_emi_mbc(jl,krow)   + &
                                     emi_bc_wf_3d(jl,jk,krow)*delta_time          
          
          ! ORGANIC CARBON
          
          !--- organic carbon insoluble fraction !!!

          pxtte(jl,jk,idt_mocki)=pxtte(jl,jk,idt_mocki) + &
                                 emi_oc_wf_3d(jl,jk,krow)*(1.-zbge_wsoc_perc) * zgdp

          
          pxtte(jl,jk,idt_nki) =pxtte(jl,jk,idt_nki)   + &
                                emi_oc_wf_3d(jl,jk,krow)*(1.-zbb_wsoc_perc)*zm2n_ocki_bb * zgdp
          
          !--- organic carbon soluble fraction !!!

          pxtte(jl,jk,idt_mocks)=pxtte(jl,jk,idt_mocks) + &
                                emi_oc_wf_3d(jl,jk,krow)*zbb_wsoc_perc * zgdp

          pxtte(jl,jk,idt_nks) =pxtte(jl,jk,idt_nks)   + &
                                emi_oc_wf_3d(jl,jk,krow)*zbb_wsoc_perc*zm2n_ocks_bb * zgdp
          
          !--- diagnostic !!
          
          d_emi_oc_wf_3d(jl,jk,krow) =d_emi_oc_wf_3d(jl,jk,krow) + &
                                      emi_oc_wf_3d(jl,jk,krow)            * delta_time
          
          !--- update total mass!!!

          d_emi_moc(jl,krow)   =d_emi_moc(jl,krow)   + &
                                emi_oc_wf_3d(jl,jk,krow)    * delta_time
       END DO
    END DO
  END SUBROUTINE carbon_emissions_aerocom
 
END MODULE mo_chem_bcond_aerocom
