MODULE mo_mvstream   

  !--------------------------------------------------------------------------
  !
  ! Module that allows to open an extra redoubled stream for some tracers
  ! of elements of any other stream (3d or 2d grid point fields) 
  ! in order to calculate for example the mean values over a certain interval
  ! -----------------------------------------
  !
  ! Authors:
  !  J.S. Rast, MPI-Met, Hamburg, December 2003, original source
  !  P. Stier,  MPI-Met, Hamburg,          2004, modifications
  !  J.S. Rast, MPI-Met, Hamburg, June     2004, optimisation
  !--------------------------------------------------------------------------

  USE mo_kind,                  ONLY: dp
  USE mo_time_event,            ONLY: io_time_event
  USE mo_mpi,                   ONLY: p_parallel_io
  USE mo_memory_base,           ONLY: maxstr
  USE mo_doctor,                ONLY: nout

  IMPLICIT NONE
  PRIVATE

  PUBLIC                           :: init_mvstream, acc_mvstream

  INTEGER, PARAMETER               :: nam_max=500
  LOGICAL                          :: lvarmean(nam_max, maxstr), &
                                      lvarsurf(nam_max, maxstr), &
                                      lstddev(nam_max, maxstr)
  CHARACTER(len=256)               :: m_stream_name(maxstr)
  CHARACTER(len=32)                :: mvarname(nam_max, maxstr)
  INTEGER                          :: nmstream,nmvar(maxstr),nlist_elements(maxstr)
  TYPE(io_time_event), SAVE        :: putmean_stream(maxstr)

  TYPE vptr
     REAL(dp), POINTER             :: ptr3d(:,:,:)
     REAL(dp), POINTER             :: ptr3d_s(:,:,:)
     REAL(dp), POINTER             :: ptr2d(:,:)
     REAL(dp), POINTER             :: ptr2d_s(:,:)
  END TYPE vptr

  TYPE tptr
     REAL(dp), POINTER             :: ptt3d(:,:,:)
     REAL(dp), POINTER             :: ptt2d(:,:)
   END TYPE tptr

  TYPE(vptr)                       :: varptr(nam_max, maxstr)

  TYPE(tptr)                       :: tarptr(nam_max, maxstr)

  CONTAINS

  SUBROUTINE init_mvstream
    !------------------------------------------------------------------------
    ! initialize mean value stream
    !
    ! Authors:
    ! J.S. Rast, MPI-Met, Hamburg, December 2003, original source
    !------------------------------------------------------------------------
  USE mo_tracdef,             ONLY: trlist, ln, ll
  USE mo_tracer,              ONLY: get_tracer
  USE mo_filename,            ONLY: find_next_free_unit
  USE mo_namelist,            ONLY: open_nml, position_nml, nnml, POSITIONED
  USE mo_exception,           ONLY: finish, message
  USE mo_mpi,                 ONLY: p_bcast, p_parallel, p_io
  USE mo_time_control,        ONLY: p_bcast_event
  USE mo_linked_list,         ONLY: t_stream, memory_info, GAUSSIAN, UNKNOWN, &
                                    list_element
  USE mo_memory_base,         ONLY: new_stream, get_stream, &
                                    add_stream_element, &
                                    default_stream_setting, &
                                    get_stream_element_info, &
                                    get_stream_element, AUTO

  LOGICAL                        :: lmean(nam_max), lsurf(nam_max), &
                                    lstd_all, lmean_all
  CHARACTER(len=256)             :: ierrchar
  CHARACTER(len=32)              :: meannam(nam_max)
  INTEGER                        :: ierr, kstream, knam, kdx, nunit
  INTEGER                        :: istd(nam_max), stddev(nam_max), ilist_cnt
  TYPE(memory_info)              :: se_info
  TYPE(t_stream), POINTER        :: mv, stream
  TYPE(io_time_event), SAVE      :: putmean
  TYPE (list_element) ,POINTER   :: this_list_element

  INCLUDE 'mvctl.inc'
  INCLUDE 'mvstreamctl.inc'

  IF (p_parallel_io) THEN
! read stream names for the variables of which mean values may be required

     m_stream_name(:)=''

     CALL position_nml ('MVSTREAMCTL', status=ierr)
     SELECT CASE (ierr)
        CASE (POSITIONED)
           READ(nnml, mvstreamctl)
     END SELECT
     nmstream = 0
     DO kstream = 1, maxstr  ! if number of streams greater than maxstr, error message by newstream
        IF ( m_stream_name(kstream) /= '' ) THEN
           IF(LEN_TRIM(m_stream_name(kstream)) > 15) THEN
              CALL finish('mo_mvstream   , init_mvstream', 'stream name '// &
                   TRIM(m_stream_name(kstream))//' exceeds 15' // &
                   ' alphanumeric characters, cannot define mean value stream name.')
           END IF
           IF ( m_stream_name(kstream) /= 'tracer' ) THEN
! proof that stream exists
              CALL get_stream ( mv, TRIM(m_stream_name(kstream)) )
              IF (.NOT.ASSOCIATED(mv)) THEN
                 CALL finish ('mo_mvstream   :', &
                 'stream '//TRIM(m_stream_name(kstream))//' does not exist')
              END IF
! proof that stream contains <= nam_max variables
              IF (mv%list_elements > nam_max) THEN 
                 WRITE(ierrchar,*) mv%list_elements
                 CALL finish ('mo_mvstream   , init_mvstream ', &
                      'number of stream elements of stream '// &
                      TRIM(m_stream_name(kstream))// &
                      'exceeds '//TRIM(ierrchar)// &
                      ', may not get all mean value variables. '// &
                      'Please, increase nam_max in mo_mvstream   .')
              END IF
           ELSE
! proof that number of tracers <= nam_max
              IF (trlist%ntrac > nam_max) THEN
                 WRITE(ierrchar,*) trlist%ntrac
                 CALL finish ('mo_mvstream   , init_mvstream ', &
                      'number of tracer exceeds '//TRIM(ierrchar)// &
                      ', may not get all mean value tracers. '// &
                      'Please, increase nam_max in mo_mvstream   .')
              END IF
           END IF
        ELSE
           EXIT   
        END IF
        nmstream = nmstream + 1
     END DO
     WRITE (nout,*)
     WRITE (nout,*) '******************************************************'
     WRITE (nout,*) 'mean values calculated for following variables:       '
     WRITE (nout,*) '******************************************************'
     nunit = find_next_free_unit(10,64)
     mvarname(:,:)      = ''
     DO kstream = 1, nmstream
! set default values
        putmean%counter    = 1
        putmean%unit       = 'months'
        putmean%adjustment = 'first'
        putmean%offset     = 0
        meannam(:)         = ''
        lmean(:)           = .FALSE.
        lsurf(:)           = .FALSE.
        stddev(:)          = 0
        istd(:)            = 0
        nmvar(kstream)     = 0
        lmean_all          = .FALSE.
        lstd_all           = .FALSE.
! read time intervals and variable names for each stream
        OPEN (nunit, file=TRIM(m_stream_name(kstream))//'.nml', &
             iostat=ierr, status='old', action='read')
        IF (ierr /= 0) THEN
           CALL finish ('mo_mvstream   , inimvstream','Could not open '// &
                TRIM(m_stream_name(kstream))//'.nml')
        END IF
        CALL position_nml ('MVCTL', nunit, status=ierr)
        SELECT CASE (ierr)
        CASE (POSITIONED)
           READ(nunit, mvctl)
        END SELECT
        CLOSE (nunit)
! save putmean interval
        putmean_stream(kstream) = putmean
! count variables for each stream and assure that variables exist
        IF (TRIM(m_stream_name(kstream)) == 'tracer') THEN
! all variables in tracer stream
           IF (meannam(1) == 'all' .OR. meannam(1) == 'ALL' ) THEN
              lmean_all=.TRUE.
           END IF
! mean of square for all variables in tracer stream
           IF (stddev(1) == -1) THEN
              lstd_all=.TRUE.
           END IF
           IF (lmean_all) THEN
              mvarname(1:trlist%ntrac,kstream)=trlist%ti(1:trlist%ntrac)%fullname
              lmean(1:trlist%ntrac)  = .TRUE.
              nmvar(kstream)         = trlist%ntrac
              IF (lstd_all) THEN
                 istd(1:trlist%ntrac)= 1
              ELSE
                 istd(1:trlist%ntrac)= stddev(1:trlist%ntrac)
              END IF
           ELSE  ! individual tracers
              ilist_cnt=0
              DO knam = 1, trlist%ntrac
                 IF (meannam(knam) /= '') THEN
                    ilist_cnt=ilist_cnt+1
                    CALL get_tracer (meannam(knam), idx=kdx, ierr = ierr)
                    IF (ierr == 1)THEN
                       CALL finish ('mo_mvstream, init_mvstream','tracer ' &
                            //TRIM(meannam(knam))//' in stream '//TRIM(m_stream_name(kstream))// &
                            ' is not defined')
                    ELSE
                       lmean(kdx) = .TRUE.
                       mvarname(kdx,kstream) = meannam(knam)
                       nmvar(kstream) = nmvar(kstream) + 1
                       IF (lstd_all) THEN
                          istd(kdx) = 1
                       ELSE
                          istd(kdx) = stddev(nmvar(kstream))
                       END IF  !lstd_all
                    END IF  ! tracer defined
                 ELSE  ! no tracer name given, so meannam(knam) = ''
                    EXIT
                 ENDIF
              END DO  ! loop over tracers
           END IF  ! lmean_all 
        ELSE 
! a general stream, not the tracer stream
           CALL get_stream ( mv, TRIM(m_stream_name(kstream)) )
           IF (TRIM(meannam(1)) == 'all' .OR. TRIM(meannam(1)) == 'ALL' ) THEN
              lmean_all = .TRUE.
           END IF
           IF (stddev(1) == -1) THEN
              lstd_all=.TRUE.
           END IF
           IF (lmean_all) THEN
              IF (lstd_all) THEN
                 istd(1:mv%list_elements) = 1
              ELSE
                 istd(1:mv%list_elements) = stddev(1:mv%list_elements)
              END IF
              this_list_element => mv%first_list_element
              DO knam = 1, mv%list_elements
                 meannam(knam) = TRIM(this_list_element%field%info%name)
                 CALL get_stream_element_info (mv, TRIM(meannam(knam)), se_info)
                 ! allow accumulation only for 3d or 2d gaussian grid point fields
                 IF (se_info%ndim == UNKNOWN) THEN
                    CALL finish ('mo_mvstream, init_mvstream','stream element ' &
                         //TRIM(meannam(knam))//' in stream '//TRIM(m_stream_name(kstream))// &
                         ' is not defined')
                 END IF
                 IF (se_info%repr == GAUSSIAN .AND. .NOT. se_info%laccu) THEN
                    lmean(knam) = .TRUE.
                    mvarname(knam,kstream) = meannam(knam)
                    nmvar(kstream) = nmvar(kstream) + 1
                    lsurf(knam) = se_info%ndim == 2
                 ELSE  ! not a gaussian grid variable or variable has accumulation flag
                    istd(knam) = 0
                    IF (se_info%repr /= GAUSSIAN ) THEN
                       CALL message ('mo_mvstream, init_mvstream','variable ' &
                            //meannam(knam)//' in stream '//TRIM(m_stream_name(kstream))// &
                            ': not gaussian => skipped')
                    ELSE IF (se_info%laccu) THEN
                       CALL message ('mo_mvstream, init_mvstream','variable ' &
                            //meannam(knam)//' in stream '//TRIM(m_stream_name(kstream))// &
                            ': laccu=.TRUE. => skipped')
                    END IF
                 END IF  ! variable is Gaussian and does not have accumulation flag
                 this_list_element => this_list_element%next_list_element
              END DO  ! loop over all stream elements
           ELSE  ! individual variables in stream
              ilist_cnt=0
              DO knam = 1, mv%list_elements
                 IF (meannam(knam) /= '') THEN
                    ilist_cnt=ilist_cnt+1
                    CALL get_stream_element_info (mv, TRIM(meannam(knam)), se_info)
! allow accumulation only for 3d or 2d gaussian grid point fields
                    IF (se_info%ndim == UNKNOWN) THEN
                       CALL finish ('mo_mvstream, init_mvstream','stream element ' &
                            //TRIM(meannam(knam))//' in stream '//TRIM(m_stream_name(kstream))// &
                            ' is not defined')
                    END IF
                    IF (se_info%repr == GAUSSIAN .AND. .NOT. se_info%laccu) THEN
                       mvarname(knam,kstream) = meannam(knam)
                       lmean(knam) = .TRUE.
                       lsurf(knam) = se_info%ndim == 2
                       nmvar(kstream) = nmvar(kstream) + 1
                       IF (lstd_all) THEN
                          istd(knam) = 1
                       ELSE
                          istd(knam) = stddev(ilist_cnt)
                       END IF
                    ELSE  ! not a gaussian grid variable or variable has accumulation flag
                       IF (se_info%repr /= GAUSSIAN ) THEN
                          CALL message ('mo_mvstream, init_mvstream','variable ' &
                               //meannam(knam)//' in stream '//TRIM(m_stream_name(kstream))// &
                               ': not gaussian => skipped')
                       ELSE IF(se_info%laccu) THEN
                          CALL message ('mo_mvstream, init_mvstream','variable ' &
                               //meannam(knam)//' in stream '//TRIM(m_stream_name(kstream))// &
                               ': laccu=.TRUE. => skipped')
                       END IF
                    END IF  ! gaussian
                 END IF  ! meannam(knam) /= '' 
              END DO  ! loop over stream elements
           END IF  ! option "all" in meannam
        END IF  ! tracer stream or other stream
! table of variables that require mean value calculation
        WRITE (nout,*)
        WRITE (nout,*) '            stream ',TRIM(m_stream_name(kstream)),':'
        WRITE (nout,*) (TRIM(meannam(knam)),' ',knam=1,nmvar(kstream))
        WRITE (nout,*) (stddev(knam),'  ',knam=1,nmvar(kstream))

        lvarmean(:, kstream) = lmean(:)
        lvarsurf(:, kstream) = lsurf(:)
        lstddev(:, kstream) = istd(:) == 1
     END DO ! kstream
!debug
     WRITE(0,*)
     WRITE(0,*) '******************** begin debug output mo_mvstream ***********************'
     WRITE(0,*) 'mo_mvstream   , init_mvstream:'
     WRITE(0,*) 'number of streams, nmstream=',nmstream
     WRITE(0,*) 'names of streams, m_stream_name=', &
          (TRIM(m_stream_name(kstream)),' ',kstream=1,nmstream)
     WRITE(0,*) 'number of variables in each stream:',nmvar(1:nmstream)
     WRITE(0,*) 'variables in each stream:'
     WRITE(0,*) ((TRIM(mvarname(knam,kstream)),' ',knam=1,nam_max),kstream=1,nmstream)
     WRITE(0,*) 'mean value flag in each stream', &
          ((lvarmean(knam,kstream),' ',knam=1,nam_max),kstream=1,nmstream)
     WRITE(0,*) '2-d flag in each stream', &
          ((lvarsurf(knam,kstream),' ',knam=1,nam_max),kstream=1,nmstream)
     WRITE(0,*) 'stddev flag in each stream', &
          ((lstddev(knam,kstream),' ',knam=1,nam_max),kstream=1,nmstream)
     WRITE(0,*) 'put mean interval for each stream', putmean_stream(1:nmstream)
     WRITE(0,*) '******************** end debug output mo_mvstream *************************'
     WRITE(0,*)
!debug.
  END IF ! (p_parallel_io)
! send information on all processors
  IF (p_parallel) THEN
     CALL p_bcast(nmstream, p_io)
     CALL p_bcast(m_stream_name, p_io)
     CALL p_bcast(nmvar, p_io)
     DO kstream = 1, nmstream
        CALL p_bcast(mvarname(:, kstream), p_io)
        CALL p_bcast(lvarmean(:, kstream), p_io)
        CALL p_bcast(lvarsurf(:, kstream), p_io)
        CALL p_bcast(lstddev(:, kstream), p_io)
        CALL p_bcast_event(putmean_stream(kstream), p_io)
     END DO
  END IF
! open new streams
  DO kstream = 1, nmstream
     IF (nmvar(kstream) /= 0) THEN
        CALL new_stream (mv,TRIM(m_stream_name(kstream))//'m', &
             interval=putmean_stream(kstream))
        CALL default_stream_setting (mv, lrerun = .TRUE., contnorest = .TRUE., &
             lpost = .TRUE., laccu = .TRUE., code = AUTO)
        IF (TRIM(m_stream_name(kstream)) == 'tracer') THEN
           DO knam = 1, trlist%ntrac
              IF (lvarmean(knam,kstream)) THEN
                 CALL add_stream_element (mv, trlist%ti(knam)%fullname, varptr(knam,kstream)%ptr3d, &
                      units = trlist%ti(knam)%units, longname = trlist%ti(knam)%longname)
              END IF  ! lvarmean is true
              IF (lstddev(knam,kstream)) THEN
                 IF (LEN_TRIM(trlist%ti(knam)%fullname) > ln - 2) THEN
                    write(ierrchar,*) ln
                    CALL finish('mo_mvstream   , init_mvstream', 'tracer name '// &
                         TRIM(trlist%ti(knam)%fullname)//' exceeds '//ierrchar// &
                         '-2 alphanumeric characters, cannot define mean of squares tracer name.')
                 END IF
                 IF (LEN_TRIM(trlist%ti(knam)%units) > ln - 5) THEN
                    write(ierrchar,*) ln
                    CALL finish('mo_mvstream   , init_mvstream', 'unit name '// &
                         TRIM(trlist%ti(knam)%units)//' of tracer '//TRIM(trlist%ti(knam)%fullname)// &
                         ' exceeds '//ierrchar// &
                         '-5 alphanumeric characters, cannot define mean of squares variable units.')
                 END IF
                 IF (LEN_TRIM(trlist%ti(knam)%longname) > ll - 10) THEN
                    write(ierrchar,*) ll
                    CALL message('mo_mvstream, init_mvstream', 'longname '// &
                         TRIM(trlist%ti(knam)%longname)//' of tracer ' &
                         //TRIM(trlist%ti(knam)%fullname)//' exceeds '//ierrchar// &
                      '-10 alphanumeric characters, cannot define longname of square.')
                 END IF
                 CALL add_stream_element (mv, TRIM(trlist%ti(knam)%fullname)//'_s', varptr(knam,kstream)%ptr3d_s, &
                      units = '('//TRIM(trlist%ti(knam)%units)//')**2', &
                      longname = 'square of '//TRIM(trlist%ti(knam)%longname))
              END IF  ! calculation of mean of square
           END DO  ! tracer list
        ELSE  ! general stream
           CALL get_stream (stream, TRIM(m_stream_name(kstream)))
           nlist_elements(kstream)=stream%list_elements
           DO knam = 1, nlist_elements(kstream)
              IF (lvarmean(knam, kstream)) THEN
                 CALL get_stream_element_info (stream, TRIM(mvarname(knam, kstream)), &
                      se_info)
                 IF (se_info%ndim == 2) THEN
                    CALL add_stream_element (mv, TRIM(se_info%name), varptr(knam,kstream)%ptr2d, &
                         units = se_info%units, longname = se_info%longname)
                    CALL get_stream_element (stream, TRIM(se_info%name), tarptr(knam,kstream)%ptt2d)
                 ELSE
                    CALL add_stream_element (mv, TRIM(se_info%name), varptr(knam,kstream)%ptr3d, &
                         units = se_info%units, longname = se_info%longname)
                    CALL get_stream_element (stream, TRIM(se_info%name), tarptr(knam,kstream)%ptt3d)
                 END IF
              END IF  ! lvarmean
              IF (lstddev(knam,kstream)) THEN
                 CALL get_stream_element_info (stream, TRIM(mvarname(knam, kstream)), &
                      se_info)
                 IF(LEN_TRIM(se_info%name) > 30) THEN
                    CALL finish('mo_mvstream   , init_mvstream', 'variable name '// &
                         TRIM(se_info%name)//' exceeds 30' // &
                         ' alphanumeric characters, cannot define mean of squares variable name.')
                 END IF
                 IF (LEN_TRIM(se_info%units) > 59) THEN
                    CALL finish('mo_mvstream   , init_mvstream', 'unit name '// &
                         TRIM(se_info%units)//' of variable '//TRIM(se_info%name)// &
                         ' exceeds 59' // &
                         'alphanumeric characters, cannot define mean of squares variable units.')
                 END IF
                 IF (LEN_TRIM(se_info%longname) > 118) THEN
                    CALL finish('mo_mvstream   , init_mvstream', 'longname '// &
                         TRIM(se_info%longname)//' of variable '//TRIM(se_info%name)// &
                         ' exceeds 118' // &
                         'alphanumeric characters, cannot define longname of square.')
                 END IF
                 IF (se_info%ndim == 2) THEN
                    CALL add_stream_element (mv, TRIM(se_info%name)//'_s', varptr(knam,kstream)%ptr2d_s, &
                         units = '('//TRIM(se_info%units)//')**2', &
                         longname = 'square of '//TRIM(se_info%longname))
                 ELSE
                    CALL add_stream_element (mv, TRIM(se_info%name)//'_s', varptr(knam,kstream)%ptr3d_s, &
                         units = '('//TRIM(se_info%units)//')**2', &
                         longname = 'square of '//TRIM(se_info%longname))
                 END IF  
              END IF  ! lstddev
           END DO  ! loop over stream elements
        ENDIF ! tracer stream
     END IF  ! there are elements in stream for which calculation is required
  END DO  ! kstream
END SUBROUTINE init_mvstream

SUBROUTINE acc_mvstream (krow,ltrac)
  !------------------------------------------------------------------------
  ! accumulate stream variables
  !
  ! Authors:
  ! J.S. Rast, MPI-Met, Hamburg, December 2003, original source
  !------------------------------------------------------------------------
  USE mo_tracdef,             ONLY: trlist
  USE mo_memory_gl,           ONLY: xt
  USE mo_linked_list,         ONLY: t_stream, memory_info, GAUSSIAN
  USE mo_time_control,        ONLY: delta_time
  USE mo_decomposition,       ONLY: ldc => local_decomposition

  INTEGER, INTENT(IN)            :: krow
  LOGICAL, INTENT(IN)            :: ltrac
  INTEGER                        :: kstream, kvar, jrow, kproma, klev
  REAL(dp), POINTER              :: zstvar3d(:,:,:), zstvar2d(:,:)
  TYPE(t_stream), POINTER        :: mv, stream
  TYPE(memory_info)              :: se_info


  IF (krow == ldc%ngpblks) THEN
     kproma = ldc%npromz
  ELSE
     kproma = ldc%nproma
  END IF
  klev = ldc%nlev
  IF (ltrac) THEN
     DO kstream = 1, nmstream
        IF (m_stream_name(kstream) == 'tracer') THEN
           IF (ANY(lvarmean(:, kstream))) THEN
              DO kvar = 1, trlist%ntrac
                 IF (lvarmean(kvar, kstream)) THEN
                    varptr(kvar,kstream)%ptr3d(1:kproma, 1:klev, krow) = &
                         varptr(kvar,kstream)%ptr3d(1:kproma, 1:klev, krow) + &
                         xt(1:kproma, 1:klev, kvar, krow) * delta_time
                    IF (lstddev(kvar,kstream)) THEN
                       varptr(kvar,kstream)%ptr3d_s(1:kproma, 1:klev, krow) = &
                            varptr(kvar,kstream)%ptr3d_s(1:kproma, 1:klev, krow) + &
                            xt(1:kproma, 1:klev, kvar, krow) * &
                            xt(1:kproma, 1:klev, kvar, krow) * delta_time
                    END IF
                 END IF
              END DO
           END IF ! ANY(lvarmean)
        END IF ! tracer stream
     END DO ! streams
  ELSE
     DO kstream = 1, nmstream
        IF (m_stream_name(kstream) /= 'tracer') THEN
           IF (ANY(lvarmean(1:nmvar(kstream), kstream))) THEN
              DO kvar = 1, nlist_elements(kstream)
                 IF (lvarmean(kvar, kstream)) THEN
                    IF (lvarsurf(kvar, kstream)) THEN
                       varptr(kvar,kstream)%ptr2d(1:kproma, krow) = &
                            varptr(kvar,kstream)%ptr2d(1:kproma, krow) + &
                            tarptr(kvar,kstream)%ptt2d(1:kproma, krow) * delta_time
                       IF (lstddev(kvar,kstream)) THEN
                          varptr(kvar,kstream)%ptr2d_s(1:kproma, krow) = &
                               varptr(kvar,kstream)%ptr2d_s(1:kproma, krow) + &
                               tarptr(kvar,kstream)%ptt2d(1:kproma, krow) * &
                               tarptr(kvar,kstream)%ptt2d(1:kproma, krow) * delta_time
                       END IF
                    ELSE
                       varptr(kvar,kstream)%ptr3d(1:kproma, 1:klev, krow) = &
                            varptr(kvar,kstream)%ptr3d(1:kproma, 1:klev, krow) + &
                       tarptr(kvar,kstream)%ptt3d(1:kproma, 1:klev, krow) * delta_time
                       IF (lstddev(kvar,kstream)) THEN
                          varptr(kvar,kstream)%ptr3d_s(1:kproma, 1:klev, krow) = &
                               varptr(kvar,kstream)%ptr3d_s(1:kproma, 1:klev, krow) + &
                                tarptr(kvar,kstream)%ptt3d(1:kproma, 1:klev, krow) * &
                                tarptr(kvar,kstream)%ptt3d(1:kproma, 1:klev, krow) * delta_time
                       END IF
                    END IF !lvarsurf
                 END IF !lvarmean
              END DO
           END IF ! ANY(lvarmean) 
        END IF ! tracer stream
     END DO ! kstream
  END IF !ltrac
END SUBROUTINE acc_mvstream

END MODULE mo_mvstream   
