MODULE mo_sub_echam
!-----------------------------------------------------------------------
! This module holds the simple submodule provided by echam. It allows to
! request tracers via the namelist and provides routines for commonly
! used processes. All these routines are called from routines within
! call_submodels.f90
!
! Authors: 
! A.Rhodin MPI/DWD April 2002 original code
!------------------------------------------------------------------------
  !-------------
  ! Modules used
  !-------------
  USE mo_kind,   ONLY: dp
  USE mo_tracer, ONLY: trlist
! uh:
    USE mo_be10,          ONLY: be7_decay, idt_be7, idt_pb210, idt_rn222, pb210conc, &
         idt_na22, idt_cl36, rn222_decay!, test
!uh-
  IMPLICIT NONE
  !----------------
  ! Public entities
  !----------------
  PUBLIC :: radionucl_sink       ! apply exponential decay
  PRIVATE

CONTAINS
!==============================================================================
  !==========
  ! Processes
  !==========
!------------------------------------------------------------------------------
  SUBROUTINE radionucl_sink (kproma, kbdim, klev, krow, pxtm1, pxtte)
    !------------------------------------------------------------
    ! Description:
    !
    ! Calculates the decrease of tracer concentration for a given 
    ! exponential decay time
    !
    ! Method:
    !
    ! The mass mixing-ratio of tracers is multiplied with
    ! exp(time-step/decay-time).
    ! This routine could also be used for emission or sink
    ! above the surface.
    !
    ! *radionucl_sink* is called from *physc*
    !
    ! Authors:
    !
    ! J. Feichter, MI, August 1991, original source
    ! L. Kornblueh, MPI, May 1998, f90 rewrite
    ! U. Schulzweida, MPI, May 1998, f90 rewrite
    !------------------------------------------------------------
    USE mo_time_control,  ONLY: time_step_len
    !----------
    ! Arguments
    !----------
    INTEGER ,INTENT(in)    :: kproma, kbdim, klev, krow             ! dimensions
    REAL(dp),INTENT(in)    :: pxtm1(kbdim,klev,trlist% ntrac) ! concentr. t-dt
    REAL(dp),INTENT(inout) :: pxtte(kbdim,klev,trlist% ntrac) ! tendency
    !----------------
    ! Local variables
    !----------------
    INTEGER :: jt                 ! tracer index
    REAL(dp):: zxtp1(kproma,klev) ! tracer concentration at t+dt
    REAL(dp):: zxtte(kproma,klev) ! tracer tendency
    REAL(dp):: ztmst              ! time step
    REAL(dp):: zqtmst             ! 1 / time step
    !----------------------
    ! Executable statements
    !----------------------
    ztmst  = time_step_len
    zqtmst = 1.0_dp/ztmst

    DO jt = 1, trlist% ntrac
       IF (trlist% ti(jt)% tdecay /= 0._dp) THEN
!uh+
         IF(jt==idt_be7.or.jt==idt_rn222) zxtte(1:kproma,:)=pxtte(1:kproma,:,jt)
          zxtp1(:,:)           = pxtm1(1:kproma,:,jt) + pxtte(1:kproma,:,jt) * ztmst

          zxtp1(:,:)           = EXP (-ztmst/trlist% ti(jt)% tdecay) * zxtp1(:,:)
          pxtte(1:kproma,:,jt) = (zxtp1(:,:)-pxtm1(1:kproma,:,jt))*zqtmst
          IF(jt==idt_rn222) THEN
             pb210conc(1:kproma,:,krow)=-(pxtte(1:kproma,:,jt)-zxtte(1:kproma,:))
             pxtte(1:kproma,:,idt_pb210)=pxtte(1:kproma,:,idt_pb210)+pb210conc(1:kproma,:,krow)
             rn222_decay(1:kproma,:,krow)=pxtte(1:kproma,:,jt)-zxtte(1:kproma,:)
          END IF
          IF(jt==idt_be7) be7_decay(1:kproma,:,krow)=pxtte(1:kproma,:,jt)-zxtte(1:kproma,:)


!          zxtp1(:,:)           = pxtm1(1:kproma,:,jt) + pxtte(1:kproma,:,jt) * ztmst
!          zxtp1(:,:)           = EXP (-ztmst/trlist% ti(jt)% tdecay) * zxtp1(:,:)
!          pxtte(1:kproma,:,jt) = (zxtp1(:,:)-pxtm1(1:kproma,:,jt))*zqtmst
       END IF
!uh-
    END DO




  END SUBROUTINE radionucl_sink
!==============================================================================
END MODULE mo_sub_echam
