SUBROUTINE xt_burden(kproma, kbdim, klev, klevp1, krow, & ! Local dimensions
                     papp1,  paphp1,                    & ! Thermodynamic quantities
                     pxtm1,  pxtte                      ) ! Tracer fields

  !  *xt_burden* calculates the atmospheric burden of all tracers
  !
  !  Author:
  !  -------
  !  Philip Stier, MPI-MET              08/2002
  !
  !  Method:
  !  -------
  !  Calculate burden [kg m-2] from mixing ratio [kg kg-1] by multiplying 
  !  with the term dp/g and summing up the vertical integral.
  !  The burden is multiplied by dt and divided by the output interval 
  !  in the stream management to give a mean over the output interval. 
  !

  USE mo_kind,         ONLY: dp
  USE mo_time_control, ONLY: delta_time, time_step_len, &
!uh+
       current_date, &
       previous_date, get_date_components
!uh-
  USE mo_constants,    ONLY: g 
  USE mo_parameters,   ONLY: jptrac
  USE mo_tracer,       ONLY: ntrac, trlist
  USE mo_doctor,       ONLY: nout
  USE mo_mpi,          ONLY: p_parallel_io
  USE mo_aero_mem,     ONLY: d_burden

!uh+
  USE mo_be10,         ONLY: str_trop_budgets, be10_massbal
  USE mo_control,      ONLY: lcn
!uh-


  IMPLICIT NONE

  !--- Dummy Variables:

  INTEGER,  INTENT(in) :: kproma,     kbdim,       klev,   klevp1, krow

  REAL(dp), INTENT(in) :: paphp1(kbdim,klevp1),    papp1(kbdim,klev)

  REAL(dp), INTENT(in) :: pxtm1(kbdim,klev,ntrac), pxtte(kbdim,klev,ntrac)


  !--- Local Variables:

  INTEGER              :: jl,       jk,       jt

!uh
!  INTEGER             :: ktropo

  REAL(dp)             :: zxtp1

  REAL(dp)             :: zdpg(kbdim,klev)

  LOGICAL, SAVE        :: lfirst=.TRUE.
 
  LOGICAL, SAVE        :: lburden(jptrac)=.FALSE.

  !--- Diagnostic stream:

  REAL(dp),    POINTER :: d_burden_p(:,:)

!uh+
  LOGICAL             :: lnewday
  INTEGER             :: ithisyear, ilastyear, ithismonth, ilastmonth,    &
       ithisday, ilastday, ithishour, ilasthour, ithisminute, ithissecond
!uh-





  !--- 0) Initializations:

  !--- Calculate burden only for tracers that are given the unit of mixing ratio,
  !    i.e. assume that the unit is mixing ratio occurs if the string kg-1 occurs
  !    in the tracer units:
  
  IF (lfirst .AND. ntrac>0) THEN

     IF (p_parallel_io) THEN
        WRITE(NOUT,*) '-------------------------------------------------------------------------------------'
        WRITE(NOUT,*) '--- xt_burden: '
     END IF

     DO jt=1, ntrac

        lburden(jt) = INDEX(trlist%ti(jt)%units,'kg-1')/=0
        
        IF (p_parallel_io) THEN
           IF(lburden(jt)) THEN
              WRITE(NOUT,*) '               Calculate burden    for tracer: ',TRIM(trlist%ti(jt)%fullname)
           ELSE
              WRITE(NOUT,*) '               Calculate NO BURDEN for tracer: ',TRIM(trlist%ti(jt)%fullname), &
                                            ' ( [', TRIM(trlist%ti(jt)%units),']', ' no mixing ratio)'
           END IF
        END IF

     END DO

     IF (p_parallel_io) THEN
        WRITE(NOUT,*) '-------------------------------------------------------------------------------------'
     END IF

     lfirst=.FALSE.

  END IF


  !--- 1) Calculate auxiliary variable dp/g :

  !--- Uppermost level:

  zdpg(1:kproma,1)=2._dp*(paphp1(1:kproma,2)-papp1(1:kproma,1))/g

  !--- Other levels:

  DO jk=2, klev
     
     zdpg(1:kproma,jk)=(paphp1(1:kproma,jk+1)-paphp1(1:kproma,jk))/g

  END DO


  !--- 2) Calculate and store d_burden:

  DO jt=1, ntrac

     IF (lburden(jt)) THEN

        d_burden_p => d_burden(jt)%ptr

        DO jk=1, klev

           DO jl=1, kproma

              zxtp1=pxtm1(jl,jk,jt)+pxtte(jl,jk,jt)*time_step_len

              d_burden_p(jl,krow)=d_burden_p(jl,krow)+zxtp1*zdpg(jl,jk)*delta_time

           END DO

        END DO

     END IF

  END DO


!uh+
  IF (lcn) THEN
     CALL get_date_components(previous_date, month=ilastmonth, year=ilastyear, &
          day=ilastday, hour=ilasthour)
     CALL get_date_components(current_date, month=ithismonth, year=ithisyear, &
         day=ithisday, hour=ithishour, minute=ithisminute, second=ithissecond)
     lnewday = ilastday /= ithisday

!     IF (lnewday) CALL be10_massbal(kproma,kbdim,klev,klevp1,papp1,paphp1,pxtm1,pxtte)
     CALL be10_massbal(kproma,kbdim,klev,klevp1,krow,papp1,paphp1,pxtm1,pxtte)
! uh: stratospheric and tropospheric budgets


     CALL str_trop_budgets(kproma,kbdim,klev,krow,pxtm1,pxtte,zdpg)

  END IF !(lcn)
!uh-


END SUBROUTINE xt_burden
