
SUBROUTINE xt_sedimentation(kproma, kbdim, klev, krow, & ! ECHAM indices
                            pxtm1,  pxtte,             & !   "   tracers
                            ptm1,   ptte,              & !   "   temperature
                            pqm1,   pqe,               & !   "   humidity
                            papp1,  paphp1             ) !   "   pressure (full/half levels)
                            
  !     -----------------------------------------------------------------------
  !
  !     **** sedimentation  -  calculation of sedimentation velocities
  !     
  !     Authors:
  !     --------
  !     Michael Schulz, LSCE                           5.2.2001
  !     Philip Stier, MPI    (adaption to ECHAM5/HAM)      2002
  !
  !     interface
  !     ---------
  !      input 
  !       pressure difference and midpoint pressure
  !       temperature
  !       accumulated tracer mixing ratios 
  !       old sedimentation flux
  !      output 
  !       accumulated tracer mixing ratios 
  !       new sedimentation flux to be found in AEROSOL_PROGNOS fields
  !
  !     method
  !     ------
  !     
  !     returns the changed tracer mixing ratio 
  !     expects that the aerosol tracers are in the tracer fields
  !     in the order explained in the aerosol module
  !
  !     Loop is done per aerosol mode, which shall be affected by sedimentation
  !     Modes treated must be "behind" each other in the tracer field
  !     and are identified by mode names such as cimode,csmode...
  !
  !     References:
  !     -----------
  !     Seinfeld &Pandis (1998)
  !     -----------------------------------------------------------------------
  !
  !     INPUT ARGUMENTS.
  !     ----- ----------
  !    Variables in routine    
  !  zxtp1     : tracer mixing ratio including tendency

  USE mo_kind,         ONLY: dp
  USE mo_time_control, ONLY: time_step_len,        delta_time
  USE mo_tracer,       ONLY: ntrac,                trlist,    &
                             AEROSOLNUMBER,        AEROSOLMASS
  USE mo_constants,    ONLY: rd,        vtmpc1,    g, tmelt
  USE mo_aero_m7,      ONLY: sigma,     sigmaln,   cmedr2mmedr
  USE mo_aero_mem,     ONLY: sed, vsediment, rwet, densaer

! uh+
  USE mo_be10,         ONLY: area_avg, BE_FLAG, idt_be10, idt_be7, idt_cl36, &
                             idt_na22, be10_sed_acc,   &
                             be7_sed_acc, cl36_sed_acc, na22_sed_acc, idt_pb210, &
                             pb210_sed_acc
  USE mo_doctor,       ONLY: nout
  USE mo_control,      ONLY: lcn, lbe10, lbe7, lcl36, lna22, lpb210
! uh-


  IMPLICIT NONE

  !--- Arguments:
  
  INTEGER, INTENT(IN)    :: kproma, kbdim, klev, krow
  REAL(dp),INTENT(IN)    :: ptm1(kbdim,klev),          ptte(kbdim,klev),    &
                            pqm1(kbdim,klev),          pqe(kbdim,klev),     &
                            papp1(kbdim,klev),         paphp1(kbdim,klev+1)
  REAL(dp),INTENT(IN)    :: pxtm1(kbdim,klev,ntrac)
  
  REAL(dp),INTENT(INOUT) :: pxtte(kbdim,klev,ntrac)

  !--- Diagnostic stream:

  REAL(dp),      POINTER :: rwet_p(:,:,:)
  REAL(dp),      POINTER :: densaer_p(:,:,:)
  REAL(dp),      POINTER :: sed_p(:,:)
  REAL(dp),      POINTER :: vsediment_p(:,:)

  !--- Local variables:

  INTEGER :: jl, jk, jt, imod, jj ! uh jl,jj added
  REAL(dp):: slinnfac

  REAL(dp):: zmd(kbdim,klev),      vt(kbdim,klev,ntrac),      & !uh ntrac added
             ztp1(kbdim,klev),     ztempc(kbdim,klev),  &
             zvis(kbdim,klev),     zlair(kbdim,klev),   &
             zdensair(kbdim,klev), zqp1(kbdim,klev),    &
             zdpg(kbdim,klev),     zdz(kbdim,klev),     &
             zsedtend(kbdim,klev), zsedflux(kbdim,klev,ntrac)  !uh ntrac added
             

  REAL(dp):: zxtp1(kbdim,klev,ntrac)

  !--- 0) Initialisations:

  !--- Calculate tracer mixing ratios at t=t+1:

  zxtp1(1:kproma,:,:)=pxtm1(1:kproma,:,:)+pxtte(1:kproma,:,:)*time_step_len


  !--- 1) Calculate meteorological/thermodynamical parameters:
  
  !--- Temperature in C at t=t+1:

  ztp1(1:kproma,:)= ptm1(1:kproma,:) + ptte(1:kproma,:)*time_step_len
  ztempc(1:kproma,:) = ztp1(1:kproma,:) - tmelt

  !---  Dynamic viscosity of air after Prup.Klett: zvis [Pa s]

  WHERE(ztempc(1:kproma,:) >= 0._dp)
     zvis(1:kproma,:)=(1.718_dp + 0.0049_dp*ztempc(1:kproma,:))*1.E-5_dp
  ELSEWHERE
     zvis(1:kproma,:)=(1.718_dp + 0.0049_dp*ztempc(1:kproma,:) - 1.2E-05_dp*(ztempc(1:kproma,:)**2))*1.E-5_dp
  ENDWHERE

  !--- Mean free path of air after Prupp. Klett: zlair [10^-6 m]

  zlair(1:kproma,:) = 0.066_dp *(1.01325E+5_dp/papp1(1:kproma,:))*(ztp1(1:kproma,:)/293.15_dp)*1.E-06_dp

  !--- Density of air:

  zqp1(1:kproma,:)=pqm1(1:kproma,:)+pqe(1:kproma,:)*time_step_len

  zdensair(1:kproma,:)=papp1(1:kproma,:)/(ztp1(1:kproma,:)*rd*(1._dp+vtmpc1*zqp1(1:kproma,:)))

  !--- Air mass auxiliary variable:

  !    Uppermost level:

  zdpg(1:kproma,1)=2._dp*(paphp1(1:kproma,2)-papp1(1:kproma,1))/g

  !    Other levels:

  DO jk=2, klev
     
     zdpg(1:kproma,jk)=(paphp1(1:kproma,jk+1)-paphp1(1:kproma,jk))/g

  END DO

  !--- Layer thickness:

  zdz(1:kproma,:)=zdpg(1:kproma,:)/zdensair(1:kproma,:)


  !--- 2) Calculate sedimentation:

  DO jt = 1, ntrac

     !--- Lookup mode of the tracer:
        
     imod=trlist%ti(jt)%mode

!uh+
!     IF (trlist%ti(jt)%nsedi/=0) THEN
     IF (trlist%ti(jt)%nsedi > 0.AND.trlist%ti(jt)%nsedi < BE_FLAG) THEN
!uh-

        rwet_p      => rwet(imod)%ptr
        densaer_p   => densaer(imod)%ptr
        sed_p       => sed(jt)%ptr
        vsediment_p => vsediment(jt)%ptr

        !--- Select diameter and limit it to maximal 50 um:
        !
        !    If tracer is:
        !
        !       aerosol number mixing ratio: use number median radius
        !       aerosol mass   mixing ratio: use mass   median radius

        IF (trlist%ti(jt)%nphase==AEROSOLNUMBER) THEN 

           zmd(1:kproma,:)=MIN( rwet_p(1:kproma,:,krow)*2._dp ,  50.E-6_dp)

        ELSE IF (trlist%ti(jt)%nphase==AEROSOLMASS) THEN

           zmd(1:kproma,:)=MIN( rwet_p(1:kproma,:,krow)*cmedr2mmedr(imod)*2._dp , 50.E-6_dp)

        END IF

        !--- Slinn correction for sedimentation velocity of a 
        !    size distribution with a given sigma:

        slinnfac=sigma(imod)**(2._dp*sigmaln(imod))


        WHERE (zmd(1:kproma,:) > 0._dp) 

           !--- Stokes-velocity (S&P, Equation 8.42):

           vt(1:kproma,:,jt)=2._dp/9._dp*(densaer_p(1:kproma,:,krow)-zdensair(1:kproma,:)) * &
                          g/zvis(1:kproma,:)*(zmd(1:kproma,:)/2._dp)**2._dp

           !--- With cunnigham slip- flow correction (S&P, Equation 8.34):

           vt(1:kproma,:,jt) = vt(1:kproma,:,jt) *                                   &
                            slinnfac       *                                   &
                            (1._dp+ 1.257_dp*zlair(1:kproma,:)/zmd(1:kproma,:)*2._dp +  &
                             0.4_dp*zlair(1:kproma,:)/zmd(1:kproma,:)*2._dp *        &
                             EXP(-1.1_dp/(zlair(1:kproma,:)/zmd(1:kproma,:)*2._dp)) )

           !--- Calculate sedimentation in terms of mixing ratio tendencies:

           !--- Limit loss to the content of box:
           !    Multilayer crossing is not realised here as sedimentation 
           !    velocity is in effect limited to dz/dt (grid velocity).
           !    
           !    vt=dz(dep)/dt

           !--- Limit vt to the grid velocity dz(grid)/dt (no multilayer crossing):

           vt(1:kproma,:,jt)=MIN( vt(1:kproma,:,jt) , zdz(1:kproma,:)/time_step_len )

           !--- Loss in terms of mixing ratio tendency:
           
           zsedtend(1:kproma,:)=MAX(0._dp,MIN(zxtp1(1:kproma,:,jt)/time_step_len, zxtp1(1:kproma,:,jt) * &
                                                                               vt(1:kproma,:,jt)/zdz(1:kproma,:) ))

           !--- Apply loss throughout the column of mixing ratio tendencies:

           pxtte(1:kproma,:,jt)= pxtte(1:kproma,:,jt)-zsedtend(1:kproma,:)

           !--- Transfer loss from mixing ratio tendency [kg/kg] 
           !    to sedimentation flux [kg m-2 s-1]:

       !uh+    
           zsedflux(1:kproma,:,jt) = zsedtend(1:kproma,:) *zdpg(1:kproma,:)
!           zsedflux(1:kproma,:) = zsedtend(1:kproma,:) *zdpg(1:kproma,:)
       !uh-

        ELSEWHERE

!uh+
!           zsedflux(1:kproma,:) = 0._dp 
           zsedflux(1:kproma,:,jt) = 0._dp 
           vt(1:kproma,:,jt) =0._dp
!           vt(1:kproma,:) =0._dp
!uh-


        END WHERE


        !--- Re-convert sedimentatation flux and add it to the mixing ratio tendency
        !    of the box below (conversion with zdpg of the box below): 
        
        pxtte(1:kproma,2:klev,jt)=pxtte(1:kproma,2:klev,jt)+(zsedflux(1:kproma,1:(klev-1),jt)/zdpg(1:kproma,2:klev))

        !--- Store sedimentation in [kg/m2] for diagnostics:
        !    Note: diagnostics has to be integrated over
        !          delta_time not time_step_len

        sed_p(1:kproma,krow)=sed_p(1:kproma,krow)+zsedflux(1:kproma,klev,jt)*delta_time

        !--- Store sedimentation velocity for diagnostics:

        vsediment_p(1:kproma,krow)=vt(1:kproma,klev,jt)

     END IF ! nsedi /=0

  END DO !jt=1, ntrac

! uh:
! ...  mo_be10 alled, area weighted average of sedimentation
! ... velocities of all tracers (subjected to sedimentation) calculated.

!uh+

  IF (lcn) THEN
     DO jt=1,ntrac

        IF (trlist%ti(jt)%nsedi==BE_FLAG) THEN

           CALL area_avg(kproma,klev,kbdim,krow,1,pxtm1,vt,jt)

           vsediment_p => vsediment(jt)%ptr
           sed_p       => sed(jt)%ptr

           zsedtend(1:kproma,:)=MAX(0._dp,MIN(zxtp1(1:kproma,:,jt)/time_step_len, zxtp1(1:kproma,:,jt) * &
                vt(1:kproma,:,jt)/zdz(1:kproma,:) ))

           pxtte(1:kproma,:,jt)= pxtte(1:kproma,:,jt)-zsedtend(1:kproma,:)
           zsedflux(1:kproma,:,jt) = zsedtend(1:kproma,:) *zdpg(1:kproma,:)
           pxtte(1:kproma,2:klev,jt)=pxtte(1:kproma,2:klev,jt)+(zsedflux(1:kproma,1:(klev-1),jt)/zdpg(1:kproma,2:klev))
           sed_p(1:kproma,krow)=sed_p(1:kproma,krow)+zsedflux(1:kproma,klev,jt)*delta_time
           vsediment_p(1:kproma,krow)=vt(1:kproma,klev,jt)

!uh: copy the calculated sed. velocity for other CN tracers
           DO jj=jt+1,ntrac

              IF (trlist%ti(jj)%nsedi==BE_FLAG) THEN                 
                 vsediment_p => vsediment(jj)%ptr
                 sed_p       => sed(jj)%ptr

                 vt(1:kproma,:,jj)=vt(1:kproma,:,jt)
                 zsedtend(1:kproma,:)=MAX(0._dp,MIN(zxtp1(1:kproma,:,jj)/time_step_len, zxtp1(1:kproma,:,jj) * &
                      vt(1:kproma,:,jj)/zdz(1:kproma,:) ))
                 pxtte(1:kproma,:,jj)= pxtte(1:kproma,:,jj)-zsedtend(1:kproma,:) 
                 zsedflux(1:kproma,:,jj) = zsedtend(1:kproma,:) *zdpg(1:kproma,:)
                 pxtte(1:kproma,2:klev,jj)=pxtte(1:kproma,2:klev,jj)+    &
                      (zsedflux(1:kproma,1:(klev-1),jj)/zdpg(1:kproma,2:klev))

                 sed_p(1:kproma,krow)=sed_p(1:kproma,krow)+zsedflux(1:kproma,klev,jj)*delta_time
                 vsediment_p(1:kproma,krow)=vt(1:kproma,klev,jj)

              END IF
           END DO
           EXIT
        END IF
     END DO
  END IF !lcn
!uh-


! uh:
! save sedimentation for mass balance

  IF (lcn) THEN
     DO jt=1,ntrac
        IF (trlist%ti(jt)%nsedi==BE_FLAG) THEN
           DO jl=1,kproma
              IF (lbe10.AND.jt==idt_be10) THEN
                 be10_sed_acc(jl,krow)=be10_sed_acc(jl,krow)+zsedflux(jl,klev,jt)*delta_time
              END IF
              IF (lbe7.AND.jt==idt_be7) THEN
                 be7_sed_acc(jl,krow)=be7_sed_acc(jl,krow)+zsedflux(jl,klev,jt)*delta_time
              END IF
              IF (lcl36.AND.jt==idt_cl36) THEN
                 cl36_sed_acc(jl,krow)=cl36_sed_acc(jl,krow)+zsedflux(jl,klev,jt)*delta_time
              END IF
              IF (lna22.AND.jt==idt_na22) THEN
                 na22_sed_acc(jl,krow)=na22_sed_acc(jl,krow)+zsedflux(jl,klev,jt)*delta_time
              END IF
              IF (lpb210.AND.jt==idt_pb210) THEN
                 pb210_sed_acc(jl,krow)=pb210_sed_acc(jl,krow)+zsedflux(jl,klev,jt)*delta_time
              END IF
           END DO
        END IF
     END DO
  END IF



END SUBROUTINE xt_sedimentation
