SUBROUTINE xt_wetdep(kbdim,  kproma,    klev,     ktop,     & ! dimensions
                     krow,                                  & ! longitude (kproma-block) index
                     pxtp10, pxtp1c,                        & ! mixing ratios (ambient, in-cloud, downdraft)
                     pmfu,   pmfuxt,    prhou,              & ! convective properties (not needed for stratiform)
                     pxtm1,                                 & ! stratiform properties (not needed for convective)
                     pmlwc,  pmiwc,     pmratepr, pmrateps, & ! cloud properties
                     pfrain, pfsnow,    pfevapr,  pfsubls,  & !   "       "
                     paclc,  pmsnowacl, lstrat,             & !   "       "
                     ptm1,   pdpg,      plfrac,             & ! thermodynamic quantities
                     pxtte                                  ) ! mixing ratio tendency


  ! *xt_wetdep* calculates the wet deposition of tracers
  !
  ! Authors:
  ! --------
  ! Philip Stier,     MPI-MET, Hamburg       2001 - 2004
  ! Claudia Timmreck, MPI-MET, Hamburg            - 2001
  ! Johann Feicher,   MPI-MET, Hamburg            - 2001
  !
  ! Method:
  ! -------
  ! Currently the scavenging ratios are described in dependence of mode
  ! stratiform or convective cloud type. This will be replaced by the 
  ! explicit formulation of nucleation and impaction scavenging. 
  !
  ! Externals:
  ! ----------
  !
  ! References:
  ! -----------
  !
  ! Variables:
  ! ----------
  ! pxtp10         = ambient tracer mass mixing ratio at t=t+dt 
  !                  units: [kg(tracer) kg-1(air)] incoming and returned
  !                         [kg(tracer) m-2]       during calculation
  ! pxtp1c         = in-cloud tracer mass mixing ratio at t=t+dt 
  !                  units: [kg(tracer) kg-1(air)] incoming and returned
  !                         [kg(tracer) m-2]       during calculation
  !
  ! zxtfracwat     = total fraction of the tracer associated with water 
  !                  that is in the water phase [%]
  ! zxtfracice     = total fraction of the tracer associated with ice 
  !                  that is in the ice phase [%]
  ! zxtfracXXX_YYY = fraction of the tracer associated with water 
  !                  that is in the XXX phase due to process YYY 
  !                  where YYY=(nuc: nucleation; imp: impaction; 
  !                  dif: gaseous diffusion) [%]
  ! zdxtwat        = total change of the tracer in the water phase during
  !                  one timestep [kg m-2]
  ! zdxtice        = total change of the tracer in the ice phase during
  !                  one timestep [kg m-2]

  USE mo_kind,         ONLY: dp
  USE mo_time_control, ONLY: time_step_len, delta_time
  USE mo_tracer,       ONLY: trlist, ntrac,     &
                             GAS, AEROSOLMASS, AEROSOLNUMBER
  USE mo_exception,    ONLY: finish
  USE mo_aero_mem,     ONLY: precipform, precipevap, uparfrac
  USE mo_aero_m7,      ONLY: nmod
  USE mo_aero_mem,     ONLY: wdep, wdepinst
  USE mo_aero_mem,     ONLY: wdep_conv, wdep_strat
  USE mo_constants,    ONLY: tmelt

!uh+
  USE mo_be10,         ONLY: area_avg, BE_FLAG, idt_be10, idt_be7, idt_cl36, &
       idt_na22, burd_wdep, burd_wdep_be7, burd_wdep_cl36,  &
       burd_wdep_na22, tend_mass, tend_wdep, &
       burd_wdep_pb210, idt_pb210, wdep_be10
  USE mo_control,      ONLY: lcn
!uh-


  IMPLICIT NONE

  LOGICAL :: lstrat ! Indicates stratiform clouds (call from cloud)

  INTEGER :: kbdim,     kproma,       klev,     ktop,  krow

  REAL(dp):: pmlwc(kbdim,klev),       pmiwc(kbdim,klev),        &
             pmratepr(kbdim,klev),    pmrateps(kbdim,klev),     &
             pfrain(kbdim,klev),      pfsnow(kbdim,klev),       &
             pfevapr(kbdim,klev),     pfsubls(kbdim,klev),      &
             paclc(kbdim,klev),       pmsnowacl(kbdim,klev),    &
             ptm1(kbdim,klev),        pdpg(kbdim,klev),         &
             pmfu(kbdim,klev),        prhou(kbdim,klev),        &
             plfrac(kbdim,klev)

  REAL(dp):: pxtm1(kbdim,klev,ntrac), pxtte(kbdim,klev,ntrac),  &
             pxtp1c(kbdim,klev,ntrac),pxtp10(kbdim,klev,ntrac), &
             pmfuxt(kbdim,klev,ntrac)


  !--- Local variables:

  LOGICAL :: laerosol,  lgas

  INTEGER :: jl,        jk,           jt,                   &
             imod

  REAL(dp):: ztmst,     zqtmst

  REAL(dp):: zilwc,                &
             zxtice,    zxtwat,       zxtp1, &
             zmprec,            &
             zfprec,    zeps,     &
             zdxtwat,   zdxtice,      zdxtcol,  &
             zdxtevap,  zdep

  REAL(dp):: zxtfracwat,              zxtfracice,           &
! uh+
!            zxtte_strat, zxtte_conv   
             zxtte_strat(kbdim,klev,ntrac),zxtte_conv(kbdim,klev,ntrac),     &                
             ztend(kbdim,klev,ntrac)
 ! uh-

  REAL(dp):: zsump(kbdim),            zsumclcp(kbdim),      &
             zdepint(kbdim)

  REAL(dp):: zice(kbdim,klev),        zclceff(kbdim,klev),  &
             zeffice(kbdim,klev),     zeffwat(kbdim,klev),  &
             zrevap(kbdim,klev),      zaclc(kbdim,klev)

  REAL(dp):: zxtfracwat_dif(kbdim,klev),                    &
             zxtfracwat_nuc(kbdim,klev),                    &
             zxtfracwat_imp(kbdim,klev),                    &
             zxtfracice_dif(kbdim,klev),                    &
             zxtfracice_nuc(kbdim,klev),                    &
             zxtfracice_imp(kbdim,klev),                    &
             zxtfrac_col(kbdim,klev)


  !--- Diagnostic stream:

  REAL(dp),        POINTER :: wdep_p(:,:)
  REAL(dp),        POINTER :: wdepinst_p(:,:)
  REAL(dp),        POINTER :: wdep_conv_p(:,:), wdep_strat_p(:,:)

!uh:
  REAL(dp),            POINTER :: tend_mass_p(:,:,:)
  REAL(dp),            POINTER :: tend_wdep_p(:,:,:)
!uh-


  !--- Constants:

  REAL(dp), PARAMETER :: zmin =1.e-10_dp, &
                         zmin2=1.e-20_dp

  !--- Mean mass scavenging coefficients normalized by rain-rate [kg m-2]:

  !    Rain: Seinfeld & Pandis, Fig 20.15:

  REAL(dp), PARAMETER :: zbcr(nmod)=(/ 5.0E-4_dp, 1.0E-4_dp, 1.0E-3_dp, 1.0E-1_dp, 1.0E-4_dp, 1.0E-3_dp, 1.0E-1_dp /)

  !    Snow: Little available, graphs in Prupbacher show similar order of magnitude in 
  !          collection efficiency as rain,therefore assume typical mean value as above:

  REAL(dp), PARAMETER :: zbcs(nmod)=(/ 5.0E-3_dp, 5.0E-3_dp, 5.0E-3_dp, 5.0E-3_dp, 5.0E-3_dp, 5.0E-3_dp, 5.0E-3_dp /)

  !--- Assumed updraft velocity in convective clouds [m s-1]:

  REAL(dp), PARAMETER :: zwu=2.0_dp

  !--- Prescribed scavenging ratios:

  !--- stratifrom cloud water in water and mixed clouds (T>238.5K):

  REAL(dp) :: csr_strat_wat(7)

  !--- stratifrom cloud ice in mixed clouds (T>238.5K):

  REAL(dp) :: csr_strat_mix(7)

  !--- stratifrom cloud ice in ice clouds (T<238.5K):

  REAL(dp) :: csr_strat_ice(7)

  !--- convective clouds:

  REAL(dp) :: csr_conv(7)


  !--- 0) Initializations:

  IF (nmod/=7) CALL finish('xt_wetdep',' incompatible number of modes')

  zeps=EPSILON(1.0_dp)

  ztmst=time_step_len
  zqtmst=1._dp/ztmst

  zsumclcp(:) = 0._dp
  zsump(:)    = 0._dp

  zaclc(:,:)  = 0._dp 


  !--- 0.1) Pre-set scavenging ratios:

  !--- Stratiform:

! uh: scavenging coefficients revisited: Bourgeois and Bey, 2011, transport of aerosol to Arctic too low

!  csr_strat_wat = (/0.10_dp, 0.25_dp, 0.85_dp, 0.99_dp, 0.20_dp, 0.40_dp, 0.40_dp/)
  csr_strat_wat = (/0.06_dp, 0.25_dp, 0.85_dp, 0.99_dp, 0.20_dp, 0.40_dp, 0.40_dp/)

!  csr_strat_mix = (/0.10_dp, 0.40_dp, 0.75_dp, 0.75_dp, 0.10_dp, 0.40_dp, 0.40_dp/)
  csr_strat_mix = (/0.06_dp, 0.06_dp, 0.06_dp, 0.75_dp, 0.06_dp, 0.06_dp, 0.40_dp/)

!  csr_strat_ice = (/0.10_dp, 0.10_dp, 0.10_dp, 0.10_dp, 0.10_dp, 0.10_dp, 0.10_dp/)
  csr_strat_ice = (/0.06_dp, 0.06_dp, 0.06_dp, 0.06_dp, 0.06_dp, 0.06_dp, 0.06_dp/)


  !--- Convective:

  csr_conv      = (/0.20_dp, 0.60_dp, 0.99_dp, 0.99_dp, 0.20_dp, 0.40_dp, 0.40_dp/)

!uh-
  IF (lcn) THEN
     DO jt=1,ntrac
        DO jk=1,klev
           DO jl=1,kproma
              ztend(jl,jk,jt)=0._dp
           END DO
        END DO
     END DO
  END IF !(lcn)
! uh-


  !--- 1) Calculate hydrological parameters:

  DO  jk=ktop,klev
     DO jl=1,kproma

        !--- 1.1) Calculate ice mass-fraction of the total water:

        pmiwc(jl,jk)=MAX(pmiwc(jl,jk),0._dp)
        pmlwc(jl,jk)=MAX(pmlwc(jl,jk),0._dp)

        zaclc(jl,jk)=paclc(jl,jk)

        zilwc=pmiwc(jl,jk)+pmlwc(jl,jk)

        IF(zilwc.GT.zmin) THEN
           zice(jl,jk)=pmiwc(jl,jk)/zilwc
        ELSE
           zaclc(jl,jk)=0._dp
           zice(jl,jk)=0._dp
        ENDIF

        !--- 1.2) Calculate autoconversion rate:

        IF(pmiwc(jl,jk).GT.zmin) THEN
           zeffice(jl,jk)=pmrateps(jl,jk)/pmiwc(jl,jk)
           zeffice(jl,jk)=MAX(0._dp,MIN(1._dp,zeffice(jl,jk)))
        ELSE
           zeffice(jl,jk)=0._dp
        ENDIF

        IF(pmlwc(jl,jk).GT.zmin) THEN
           zeffwat(jl,jk)=(pmratepr(jl,jk)+pmsnowacl(jl,jk))/pmlwc(jl,jk)
           zeffwat(jl,jk)=MAX(0._dp,MIN(1._dp,zeffwat(jl,jk)))
        ELSE
           zeffwat(jl,jk)=0._dp
        ENDIF

        !--- 1.3) Calculate the effective grid-box fraction 
        !         affected by precipitation (zclceff):

        !--- Preciptiation formation:

        zmprec=pmratepr(jl,jk)+pmrateps(jl,jk)+pmsnowacl(jl,jk)

        !--- Store for diagnostics [kg m-2 s-1]:

        precipform(jl,jk,krow)=zmprec*pdpg(jl,jk)/ztmst

        !--- 1.3.1) Stratiform clouds:

        IF (lstrat) THEN

           !--- Precipitaion-formation weighted averaging:

           zsumclcp(jl)=zsumclcp(jl)+zaclc(jl,jk)*zmprec
           zsump(jl)=zsump(jl)+zmprec

           IF(zsump(jl).GT.zmin2) THEN
              zclceff(jl,jk)=zsumclcp(jl)/zsump(jl)
           ELSE
              zclceff(jl,jk)=0._dp
           ENDIF

        !--- 1.3.2) Convective clouds:

        ELSE

           !--- Assume updraft area fraction as precipitiating 
           !    fraction of grid box
           !
           !    Estimate updraft area from prescribed updraft
           !    velocities

           zclceff(jl,jk)=pmfu(jl,jk)/(zwu*prhou(jl,jk))

           uparfrac(jl,jk,krow)=zclceff(jl,jk)

        END IF

        !--- 1.4) Re-evaporation

        zfprec=pfrain(jl,jk)+pfsnow(jl,jk)

        IF(zfprec.GT.zmin) THEN
           zrevap(jl,jk)=(pfevapr(jl,jk)+pfsubls(jl,jk))/zfprec
        ELSE
           zrevap(jl,jk)=0._dp
        ENDIF

        zrevap(jl,jk)=MAX(0._dp,MIN(1._dp,zrevap(jl,jk)))

        !--- Store re-evaporatio + sublimation for diagnostics:
        
        precipevap(jl,jk,krow)=pfevapr(jl,jk)+pfsubls(jl,jk)

     END DO
  END DO


  !--- 2) Scavenging:

  tracer : DO jt=1, ntrac

     !--- Initialization:

     imod    =trlist%ti(jt)%mode
     laerosol=(trlist%ti(jt)%nphase==AEROSOLMASS .OR. trlist%ti(jt)%nphase==AEROSOLNUMBER)
     lgas    =trlist%ti(jt)%nphase==GAS

     zdepint(1:kproma)    = 0._dp

     zxtfracwat           = 0._dp
     zxtfracice           = 0._dp

     zxtfracwat_nuc(:,:)  = 0._dp
     zxtfracwat_imp(:,:)  = 0._dp
     zxtfracwat_dif(:,:)  = 0._dp
     zxtfracice_nuc(:,:)  = 0._dp
     zxtfracice_imp(:,:)  = 0._dp
     zxtfracice_dif(:,:)  = 0._dp
     zxtfrac_col(:,:)     = 0._dp


! uh+:
     !     nwetdep :IF (trlist%ti(jt)%nwetdep>0) THEN

     nwetdep :IF (trlist%ti(jt)%nwetdep>0 .AND. trlist%ti(jt)%nwetdep < BE_FLAG) THEN
!uh-


        !--- Consistency check:

        IF(.NOT.(laerosol .OR. &
                 trlist%ti(jt)%fullname=='SO2'.OR.trlist%ti(jt)%fullname=='SO4_gas')) THEN
           CALL finish('xt_wetdep',' scavenging parameters not defined for '//trlist%ti(jt)%fullname)
        END IF

        !--- Get diagostic fields:

        wdep_p     => wdep(jt)%ptr
        wdep_strat_p => wdep_strat(jt)%ptr
        wdep_conv_p  => wdep_conv(jt)%ptr
        wdepinst_p => wdepinst(jt)%ptr

        wdepinst_p(1:kproma,krow)=0._dp

        !--- 2.1) Calculate scavenging parameters:

        !--- 2.1.1) Aerosols:

        IF(lstrat) THEN

        IF(laerosol) THEN

           !--- Nucleation and impaction in water clouds:

           WHERE(ptm1(1:kproma,:)> tmelt)

              zxtfracwat_nuc(1:kproma,:)=csr_strat_wat(imod)
              zxtfracice_nuc(1:kproma,:)=csr_strat_wat(imod)

           !--- Nucleation and impaction in mixed clouds:

           ELSEWHERE(ptm1(1:kproma,:)>238.15_dp)

              zxtfracwat_nuc(1:kproma,:)=csr_strat_mix(imod)
              zxtfracice_nuc(1:kproma,:)=csr_strat_mix(imod)

           ELSEWHERE

           !--- Nucleation and impaction in cold clouds:

              zxtfracwat_nuc(1:kproma,:)=csr_strat_wat(imod)
              zxtfracice_nuc(1:kproma,:)=csr_strat_ice(imod)

           ENDWHERE

           !--- Below-cloud scavenging:
           !    Calculate fraction of below cloud scavenged tracer
           !    dM/M = -C dt where the mean scavenging coefficient C = Cn x R where Cn is normalized 
           !    scavenging coefficient (above) and R is the rain-rate kg m-2 s-1, dt is the timestep

           WHERE(zaclc(1:kproma,:) < zmin)

              zxtfrac_col(1:kproma,:)=ztmst*(zbcr(imod)*pfrain(1:kproma,:)+zbcs(imod)*pfsnow(1:kproma,:))
              zxtfrac_col(1:kproma,:)=MAX(0._dp,zxtfrac_col(1:kproma,:))

           ELSEWHERE

              zxtfrac_col(1:kproma,:)=0._dp

           ENDWHERE

        !--- 2.2.2) Gases:

        ELSE IF (lgas) THEN

           IF(trlist%ti(jt)%basename=='SO2') THEN

              !--- Gas phase scavenging utilizing the liquid fraction of total SO2 
              !    calculated from Henry's law in xt_chemistry:

              zxtfracwat_dif(1:kproma,:)=plfrac(1:kproma,:)
              zxtfracice_dif(1:kproma,:)=0._dp
                 
           ELSE IF(trlist%ti(jt)%fullname=='SO4_gas') THEN

              zxtfracwat_dif(1:kproma,:)=1.0_dp
              zxtfracice_dif(1:kproma,:)=0._dp

           END IF

        END IF

        DO  jk=ktop,klev
           DO jl=1,kproma

           !--- Weight mixing ratios with cloud fraction:

           pxtp1c(jl,jk,jt)=pxtp1c(jl,jk,jt)*zaclc(jl,jk)
           pxtp10(jl,jk,jt)=pxtp10(jl,jk,jt)*(1._dp-zaclc(jl,jk))


           !--- 2.2) Partitioning between ice and liquid phase:

           !--- Associate tracer masses in the cloud fraction to water phase
           !    and ice phase according to the respective mass fractions:

           zxtice=pxtp1c(jl,jk,jt)*zice(jl,jk)
           zxtwat=pxtp1c(jl,jk,jt)*(1._dp-zice(jl,jk))


           !--- 2.3) Calculate the scavenging parameter as sum over the processes:

           zxtfracwat=zxtfracwat_nuc(jl,jk)+zxtfracwat_imp(jl,jk)+zxtfracwat_dif(jl,jk)
           zxtfracice=zxtfracice_nuc(jl,jk)+zxtfracice_imp(jl,jk)+zxtfracice_dif(jl,jk)

           !--- Confine the fraction between 0% and 100% :
              
           zxtfracwat=MAX(0._dp,MIN(1._dp,zxtfracwat))
           zxtfracice=MAX(0._dp,MIN(1._dp,zxtfracice))


           !---  2.4) Calculate the changes in concentration and the deposition flux
           !          (R:Rain S:Snow 0:out of cloud)

           zdxtwat=0._dp
           zdxtice=0._dp

           !--- 3.1) In-cloud scavenging:

           !--- Change in in-cloud tracer concentration:

           IF(zaclc(jl,jk).GT.zmin) THEN
              zdxtwat=zxtwat*zxtfracwat*zeffwat(jl,jk)
              zdxtice=zxtice*zxtfracice*zeffice(jl,jk)
           ENDIF

           !--- In-cloud tracer concentration:

           zxtwat = zxtwat - zdxtwat
           zxtice = zxtice - zdxtice

           pxtp1c(jl,jk,jt)=zxtwat+zxtice

           !--- Local deposition mass-flux [grid-box mean kg m-2 s-1]:

           zdep=(zdxtwat+zdxtice)*pdpg(jl,jk) / ztmst

           !--- Integrated deposition mass flux:

           zdepint(jl) = zdepint(jl) + zdep


           !--- 3.2) Below cloud scavenging:

           IF(laerosol .AND. zclceff(jl,jk).GT.zmin) THEN

              zdxtcol=pxtp10(jl,jk,jt)*zclceff(jl,jk)*zxtfrac_col(jl,jk)

              pxtp10(jl,jk,jt)=pxtp10(jl,jk,jt)-zdxtcol

              zdepint(jl)=zdepint(jl)+zdxtcol*pdpg(jl,jk)/ztmst

           ENDIF


           !--- 3.3) Re-evaporation:

           zdxtevap=zdepint(jl)*zrevap(jl,jk)

           !--- Reduce integrated deposition mass flux:

           zdepint(jl)=zdepint(jl)-zdxtevap

           !--- Stratiform: attribute evaporated mass to ambient air:

           pxtp10(jl,jk,jt)=pxtp10(jl,jk,jt)+zdxtevap/pdpg(jl,jk)*ztmst


           !--- 3.4) Total tendency:

           zxtp1=pxtm1(jl,jk,jt)+pxtte(jl,jk,jt)*ztmst

           zxtte_strat(jl,jk,jt)=((pxtp10(jl,jk,jt)+pxtp1c(jl,jk,jt))-zxtp1) / ztmst

!uh+
!           pxtte(jl,jk,jt)=pxtte(jl,jk,jt)+zxtte_strat
           pxtte(jl,jk,jt)=pxtte(jl,jk,jt)+zxtte_strat(jl,jk,jt)
!uh-
           END DO
        END DO

        !--- 3.5 Diagnostics:

        DO jl=1, kproma
           wdep_p(jl,krow)    =wdep_p(jl,krow)+zdepint(jl)*delta_time
           wdep_strat_p(jl,krow)=wdep_strat_p(jl,krow)+zdepint(jl)*delta_time
           wdepinst_p(jl,krow)=zdepint(jl)
        END DO


        ELSE ! Convective:

        IF(laerosol) THEN

           !--- Nucleation and impaction:

           zxtfracwat_nuc(1:kproma,:)=csr_conv(imod)
           zxtfracice_nuc(1:kproma,:)=csr_conv(imod)

           !--- Below-cloud scavenging:
           !    Calculate fraction of below cloud scavenged tracer
           !    dM/M = -C dt where the mean scavenging coefficient C = Cn x R where Cn is normalized 
           !    scavenging coefficient (above) and R is the rain-rate kg m-2 s-1, dt is the timestep

           WHERE(zaclc(1:kproma,:) < zmin)

              zxtfrac_col(1:kproma,:)=ztmst*(zbcr(imod)*pfrain(1:kproma,:)+zbcs(imod)*pfsnow(1:kproma,:))
              zxtfrac_col(1:kproma,:)=MAX(0._dp,zxtfrac_col(1:kproma,:))

           ELSEWHERE

              zxtfrac_col(1:kproma,:)=0._dp

           ENDWHERE

        !--- 2.2.2) Gases:

        ELSE IF (lgas) THEN

           IF(trlist%ti(jt)%basename=='SO2') THEN

              !--- Gas phase scavenging utilizing the liquid fraction of total SO2 
              !    calculated from Henry's law in xt_chemistry:

              zxtfracwat_dif(1:kproma,:)=plfrac(1:kproma,:)
              zxtfracice_dif(1:kproma,:)=0._dp
                 
           ELSE IF(trlist%ti(jt)%fullname=='SO4_gas') THEN

              zxtfracwat_dif(1:kproma,:)=1.0_dp
              zxtfracice_dif(1:kproma,:)=0._dp

           END IF

        END IF

        DO  jk=ktop,klev
           DO jl=1,kproma

              !--- 2.2) Partitioning between ice and liquid phase:

              !--- Associate tracer masses in the cloud fraction to water phase
              !    and ice phase according to the respective mass fractions:

              zxtice=pxtp1c(jl,jk,jt)*zice(jl,jk)
              zxtwat=pxtp1c(jl,jk,jt)*(1._dp-zice(jl,jk))


              !--- 2.3) Calculate the scavenging parameter as sum over the processes:

              zxtfracwat=zxtfracwat_nuc(jl,jk)+zxtfracwat_imp(jl,jk)+zxtfracwat_dif(jl,jk)
              zxtfracice=zxtfracice_nuc(jl,jk)+zxtfracice_imp(jl,jk)+zxtfracice_dif(jl,jk)

              !--- Confine the fraction between 0% and 100% :
              
              zxtfracwat=MAX(0._dp,MIN(1._dp,zxtfracwat))
              zxtfracice=MAX(0._dp,MIN(1._dp,zxtfracice))


              !---  2.4) Calculate the changes in concentration and the deposition flux
              !          (R:Rain S:Snow 0:out of cloud)

              zdxtwat=0._dp
              zdxtice=0._dp

              !--- 3.1) In-cloud scavenging:

              !--- Change in updraft tracer concentration:

              IF(zaclc(jl,jk).GT.zmin) THEN
                 zdxtwat=zxtwat*zxtfracwat*zeffwat(jl,jk)
                 zdxtice=zxtice*zxtfracice*zeffice(jl,jk)
              ENDIF


              !--- Update updraft tracer concentration:

              zxtwat = zxtwat - zdxtwat
              zxtice = zxtice - zdxtice

              pxtp1c(jl,jk,jt)=zxtwat+zxtice

              !--- Local deposition mass-flux [grid-box mean kg m-2 s-1]:

              zdep=(zdxtwat+zdxtice)*pmfu(jl,jk)

              !--- Integrated deposition mass flux:

              zdepint(jl) = zdepint(jl) + zdep

              !--- Update updraft mass flux:

              pmfuxt(jl,jk,jt)=pxtp1c(jl,jk,jt)*pmfu(jl,jk)

              !--- Local tendency:

!uh+
!              zxtte_conv=-zdep/pdpg(jl,jk)
              zxtte_conv(jl,jk,jt)=-zdep/pdpg(jl,jk)
!uh-
              !--- 3.2) Below cloud scavenging:

              IF(laerosol .AND. zclceff(jl,jk).GT.zmin) THEN

                 zdxtcol=pxtp10(jl,jk,jt)*zclceff(jl,jk)*zxtfrac_col(jl,jk)
!uh+
!                 zxtte_conv=zxtte_conv-zdxtcol/ztmst
                 zxtte_conv(jl,jk,jt)=zxtte_conv(jl,jk,jt)-zdxtcol/ztmst
!uh-
                 zdepint(jl)=zdepint(jl)+zdxtcol/ztmst*pdpg(jl,jk)

              ENDIF

              !--- 3.3) Re-evaporation:

              zdxtevap=zdepint(jl)*zrevap(jl,jk)

              !--- Reduce integrated deposition mass flux:

              zdepint(jl)=zdepint(jl)-zdxtevap

              !--- Convective: change local tendency:
!uh+
!              zxtte_conv=zxtte_conv+zdxtevap/pdpg(jl,jk)
              zxtte_conv(jl,jk,jt)=zxtte_conv(jl,jk,jt)+zdxtevap/pdpg(jl,jk)
!uh-

              !--- 4.4) Total tendency:
!uh+
!              pxtte(jl,jk,jt)=pxtte(jl,jk,jt)+zxtte_conv
              pxtte(jl,jk,jt)=pxtte(jl,jk,jt)+zxtte_conv(jl,jk,jt)
!uh-
           END DO
        END DO

        !--- 4.5) Diagnostics:

        DO jl=1, kproma
           wdep_p(jl,krow)=wdep_p(jl,krow)+zdepint(jl)*delta_time
           wdep_conv_p(jl,krow)=wdep_conv_p(jl,krow)+zdepint(jl)*delta_time
           wdepinst_p(jl,krow)=wdepinst_p(jl,krow)+zdepint(jl)
        END DO

     END IF

     END IF nwetdep

  END DO tracer

!uh+
  ! ... Area average the local stratiform + convective tendency
  ! ... for Be-10 (Be-7, Cl-36).

  IF(lcn) THEN
     IF (lstrat) THEN                !... STRATIFORM
        DO jt=1,ntrac

           IF(trlist%ti(jt)%nphase==AEROSOLMASS .AND.     &
                trlist%ti(jt)%nwetdep > 0 ) THEN
              tend_mass_p => tend_mass(jt)%ptr

              DO jk =ktop,klev
                 DO jl=1,kproma

                    tend_mass_p(jl,jk,krow)=tend_mass_p(jl,jk,krow)+zxtte_strat(jl,jk,jt)
                 END DO
              END DO

           ENDIF
        END DO


     ELSE                            !... CONVECTIVE

       DO jt=1,ntrac
           IF(trlist%ti(jt)%nphase==AEROSOLMASS .AND.    &
                trlist%ti(jt)%nwetdep > 0) THEN
              tend_mass_p => tend_mass(jt)%ptr
              tend_mass_p(1:kproma,1:klev,krow)=0.
              DO jk =ktop,klev
                 DO jl=1,kproma
                    tend_mass_p(jl,jk,krow)=zxtte_conv(jl,jk,jt)
                 END DO
              END DO
           ENDIF
        END DO
     END IF !strat


!!!!!! before the following: call of xt_massbal -> ztend=ztend+zxttefix -> ok !!!!!!!!
     IF(lstrat) THEN

        DO jt=1,ntrac

           IF(trlist%ti(jt)%nphase==AEROSOLMASS .AND. trlist%ti(jt)%nwetdep > 0  ) THEN
              tend_mass_p => tend_mass(jt)%ptr
              DO jl=1,kproma
                 DO jk=ktop,klev
                    ztend(jl,jk,jt)=tend_mass_p(jl,jk,krow)
                 END DO
              END DO
           END IF
        END DO

        DO jt=1,ntrac
           IF(trlist%ti(jt)%nwetdep==BE_FLAG) THEN
              CALL wdep_be10(kproma,klev,kbdim,krow,3,pxtm1,ztend,jt)

              wdep_p       => wdep(jt)%ptr
              wdepinst_p   => wdepinst(jt)%ptr

              wdepinst_p(1:kproma,krow)=0.

              !... Save wet deposition for diagnostics for Be-10.
              DO jl=1,kproma
                 zdepint(jl)=0.
              END DO
              !!... Sum the deposition flux over the atmospheric column
              DO jk=ktop,klev
                 DO jl=1,kproma
                    zdepint(jl)=zdepint(jl)-(ztend(jl,jk,jt)*pdpg(jl,jk))
                 END DO
              END DO

              DO jl=1,kproma
                 wdep_p(jl,krow)=wdep_p(jl,krow)+zdepint(jl)*delta_time
                 wdepinst_p(jl,krow)=wdepinst_p(jl,krow)+zdepint(jl)
              END DO
              DO jk=ktop,klev
                 DO jl=1,kproma

                    pxtte(jl,jk,jt)=pxtte(jl,jk,jt)+ztend(jl,jk,jt)

                 END DO
              END DO
           END IF
        END DO


     END IF   !lstrat

     ! only defined for massbal
     IF (lstrat) THEN
        DO jt=1,ntrac
           IF (trlist%ti(jt)%nwetdep==BE_FLAG) THEN
              DO jk=1,klev
                 DO jl=1,kproma

                    IF (jt==idt_be10)  burd_wdep(jl,jk,krow)=-ztend(jl,jk,idt_be10)
                    IF (jt==idt_be7)   burd_wdep_be7(jl,jk,krow)= -ztend(jl,jk,idt_be7)
                    IF (jt==idt_cl36)  burd_wdep_cl36(jl,jk,krow)= -ztend(jl,jk,idt_cl36)
                    IF (jt==idt_na22)  burd_wdep_na22(jl,jk,krow)= -ztend(jl,jk,idt_na22)
                    IF (jt==idt_pb210) burd_wdep_pb210(jl,jk,krow)= -ztend(jl,jk,idt_pb210)
                 END DO
              END DO
           END IF
        END DO
     END IF

  END IF !(lcn)
!uh-


END SUBROUTINE xt_wetdep
